/*
 * Copyright (c) 1996 The Regents of the University of Michigan
 * All Rights Reserved
 * 
 * License to use, copy, modify, and distribute this software and its
 * documentation can be obtained from Merit at the University of Michigan.
 * 
 * Merit GateDaemon Project
 * 4251 Plymouth Road, Suite C
 * Ann Arbor, MI 48105
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE REGENTS OF THE
 * UNIVERSITY OF MICHIGAN AND MERIT DO NOT WARRANT THAT THE
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the
 * University of Michigan and Merit shall not be liable for
 * any special, indirect, incidental or consequential damages with respect
 * to any claim by Licensee or any third party arising from use of the
 * software. GateDaemon was originated and developed through release 3.0
 * by Cornell University and its collaborators.
 * 
 * Please forward bug fixes, enhancements and questions to the
 * gated mailing list: gated-people@gated.merit.edu.
 * 
 * This copyright has ben automaticly added by the util/addcopyright.pl program.
 * __END_OF_COPYRIGHT__
 */

#define INCLUDE_IF
#define INCLUDE_RT_VAR

#include "include.h"
#ifdef PROTO_OSPF2
#include "new_ospf.h"
#include "new_ospf_ls.h"
#endif
#include "gii.h"
#include "inet.h"
#include <pwd.h>
#include <arpa/telnet.h>
#include "list.h"
#include "krt.h"
#include "krt_var.h"
#ifdef IP_MULTICAST_ROUTING
#include "inet_multi.h"
#endif
#ifdef  PROTO_DVMRP
#include "dvmrp.h"
#include "dvmrp_targets.h"
#endif
#ifdef SHADOW_PASSWD
#include <shadow.h>
#endif

giimenu_cmd_t giicmd_route[] = {
	{ "route",	NULL,		gii_showrtipall,gii_showrtip,
				" [x.x.x/len]: Show info about IP routes" },
	{ "walkup",	NULL,		gii_help,	gii_showipup,
				" x.x.x/len: Show less specific routes" },
	{ "walkdown",	NULL,		gii_help,	gii_showipdown,
				" x.x.x/len: Show more specific routes" },
	{ NULL, NULL, NULL, NULL, NULL }
};

giimenu_cmd_t giicmd_dvmrp[] = {
	{ "mfc",	NULL,	gii_dvmrp_mfcall,	gii_dvmrp_mfc,
				" []: Show Multicast Forwarding Cache" },
	{ "targets",	NULL,	gii_dvmrp_targets,	gii_dvmrp_targets,
				" []: Show DVMRP targets (peers)"},
	{ NULL, NULL, NULL, NULL, NULL }
};

giimenu_cmd_t giicmd_ospf[] = {
	{ "global",	NULL,	gii_ospf_global,	gii_ospf_global,
				": Show genral OSPF options" },
	{ "interface",	NULL,	gii_ospf_allintf,	gii_ospf_intf,
				"[address]: Show interface status" },
	{ "area",	NULL,	gii_ospf_allarea,	gii_ospf_area,
				" [area ID]: Show area info" },
	{ NULL, NULL, NULL, NULL, NULL }
};

giimenu_cmd_t giicmd_show[] = {
	{ "version", 	NULL, 		NULL,		gii_showversion,
					": Show the current GateD version" },
	{ "kernel", 	NULL,		NULL,		gii_showkernel,
					": Show the Kernel support" },
	{ "interface",	NULL, 		gii_showallif,	gii_showif,
					" [name|index]: Show interface status"},
	{ "memory",	NULL,		NULL,		gii_showmem,
					": Show the memory allocation" },
	{ "ip",		giicmd_route,	gii_help,	gii_help,
					": Show info about IP protocol" },
	{ "task",	NULL,		gii_showalltask,gii_showalltask,
					": Show list of active tasks" },
	{ "dvmrp",	giicmd_dvmrp,	gii_help,	gii_help,
					": Show info about DVMRP protocol" },
	{ "ospf",	giicmd_ospf,	gii_help,	gii_help,
					": Show info about OSPF protocol" },
	{ "timer",	NULL,		gii_showalltimer,gii_showalltimer,
					": Show list of timers" },
	{ NULL, NULL, NULL, NULL, NULL }
};

giimenu_cmd_t giicmd_top[] = {
	{ "help",	NULL, 		NULL, 		gii_help,
					": Print help messages" },
	{ "show",	giicmd_show,	gii_help, 	gii_help,
					": Show internal values" },
	{ "quit",	NULL,		NULL, 		gii_quit,
					": Close the session" },
	{ NULL, NULL, NULL, NULL, NULL }
};
/* YYYYUK!
 */
giimenu_cmd_t *gii_helpcmd = &giicmd_top[0];

/* List of state names
 */
char *gii_statenames[] = {
	"GIIS_OPEN",		/* GIIS_OPEN */
	"GIIS_SESSION",		/* GIIS_SESSION */
	"GIIS_JOB",		/* A job is running */
};

trace *gii_trace_options;
const bits gii_trace_types[] = {
	{ TR_DETAIL,        "detail packets" },
    	{ TR_DETAIL_SEND,   "detail send packets" },
    	{ TR_DETAIL_RECV,   "detail recv packets" },
	{ 0, NULL }
};

task *gii_task_listen = NULL;
u_short gii_port = 0;
flag_t gii_flags = GIIF_ON;
gii_ctl_t *gii_ctl_list = NULL;
block_t gii_ctl_alloc = NULL;
char cmd_errmsg[BUFSIZ];

/* Initialisation of the Gated Interactive Interface
 */
void
gii_init __PF0(void)
{
	int s;

	/* If GII is not used, clean up!
	 */
	if (!BIT_TEST(gii_flags, GIIF_ON)) {
		gii_cleanup();
		return;
	}

	/* Get the trace options from the global trace
	 */
	trace_inherit_global(gii_trace_options, gii_trace_types, (flag_t) 0);

	/* Do we need to create a task for listening?
	 */
	if (!gii_task_listen) {

		/* Get one...
		 */
		gii_task_listen = task_alloc("GII_LISTEN", TASKPRI_NETMGMT,
			gii_trace_options);

		/* Get the port number to listen from
		 */
		if (!gii_port)
			gii_port = task_get_port(gii_trace_options,
				"gii", "tcp", htons(GII_PORT));

		/* Some method initilaisation...
		 */
		/*task_set_recv(gii_task_listen, gii_recv); */
		/*task_set_cleanup(gii_task_listen, gii_cleanup);*/
		/*task_set_reinit(gii_task_listen, gii_reinit);*/
		gii_task_listen->task_flags |= TASKF_ACCEPT;
		task_set_dump(gii_task_listen, gii_dump);
		task_set_terminate(gii_task_listen, gii_terminate);
		task_set_accept(gii_task_listen, gii_accept);

		/* And the socket...
		 */
		if ((s = task_get_socket(gii_task_listen, AF_INET,
				SOCK_STREAM, 0)) < 0) {
			s = errno;
			trace_log_tp(gii_task_listen, 0, LOG_ERR,
				("gii_init: cannot get socket"));
			task_quit(s);
		}
		task_set_socket(gii_task_listen, s);

		/* We want to reuse the address.
		 */
		if (task_set_option(gii_task_listen, TASKOPTION_REUSEADDR,
				TRUE) < 0)
			task_quit(errno);

		/* The bind() call
		 */
		gii_task_listen->task_addr = sockdup(inet_addr_any);
		sock2port(gii_task_listen->task_addr) = gii_port;
		if (task_addr_local(gii_task_listen,gii_task_listen->task_addr))
			task_quit(errno);

		/* Now listen to the socket. No task_xxx function to do that!
		 */
		if (listen(s, 5) < 0) {
			trace_log_tp(gii_task_listen, 0, LOG_ERR,
				("gii_init: cannot listen: %m"));
			task_quit(errno);
		}

		/* Start the task
		 */
		task_create(gii_task_listen);
	}

	/* Memory initialisation
	 */
	if (!gii_ctl_alloc)
		gii_ctl_alloc = task_block_init(sizeof(gii_ctl_t), "gii_ctl_t");
}

void
gii_var_init __PF0(void)
{
}

/* Accept an incoming connection
 */
void
gii_accept __PF1(listen_tp, task *)
{
	int s, inaddrlen;
	struct sockaddr_in inaddr;
	sockaddr_un *paddr;
	task *tp;
	gii_ctl_t *gii_ctl;

	/* Accept the connection
	 */
	inaddrlen = sizeof(inaddr);
	if ((s = accept(listen_tp->task_socket, (struct sockaddr *) &inaddr,
			(int *) &inaddrlen)) < 0) {
		trace_log_tp(listen_tp, 0, LOG_ERR,
			("gii_accept: accept: accept(%d): %m",
				listen_tp->task_socket));
		return;
	}

	/* Make a copy of the address for later use
	 */
	paddr = sockdup(sock2gated((struct sockaddr *) &inaddr, inaddrlen));

#ifdef TCPWRAPPER
	/* Do we want the guy?
	 */
	if (!hosts_ctl("gii", "", inet_ntoa((struct sockaddr_in *)&inaddr,
			"")) {
		trace_tp(listen_tp, TR_NORMAL, 0,
			("gii_accept: connection refused for %#A", paddr));

		/* Get rid of him!
		 */
		(void)close(s);
		return;
	}
#endif	/* TCPWRAPPER */

	trace_tp(listen_tp, TR_NORMAL, 0, ("gii_accept: connection from %#A",
		paddr));

	/* Now we create a task to take care of that connection
         */
        tp = task_alloc("GII_SESSION", TASKPRI_NETMGMT, gii_trace_options);
        task_set_socket(tp, s);
        tp->task_addr = paddr;
        if (task_set_option(tp, TASKOPTION_REUSEADDR, TRUE) < 0)
                task_quit(errno);

	/* Set all the methods
	 */
	task_set_dump(tp, gii_dump);
	task_set_terminate(tp, gii_terminate);
	task_set_recv(tp, gii_recv);

	/* Create a local structure to keep the state of the connection
	 */
	gii_ctl = (gii_ctl_t *)task_block_alloc(gii_ctl_alloc);
	gii_ctl->g_state = GIIS_OPEN;
	gii_ctl->g_iden_tries = 0;
	gii_ctl->g_task = tp;
	gii_ctl->g_cmdmenu = giicmd_top;
	gii_ctl->g_cmd = NULL;
	gii_ctl->g_cmdStr[0] = '\0';
	gii_ctl->g_bufflen = 0;
	gii_ctl->g_buffLeft = 2*GIIMAXLINELEN;
	LIST_ADD(gii_ctl_list, gii_ctl);

	tp->task_data = (caddr_t)gii_ctl;
	task_create(tp);

	/* Identify the user
	 */
	(void) gii_iden(gii_ctl);
}

/* Identify an user. In fact, just ask for the password
 */
int
gii_iden __PF1(gii_ctl, gii_ctl_t *)
{
	/* If we have tryes too many time, give up!
	 */
	if (++gii_ctl->g_iden_tries > GII_MAXTRIES) {
		trace_tp(gii_ctl->g_task, TR_NORMAL, 0,
			("gii_iden: logging failed from %#A",
			gii_ctl->g_task->task_addr));
		gii_terminate(gii_ctl->g_task);
		return(1);
	}

	/* Turn off echo on a telnet connection
	 */
	telnet_echooff(gii_ctl);

	/* Ask for a password.
	 */
	return(gii_write(gii_ctl, GW_NONE, "Password? "));
}

/* what to do when receiving a packet
 */
void
gii_recv __PF1(tp, task *)
{
	int err, lnSize;
	gii_ctl_t *gii_ctl;
	char *ln, *cr;

	/* Our control structure
	 */
	gii_ctl = (gii_ctl_t *)tp->task_data;

	/* read the data and copy it into our internal bugger
	 */
	err = read(tp->task_socket, gii_ctl->g_buff + gii_ctl->g_bufflen,
		gii_ctl->g_buffLeft);
	if (err < 0) {
		switch(errno) {
		case EWOULDBLOCK:
#if     defined(EAGAIN) && EAGAIN != EWOULDBLOCK
		case EAGAIN:
#endif  /* EAGAIN */
		case EINTR:
			/* Well, nothing to do, just return.
			 */
			return;
		default:
			/* Get rid of the peer
			 */
			trace_tp(tp, 0, LOG_ERR,
				("gii_recv: read from peer %#A failed: %m",
				tp->task_addr));
			gii_terminate(tp);
			return;
		}
	}

	if (err == 0) {
		/* End of file. Terminate.
		 */
		trace_tp(tp, TR_NORMAL, 0,
			("gii_recv: end of session from %#A", tp->task_addr));
		gii_terminate(tp);
		return;
	}

	gii_ctl->g_bufflen += err;
	gii_ctl->g_buffLeft -= err;
	telnet_strip(gii_ctl->g_buff, &gii_ctl->g_bufflen);

	/* Get as much "line" as possible, freeing the buffer
	 */
	while(gii_ctl->g_bufflen &&
			(cr = (char *)index(gii_ctl->g_buff, '\n'))) {
		*cr = '\0';
		ln = gii_ctl->g_buff;
		STRIP(ln);
		if (gii_process(gii_ctl, ln))
			return;

		lnSize = GA2S(cr) - GA2S(gii_ctl->g_buff) + 1;
		gii_ctl->g_bufflen -= lnSize;
		gii_ctl->g_buffLeft += lnSize;
		(void)bcopy(cr, gii_ctl->g_buff, lnSize);
	}
}

/* Process the line. Just a parser for the interpreter.
 * Return 1 if the task has terminated.
 */
int
gii_process __PF2(gii_ctl, gii_ctl_t *, ln, char *)
{
	struct passwd *pw;
#ifdef SHADOW_PASSWD
	struct spwd *spw;
#endif

	switch(gii_ctl->g_state) {
	case GIIS_OPEN:
		/* Echo is turned off, so echo only the last '\n'
		 */
		if (gii_write(gii_ctl, GW_NONE, "\r\n"))
			return(1);

		/* If the state is GIIS_OPEN we need to read the 
	 	 * password. But first get the crypted password.
	 	 */
		pw = getpwnam(GII_USER);
#ifdef SHADOW_PASSWD
		if (pw) {
			spw = getspnam (pw->pw_name);
			if (spw)
				pw->pw_passwd = spw->sp_pwdp;
			endspent();
		}
#endif
		if (!pw || !pw->pw_passwd[0]) {
			trace_tp(gii_ctl->g_task, 0, LOG_ERR,
				("gii_process: no password or user %s",
					GII_USER));
			if (gii_write(gii_ctl, GW_ERR,
					"Configuration error. Disconnecting!"))
				return(1);
			gii_terminate(gii_ctl->g_task);
			return(1);
		}

		/* Check the password we got
		 */
		if (strcmp(pw->pw_passwd, (char *)crypt(ln, pw->pw_passwd)))
			return(gii_iden(gii_ctl));

		/* Autentifocation succeeded. Go to state GIIS_SESSION.
		 * Also turn on the echo back.
		 */
		telnet_echoon(gii_ctl);
		if (gii_write(gii_ctl, GW_INFO,
				"Gated Interactive Interface. Version %s",
				gated_version))
			return(1);
		gii_ctl->g_state = GIIS_SESSION;
		return(gii_prompt(gii_ctl));

	case GIIS_SESSION:
		/* Parse the command.
		 */
		if (gii_parse_cmd(gii_ctl, ln))
			return(1);
		return(gii_prompt(gii_ctl));
	case GIIS_JOB:
		/* If we get a ^C or ^D or ^Z then we stop the job.
		 */
		if (index(ln, '\013') || index(ln, '\014') ||
				index(ln, '\032')) {
			ASSERT(gii_ctl->g_job && gii_ctl->g_walk);
			gii_ctl->g_walk = rt_walk_free(gii_ctl->g_walk);
			task_job_delete(gii_ctl->g_job);
			gii_ctl->g_job = NULL;
			GII_WRITE((gii_ctl, GW_INFO,
				"Current job killed"));
			return(gii_prompt(gii_ctl));
		}
		return(0);
	}
	ASSERT(0);
	return(0);
}

/* Write a message to the peer. Return 1 if the task has been terminated.
 */
int
#ifdef  STDARG
gii_write(gii_ctl_t *gii_ctl, int code, char *msg, ...)
#else   /* STDARG */
gii_write(va_alist)
	va_dcl
#endif  /* STDARG */
{
	va_list ap;
	char buf[GIIMAXLINELEN];
	char buf2[GIIMAXLINELEN];
	int err, len, try;
	
#ifdef  STDARG
	va_start(ap, msg);
#else   /* STDARG */
	gii_ctl_t *gii_ctl;
	int code;
	char *msg;
	int pri;
	
	va_start(ap);
	gii_ctl = va_arg(ap, gii_ctl_t *);
	code = va_arg(ap, int);
 	msg = va_arg(ap, const char *);
#endif  /* STDARG */

	(void) vsprintf(buf, msg, &ap);
	if (code != GW_NONE)
		(void) sprintf(buf2, "%1d00 %s\n", code, buf);
	else
		(void) strcpy(buf2, buf);
	len = strlen(buf2);

	/* write the buffer to the socket and check for errors.
	 */
#define MAXWRITETRIES 3
	try = 0;
	do {
		err = write(gii_ctl->g_task->task_socket, buf2, len);
		if (err < 0) {
			switch(errno) {
			case EWOULDBLOCK:
				/* The buffer are full, that's bad. Let's
				 * give up too
			 	 */
			case EHOSTUNREACH:
			case ENETUNREACH:
			default:
				/* give up
			 	 */
				trace_tp(gii_ctl->g_task, 0, LOG_ERR,
				("gii_write: write from peer %#A failed: %m",
					gii_ctl->g_task->task_addr));
				gii_terminate(gii_ctl->g_task);
				return(1);

#if defined(EAGAIN) && EAGAIN != EWOULDBLOCK
               		case EAGAIN:
#endif  /* EAGAIN */
               		case EINTR:
				/* Let's try again?
			 	*/
				try++;
				continue;
			}
		}
		else if (err == 0) {
			/* Connection closed?
			 */
			trace_tp(gii_ctl->g_task, 0, LOG_ERR,
			("gii_write: write: connection closed from peer %#A",
				gii_ctl->g_task->task_addr));
			gii_terminate(gii_ctl->g_task);
			return(1);
		}
		else if (err < len) {
			/* We didn't write every thing. that's bad. Let's
			 * close the connection.
			 */
			trace_tp(gii_ctl->g_task, 0, LOG_ERR,
			("gii_write: write: congestion from peer %#A",
				gii_ctl->g_task->task_addr));
			gii_terminate(gii_ctl->g_task);
			return(1);
		}

		/* We are happy, everything gas been written
		 */
		return(0);

	} while(try < MAXWRITETRIES);

	/* something was wrong. Let's given up.
	 */
	trace_tp(gii_ctl->g_task, 0, LOG_ERR,
		("gii_write: write: too many tryes (%d) from peer %#A: %m", 
		MAXWRITETRIES, gii_ctl->g_task->task_addr));
	gii_terminate(gii_ctl->g_task);
	return(1);
}

/* Terminate a session.
 */
void
gii_terminate __PF1(tp, task *)
{
	gii_ctl_t *gii_ctl;

	/* Is it a session task? If so, free the control block.
	 */
	if ((gii_ctl = (gii_ctl_t *)tp->task_data)) {
		if (gii_ctl->g_walk)
	                gii_ctl->g_walk = rt_walk_free(gii_ctl->g_walk);
		if (gii_ctl->g_job)
                	task_job_delete(gii_ctl->g_job);
                gii_ctl->g_job = NULL;
#ifdef PROTO_OSPF2
		if (gii_ctl->g_ospfwalk)
			gii_ctl->g_ospfwalk =
				ospf_pt_walkdel(gii_ctl->g_ospfwalk);
#endif

		LIST_REM(gii_ctl_list, gii_ctl);
		task_block_free(gii_ctl_alloc, (void_t)gii_ctl);
	}

	/* close the socket and free the task.
	 */
	task_delete(tp);
}

/* clean up. I.e. get read of all connections, and the listening task.
 */
void
gii_cleanup __PF0(void)
{
	gii_ctl_t *gii_ctl, *next;

	/* Get all gii_ctl. Delete all tasks.
	 */
	for(gii_ctl = gii_ctl_list; gii_ctl;) {
		task_delete(gii_ctl->g_task);
		next = gii_ctl->g_next;
		task_block_free(gii_ctl_alloc, (void_t)gii_ctl);
		gii_ctl = next;
	}
	gii_ctl_list = (gii_ctl_t *)NULL;

	/* Delete the listen task
	 */
	if (gii_task_listen)
		task_delete(gii_task_listen);
}

/* Dump a task.
 */
void
gii_dump __PF2(tp, task *, fd, FILE *)
{
	gii_ctl_t *gii_ctl;

	gii_ctl = (gii_ctl_t *)tp->task_data;

	if (!gii_ctl) {
		/* This is the listen task
		 */
		(void) fprintf(fd, "\tGII LISTEN on port %-4d\n",
			ntohs(gii_port));
		return;
	}

	/* This is a session task
	 */
	(void) fprintf(fd, "\tPeer %#A\tState %s\n",
		tp->task_addr, GIISTATESTR(gii_ctl->g_state));
}

/* Parse a command from the peer
 */
int
gii_parse_cmd __PF2(gii_ctl, gii_ctl_t *, ln, char *)
{
	int nbTok, tok;
	char tokens[GIITOKENLEN][GIINBMAXTOKENS];
	giimenu_cmd_t *cmd_table, *cmd;

	/* Separate all the tokens
	 */
	nbTok = split(ln, tokens);

	/* No token means an empty line.
	 */
	if (!nbTok)
		return(0);

	/* Find out what command   to execute
	 */
	cmd_table = giicmd_top;
	gii_ctl->g_cmdStr[0] = '\0';

	for(tok = 0; tok < nbTok; tok++) {

		/* look up the command name.
		 */
		if (!(gii_ctl->g_cmd = cmd = cmd_find(cmd_table, tokens[tok])))
			return(gii_cmd_error(gii_ctl));

		(void)strcat(gii_ctl->g_cmdStr, " ");
		(void)strcat(gii_ctl->g_cmdStr, cmd->cmd_name);

		/* If no more token, execute the 'last token' method
		 */
		if (tok == nbTok - 1) {
			trace_tp(gii_ctl->g_task, TR_NORMAL, 0,
				("Executing \"%s\"", gii_ctl->g_cmdStr));
			if (cmd->cmd_lasttok) {
				return(cmd->cmd_lasttok(gii_ctl));
			}
			else if (!cmd->cmd_nextCmdTable) {
				ASSERT(cmd->cmd_exec);
				return(cmd->cmd_exec(gii_ctl, tokens + tok, 0));
			}
			else {
				(void)sprintf(cmd_errmsg,
					"Insufficiant command \"%s\"",
					gii_ctl->g_cmdStr);
				return(gii_cmd_error(gii_ctl));
			}
		}

		/* Get the next command table. If this is the last
		 * one, excetute the command method
		 */
		if (!cmd->cmd_nextCmdTable) {
			ASSERT(cmd->cmd_exec);
			trace_tp(gii_ctl->g_task, TR_NORMAL, 0,
				("Executing \"%s\"", gii_ctl->g_cmdStr));
			return(cmd->cmd_exec(gii_ctl, tokens + tok + 1,
				nbTok - tok - 1));
		}

		gii_ctl->g_cmdmenu = cmd_table = cmd->cmd_nextCmdTable;
	}
	/* By now we should have executed something
	 */
	ASSERT(0);
	return(0);
}

int
gii_cmd_error(gii_ctl)
	gii_ctl_t *gii_ctl;
{
	return(gii_write(gii_ctl, GW_ERR, cmd_errmsg));
}

int
split(str, tokens)
	char *str;
	char tokens[GIITOKENLEN][GIINBMAXTOKENS];
{
	int tok;
	char *strTok;

	for(tok = 0, strTok = (char *)strtok(str, "\t\n ");
			strTok && tok < GIINBMAXTOKENS; tok++,
			strTok = (char *)strtok(NULL, "\t\n "))
		(void)strncpy(tokens[tok], strTok, GIITOKENLEN);
	return(tok);
}

/* Given the name of a command, find out the giimenu_cmd_t entry in
 * cmd_table which describe this command
 */
giimenu_cmd_t *
cmd_find(cmd_table, cmdStr)
	giimenu_cmd_t cmd_table[];
	char *cmdStr;
{
	char match[GIITOKENLEN];
	int nbMatch;
	giimenu_cmd_t *cmd, *rtnCmd;

	/* Go through all the entries in cmd_table
	 */
	match[0] = '\0';
	nbMatch = 0;
	for(cmd = &cmd_table[0]; cmd->cmd_name; cmd++) {

		if (strncmp(cmdStr, cmd->cmd_name, strlen(cmdStr)))
			continue;

		nbMatch++;
		rtnCmd = cmd;
		(void)strcat(match, " "); (void)strcat(match, cmd->cmd_name);
	}

	/* If we got more than one match, print an error and return NULL
	 */
	if (nbMatch > 1) {
		(void)sprintf(cmd_errmsg,
			"Unbiguous keywork \"%s\". Possible values are\"%s\"",
			cmdStr, match);
		return(NULL);
	}
	else if (!nbMatch) {
		(void)sprintf(cmd_errmsg, "Unknown command \"%s\"", cmdStr);
		return(NULL);
	}
	return(rtnCmd);
}
	
/* close the session
 */
int
gii_quit(gii_ctl)
	gii_ctl_t *gii_ctl;
{
	gii_terminate(gii_ctl->g_task);
	return(1);
}

/* Print the version
 */
int
gii_showversion(gii_ctl)
        gii_ctl_t *gii_ctl;
{
        return(gii_write(gii_ctl, GW_INFO, "Version %s (%s)", gated_version,
		build_date));
}

/* Print some help messages
 */
int
gii_help(gii_ctl)
        gii_ctl_t *gii_ctl;
{
	giimenu_cmd_t *cmd;
	char *sub = "";

	ASSERT(gii_ctl->g_cmdmenu);

	/* Top level help command
	 */
	if (gii_ctl->g_cmd == gii_helpcmd) {
		/* Default are ok.
		 */
	}
	/* Is there any sub command for that one?
	 */
	else if (gii_ctl->g_cmd && gii_ctl->g_cmd->cmd_nextCmdTable) {
		sub = "sub";
		gii_ctl->g_cmdmenu = gii_ctl->g_cmd->cmd_nextCmdTable;
	}

        if (gii_write(gii_ctl, GW_INFO, "HELP: The possible %scommands are:",
			sub))
		return(1);

	for(cmd = gii_ctl->g_cmdmenu; cmd->cmd_name; cmd++) {
		if (gii_write(gii_ctl, GW_INFO, "\t%s%s",
			cmd->cmd_name,cmd->cmd_helpMsg))
			return(1);
	}
	return(0);
}

/* Reset all command processing and print a prompt. If a job is running,
 * we do not print anything.
 */
int
gii_prompt(gii_ctl)
	gii_ctl_t *gii_ctl;
{
	gii_ctl->g_cmdmenu = giicmd_top;
	gii_ctl->g_cmd = NULL;

	if (gii_ctl->g_state == GIIS_SESSION)
		return(gii_write(gii_ctl, GW_NONE, GIIPROMPT, task_hostname));

	return(0);
}

/* Get rid of all telnet out of band comand
 */
void
telnet_strip(ln, len)
	char *ln;
	int *len;
{
	char *c;
	int left;

	for(c = ln, left = *len; left; c++, left--) {
		switch((unsigned char)*c) {
		case IAC:
			/* Get rid of the telnet option
			 */
			if (left < 3)
				break;
			(void)bcopy(c + 3, c, left - 3);
			c--;
			left -= 2;
			*len -= 3;
			break;
		default:
			break;
		}
	}
}

/* Turn off/on telnet echo at the other end
 */
int
telnet_echooff(gii_ctl)
	gii_ctl_t *gii_ctl;
{
	static char buff[] = { IAC, WILL, TELOPT_ECHO, '\0' };

	return(gii_write(gii_ctl, GW_NONE, buff));
}

int
telnet_echoon(gii_ctl)
        gii_ctl_t *gii_ctl;
{
        static char buff[] = { IAC, WONT, TELOPT_ECHO, '\0' };

        return(gii_write(gii_ctl, GW_NONE, buff));
}

int
gii_showkernel(gii_ctl)
        gii_ctl_t *gii_ctl;
{
	
	if (gii_write(gii_ctl, GW_INFO, "Kernel options: <%s> Support: <%s>",
			trace_bits(kernel_option_bits, krt_options),
			trace_bits(kernel_support_bits, krt_rt_support)))
		return(1);
	if (gii_write(gii_ctl, GW_INFO, "IP forwarding: %d UDP checksums %d",
			inet_ipforwarding, inet_udpcksum))
		return(1);
	return(0);
}

/* Print interface info. First parameter is the interface name or index
 */
int
gii_showif(gii_ctl, tokens, nbTok)
	gii_ctl_t *gii_ctl;
	char tokens[GIITOKENLEN][GIINBMAXTOKENS];
	int nbTok;
{
	if_link *ifl;
	if_addr *ifap;
	int proto;

	assert(nbTok);

	/* We care about the first interface name
	 */
	if (nbTok > 1)
		return(gii_write(gii_ctl, GW_ERR,
		"Syntax error (Syntax: show interface [name|index])"));

	/* Try with the name first...
	 */
	if (!(ifl = ifl_locate_name(tokens[0], strlen(tokens[0]))) &&
			!(ifl = ifl_locate_index(atoi(tokens[0]))))
		return(gii_write(gii_ctl, GW_ERR, "%s: No such interface",
			tokens[0]));

	/* Print the full dump of an interface. All from if_dump().
	 */
	GII_WRITE((gii_ctl, GW_INFO, "%s\tIndex %u%s%A",
		ifl->ifl_name, ifl->ifl_index,
		ifl->ifl_addr ? " Address ":" ",
		ifl->ifl_addr ? ifl->ifl_addr : sockbuild_str("")));
	GII_WRITE((gii_ctl, GW_INFO, "\tChange: <%s>\tState: <%s>",
		trace_bits(if_change_bits, ifl->ifl_change),
		trace_bits(if_state_bits, ifl->ifl_state)));
	GII_WRITE((gii_ctl, GW_INFO, "\tRefcount: %d\tUp-down transitions: %u",
		ifl->ifl_refcount, ifl->ifl_transitions));

	IF_ADDR(ifap) {
		if (ifap->ifa_link != ifl)
			continue;

		GII_WRITE((gii_ctl, GW_INFO,
			"\t%A\tMetric: %d\tMTU: %d",
			ifap->ifa_addr, ifap->ifa_metric,
			ifap->ifa_mtu));
		GII_WRITE((gii_ctl, GW_INFO,
			"\t\tRefcount: %d\tPreference: %d\tDown: %d",
			ifap->ifa_refcount, ifap->ifa_preference,
			ifap->ifa_preference_down));
		GII_WRITE((gii_ctl, GW_INFO,
			"\t\tChange: <%s>\tState: <%s>",
			trace_bits(if_change_bits, ifap->ifa_change),
			trace_bits(if_state_bits, ifap->ifa_state)));
		GII_WRITE((gii_ctl, GW_INFO,
			"\t\tBroadcast Address: %A\tLocal %A",
			ifap->ifa_addr_broadcast? ifap->ifa_addr_broadcast:
				sockbuild_str(""),
			ifap->ifa_addr_local? ifap->ifa_addr_local:
				sockbuild_str("")));
		GII_WRITE((gii_ctl, GW_INFO,
			"\t\tSubnet Number: %A\tSubnet Mask: %A",
			ifap->ifa_net? ifap->ifa_net: sockbuild_str(""),
			ifap->ifa_netmask?ifap->ifa_netmask:sockbuild_str("")));

#ifdef  PROTO_ASPATHS
		GII_WRITE((gii_ctl, GW_INFO, "\t\tAutonomous System: %u",
			ifap->ifa_as));
#endif  /* PROTO_ASPATHS */

		GII_WRITE((gii_ctl, GW_INFO,
			"\t\tRouting protocols active: %s",
			trace_state_all(ifap->ifa_rtactive)));

		for (proto = 0; proto < RTPROTO_MAX; proto++) {
			struct ifa_ps *ips = &ifap->ifa_ps[proto];

			if (!BIT_TEST(ifap->ifa_rtactive, RTPROTO_BIT(proto)))
				continue;
			GII_WRITE((gii_ctl, GW_INFO,
		"\t\tproto: %-6.6s Metricin: %-3u Metricout: %-3u State: <%s>",
				trace_state(rt_proto_bits, proto),
				ips->ips_metric_in, ips->ips_metric_out,
				trace_bits2(if_proto_bits, int_ps_bits[proto],
					ips->ips_state)));
		}
	} IF_ADDR_END(ifap);

	return(0);
}

char *
trace_state_all(proto)
	flag_t proto;
{
	int i = RTPROTO_MAX;
	static char lbuff[BUFSIZ] = { '\0' };

        while (i--)
                if (BIT_TEST(proto, RTPROTO_BIT(i))) {
			(void)strcat(lbuff, " ");
			(void)strcat(lbuff, trace_state(rt_proto_bits, i));
		};
	return(lbuff);
}

/* Print a list of all available interfaces.
 */
int
gii_showallif(gii_ctl)
        gii_ctl_t *gii_ctl;
{
	if_link *ifl;
	if_addr *ifap;

	GII_WRITE((gii_ctl, GW_INFO,
		"#ind name     address         mtu       flags"));

	IF_LINK(ifl) {

		IF_ADDR(ifap) {

			if (ifap->ifa_link != ifl)
				continue;

			GII_WRITE((gii_ctl, GW_INFO,
				"#%-3u %-8s %-15A %4d/%-4d %s",
				ifl->ifl_index, ifl->ifl_name,
				ifap->ifa_addr,
				ifl->ifl_mtu, ifap->ifa_mtu,
				trace_bits(if_state_bits, ifl->ifl_state)));

		} IF_ADDR_END(ifap);

	} IF_LINK_END(ifl);

	return(0);
}


/* Show memory allocation.
 */
int
gii_showmem(gii_ctl)
        gii_ctl_t *gii_ctl;
{
	int freemem = 0, used = 0;
	struct task_size_block *tsb = task_block_head.tsb_forw;

	GII_WRITE((gii_ctl, GW_INFO, "Allocation size: %5d",
		task_pagesize));

	GII_WRITE((gii_ctl, GW_INFO,
   "Bck Size Free   Block Name        Init   Alloc  Free   InUse  (bytes)"));
/*0    12345 123456 1234567890123456  123456 123456 123456 123456 ()*/

	do {
		struct task_block *tbp = &tsb->tsb_block;

		freemem += tsb->tsb_size * tsb->tsb_n_free;

		do {
			GII_WRITE((gii_ctl, GW_INFO,
			"   %-5u %-6u %-16.16s  %-6u %-6u %-6u %-6u (%d)",
				tsb->tsb_size, tsb->tsb_n_free,
				tbp->tb_name, tbp->tb_n_init,
				tbp->tb_n_alloc, tbp->tb_n_free,
				tbp->tb_n_alloc - tbp->tb_n_free,
				(tbp->tb_n_alloc - tbp->tb_n_free)
					* tsb->tsb_size));
			/*freemem += tbp->tb_n_free * tsb->tsb_size;*/
			used += (tbp->tb_n_alloc - tbp->tb_n_free)
				* tsb->tsb_size;
		} while ((tbp = tbp->tb_next));
	} while ((tsb = tsb->tsb_forw) != &task_block_head);

	GII_WRITE((gii_ctl, GW_INFO,
		"Total Memory: %-9d Total Free: %-9d Total Allocated: %-9d",
		freemem + used, freemem, used));

	return(0);
}

int
gii_showrtip(gii_ctl, tokens, nbTok)
        gii_ctl_t *gii_ctl;
        char tokens[GIITOKENLEN][GIINBMAXTOKENS];
        int nbTok;
{
	int i, proto, nbHop, count;
	rt_head *rth;
	rt_entry *rt;
	sockaddr_un *addr, *mask;
	char active, lbuff[BUFSIZ];
	char lbuff2[BUFSIZ];
	
	/* The only argument should be an ip address in the format
	 * X.X.X.X/len
	 */
	if (nbTok != 1 || (sockstr(tokens[0], &addr, &mask)))
		return(gii_write(gii_ctl, GW_ERR,
			"Syntax error: show route ip [x.x.x.x/len]"));

	/* find and print the route
	 */
	if (!(rth = rt_table_locate(addr, mask)))
		return(gii_write(gii_ctl, GW_ERR,
			"No IP route %A mask %A", addr, mask));

	GII_WRITE((gii_ctl, GW_INFO,
		"Route %A - %A entries %d Announced %d Depth %d <%s>",
		rth->rth_dest, rth->rth_dest_mask, rth->rth_entries,
		rth->rth_n_announce, rth->rth_aggregate_depth,
		trace_bits(rt_state_bits, rth->rth_state)));
	GII_WRITE((gii_ctl, GW_INFO,
  "  Proto Route          Next Hop        Preference/2 Metric/2 etc..."));
/*"* OSPF 192.168.10.100  192.168.10.100  1/0     0        17:15:22 IGP (Id 1) <Int Active Gateway> */

	/* Print all sources,
	 */
	RT_ALLRT(rt, rth) {
		if (rt == rth->rth_active && rt == rth->rth_last_active)
			active = '*';
		else if (rt == rth->rth_active)
			active = '+';
		else if (rt == rth->rth_last_active)
			active = '-';
		else
			active = ' ';

		/* by next hop... We force at least 'one' next hop.
		 */
		nbHop = rt->rt_n_gw? rt->rt_n_gw: 1;
		count = 0;
		for (i = 0; i < nbHop; i++) {
			/* ... and protocol. We force at least 'one' proto.
			 */
			(void)sprintf(lbuff2, "%d/%d", rt->rt_preference,
				rt->rt_preference2);
			(void)sprintf(lbuff, "%d/%d", rt->rt_metric,
				rt->rt_metric2);

#ifdef PROTO_ASPATHS
			GII_WRITE((gii_ctl, GW_INFO,
				"%c %-4.4s %-15A %-15A %-7s %-13s %-8X %8T %s <%s>",
				(!count++)? active: '|',
				(rt->rt_gwp)? trace_state(rt_proto_bits,
					rt->rt_gwp->gw_proto): "---",
/*
		(!rt->rt_n_bitsset)? "---":
		task_name(rtbit_map[proto-1].rtb_task),
 */
				(rt->rt_routers[i])? rt->rt_routers[i]:
					sockbuild_str("---"),
				(rt->rt_gwp && rt->rt_gwp->gw_addr)?
					rt->rt_gwp->gw_addr:
						 sockbuild_str("---"),
				lbuff2, lbuff,
				rt->rt_tag,
				rt_age(rt),
				aspath_str(rt->rt_aspath),
				trace_bits(rt_state_bits, rt->rt_state)
				));
#else
			GII_WRITE((gii_ctl, GW_INFO,
				"%c %-4.4s %-15A %-15A %-7s %-13s %-8X %8T %s <%s>",
				(!count++)? active: '|',
				(rt->rt_gwp)? trace_state(rt_proto_bits,
					rt->rt_gwp->gw_proto): "---",
				(rt->rt_routers[i])? rt->rt_routers[i]:
					sockbuild_str("---"),
				(rt->rt_gwp && rt->rt_gwp->gw_addr)?
					rt->rt_gwp->gw_addr:
						 sockbuild_str("---"),
				lbuff2, lbuff,
				rt->rt_tag,
				rt_age(rt),
				"",
				trace_bits(rt_state_bits, rt->rt_state)
				));
#endif     /* PROTO_ASPATHS  */
					
/** STA 12.67.01.34 23.57.01.35 13/57 00 12:45:67 AS34 AS1800 IGP <Retain>*/


		}
	} RT_ALLRT_END(rt, rth);

	return(0);
}

/* display info about the IP route radix tree
 */
int
gii_showrtipall(gii_ctl)
        gii_ctl_t *gii_ctl;
{
	GII_WRITE((gii_ctl, GW_INFO,
		"IP radix tree: %d nodes, %d routes",
		rt_table_nodes(AF_INET), rt_table_routes(AF_INET)));
		
	return(0);
}

int
gii_showalltask(gii_ctl)
        gii_ctl_t *gii_ctl;
{
	task *tp;

	TASK_TABLE(tp) {

		GII_WRITE((gii_ctl, GW_INFO,
			"%-10.10s %-3d %-3d %-15A %-4d %2d <%s - %s>",
			tp->task_name, tp->task_proto, 
			tp->task_priority, 
			(tp->task_addr)? tp->task_addr: sockbuild_str("---"),
			(tp->task_addr)? ntohs(sock2port(tp->task_addr)): -1,
			tp->task_socket,
			trace_state(rt_proto_bits, tp->task_rtproto),
			trace_bits(task_flag_bits, tp->task_flags)));

	} TASK_TABLE_END(tp);

	return(0);
}

#define GII_TIMER_DUMP(tip, tpname) do { \
	GII_WRITE((gii_ctl, GW_INFO, \
	"%-20.20s %-8.8s %02d:%02ds %02d:%02ds %02d:%02ds %02d:%02ds <%s>", \
		(tip)->task_timer_name, tpname, \
		BIT_TEST(tip->task_timer_flags, TIMERF_ONESHOT)? 0: \
			(time_sec - tip->task_timer_last_time) / 60, \
		BIT_TEST(tip->task_timer_flags, TIMERF_ONESHOT)? 0: \
			(time_sec - tip->task_timer_last_time) % 60, \
		BIT_TEST(tip->task_timer_flags, TIMERF_INACTIVE)? 0: \
			(tip->task_timer_next_time - time_sec) / 60, \
		BIT_TEST(tip->task_timer_flags, TIMERF_INACTIVE)? 0: \
			(tip->task_timer_next_time - time_sec) % 60, \
		tip->task_timer_interval /60, tip->task_timer_interval % 60, \
		tip->task_timer_jitter /60, tip->task_timer_jitter % 60, \
		trace_bits(task_timer_flag_bits, tip->task_timer_flags))); \
} while(0)

/* Show all timers defined in gated.
 */
int
gii_showalltimer(gii_ctl)
        gii_ctl_t *gii_ctl;
{
	task *tp;
	task_timer *tip;

	GII_WRITE((gii_ctl, GW_INFO,
"Name                           Task     Last   Next   Intrvl Jitter flags"));

	/* Timer are listed per task.
	 */
	TASK_TABLE(tp) {
		for (tip = tp->task_timers; tip; tip = tip->task_timer_next)
			GII_TIMER_DUMP(tip, tp->task_name);

	} TASK_TABLE_END(tp);

	/* and are also global!
	 */
	for (tip = task_head.task_timers; tip; tip = tip->task_timer_next)
		GII_TIMER_DUMP(tip, "GLOBAL");

	return(0);
}

int
gii_showipup(gii_ctl, tokens, nbTok)
        gii_ctl_t *gii_ctl;
        char tokens[GIITOKENLEN][GIINBMAXTOKENS];
        int nbTok;
{
	sockaddr_un *addr, *mask;
	rt_head *rth;

	/* The only argument should be an ip address in the format
	 * X.X.X.X/len
	 */
	if (nbTok != 1 || (sockstr(tokens[0], &addr, &mask)))
                return(gii_write(gii_ctl, GW_ERR,
                        "Syntax error: show route ip [x.x.x.x/len]"));

	/* find and print the route
         */
	if (!(rth = rt_walk_start(addr, mask, RTW_UP)))
		return(gii_write(gii_ctl, GW_ERR,
			"No less specific IP route %A mask %A", addr, mask));

	/* No we create a job to walk up the tree. Everytime the job
	 * is dispatched, one route is printed.
	 */
	gii_ctl->g_walk = rt_walk_alloc();
	gii_ctl->g_walk->rw_rth = rth;
	gii_ctl->g_walk->rw_way = RTW_UP;
	gii_ctl->g_walk->rw_len = mask_bits(mask);
	assert(!gii_ctl->g_job);
	gii_ctl->g_job = task_job_create(gii_ctl->g_task, TASK_JOB_PRIO_WORST,
		"GII_WALKUP", gii_job_walk, (void_t)gii_ctl);
	gii_ctl->g_state = GIIS_JOB;
	return(0);
}

/* Print all more specific routes.
 */
int
gii_showipdown(gii_ctl, tokens, nbTok)
        gii_ctl_t *gii_ctl;
        char tokens[GIITOKENLEN][GIINBMAXTOKENS];
        int nbTok;
{
        sockaddr_un *addr, *mask;
        rt_head *rth;

        /* The only argument should be an ip address in the format
         * X.X.X.X/len
         */
        if (nbTok != 1 || (sockstr(tokens[0], &addr, &mask)))
                return(gii_write(gii_ctl, GW_ERR,
                        "Syntax error: show route ip [x.x.x.x/len]"));

        /* find and print the route
         */
        if (!(rth = rt_walk_start(addr, mask, RTW_DOWN)))
                return(gii_write(gii_ctl, GW_ERR,
                        "No more specific IP route %A mask %A", addr, mask));

        /* No we create a job to walk up the tree. Everytime the job
         * is dispatched, one route is printed.
         */
        gii_ctl->g_walk = rt_walk_alloc();
        gii_ctl->g_walk->rw_rth = rth;
        gii_ctl->g_walk->rw_way = RTW_DOWN;
        gii_ctl->g_walk->rw_len = mask_bits(mask);
        assert(!gii_ctl->g_job);
        gii_ctl->g_job = task_job_create(gii_ctl->g_task, TASK_JOB_PRIO_WORST,
                "GII_WALKDOWN", gii_job_walk, (void_t)gii_ctl);
        gii_ctl->g_state = GIIS_JOB;
        return(0);
}

#define RTWALK_BULK	10
void
gii_job_walk __PF1(tjp, task_job *)
{
	gii_ctl_t *gii_ctl;
	rt_head *rth;
	int ind;

	gii_ctl = (gii_ctl_t *)tjp->task_job_data;

	/* We print a bulk of routes, then exit.
	 */
	for(ind = 0; ind < RTWALK_BULK; ind++) {
		/* Are we at the end of the tree? Then delete the job & data.
	 	 */
		if (!(rth = gii_ctl->g_walk->rw_rth)) {
			gii_ctl->g_walk = rt_walk_free(gii_ctl->g_walk);
			task_job_delete(gii_ctl->g_job);
			gii_ctl->g_job = NULL;
			gii_ctl->g_state = GIIS_SESSION;
			gii_prompt(gii_ctl);
			return;
		}

		/* Print the route
	 	 */
		if (gii_write(gii_ctl, GW_INFO,
			"%-3.3s %15A/%-2d %-15A %s",
			(rth->rth_active && rth->rth_active->rt_gwp)?
				trace_state(rt_proto_bits,
					rth->rth_active->rt_gwp->gw_proto):
				"---",
			rth->rth_dest,
			mask_bits(rth->rth_dest_mask),
			(rth->rth_active && rth->rth_active->rt_routers[0])?
				rth->rth_active->rt_routers[0]:
				sockbuild_str("---"),
#ifdef PROTO_ASPATHS
				(rth->rth_active && rth->rth_active->rt_aspath)?
				aspath_str(rth->rth_active->rt_aspath): "")) {
#else
				"")) {
#endif
			gii_terminate(gii_ctl->g_task);
			return;
		}
	
		/* Get the next one...
	 	 */
		gii_ctl->g_walk->rw_rth = rt_walk(gii_ctl->g_walk);
	}
}

/* List the DVMDP MFCs
 */
int
gii_dvmrp_mfcall(gii_ctl)
        gii_ctl_t *gii_ctl;
{
#ifndef PROTO_DVMRP
	return(GII_NOAVAILABLE);
#else
	mfc_list *pm;
	mfc *mfcp;

	GII_WRITE((gii_ctl, GW_INFO,
 "Group           Source          Count     Timeout  Refresh  Interface"));
/*123.567.901.345 123.567.901.345 123456789 12:45:78 12:45:78 123.567.901.345 */

	for(pm = dvmrp_mfc_head.forw; pm != &dvmrp_mfc_head; pm = pm->forw) {

		mfcp = pm->mfcp;
		GII_WRITE((gii_ctl, GW_INFO,
			"%-15A %-15A %-9d %8T %8T %-15A",
			sockbuild_in(0, mfcp->mfc_group->group_key),
			sockbuild_in(0, mfcp->mfc_src),
			mfcp->mfc_use, mfcp->mfc_rtime, pm->mfc_timeout,
			mfcp->upstream_ifap->ifa_addr));
	}
	return(0);
#endif
}


/* List one DVMDP MFC
 */
int
gii_dvmrp_mfc(gii_ctl, tokens, nbTok)
        gii_ctl_t *gii_ctl;
        char tokens[GIITOKENLEN][GIINBMAXTOKENS];
        int nbTok;
{
	return(0);
}

int
gii_dvmrp_targets(gii_ctl)
	gii_ctl_t *gii_ctl;
{
#ifndef PROTO_DVMRP
        return(GII_NOAVAILABLE);
#else
	dvmrp_target *tlp;
	int nbRoutes, i;

	GII_WRITE((gii_ctl, GW_INFO,
"Destination     Source          Interface       Gateway         Rtes  flags"));
/*123.123.123.123 123.123.123.123 123.123.123.123 123.123.123.123 12345 <...> */

	TARGET_LIST(tlp, &dvmrp_targets) {

		nbRoutes = 0;
		for(i = 0;  i < INETMAXMASKLEN; i++)
			nbRoutes += tlp->target_td_count[i];

		GII_WRITE((gii_ctl, GW_INFO,
		"%-15A %-15A %-15A %-15A %-5d %s",
			(tlp->target_dst && *tlp->target_dst)? *tlp->target_dst:
				sockbuild_str("---"),
			(tlp->target_src && *tlp->target_src)? *tlp->target_src:
				sockbuild_str("---"),
			(tlp->target_ifap && tlp->target_ifap->ifa_addr)?
				tlp->target_ifap->ifa_addr:
				sockbuild_str("---"),
			(tlp->target_gwp && tlp->target_gwp->gw_addr)?
				tlp->target_gwp->gw_addr:
				sockbuild_str("---"),
			nbRoutes,
			trace_bits(target_flag_bits, tlp->target_flags)));

	} TARGET_LIST_END(tlp, &dvmrp_targets);

	return(0);
#endif
}

/* Print global OSPF options
 */
int
gii_ospf_global(gii_ctl)
        gii_ctl_t *gii_ctl;
{
#ifndef PROTO_OSPF2
	return(GII_NOAVAILABLE);
#else
	GII_WRITE((gii_ctl, GW_INFO, "OSPF protocol: %s, Preference: %d",
		BIT_TEST(ospf_flags, FOSPF_ON)? "ON": "OFF", ospf_preference));

	if (!BIT_TEST(ospf_flags, FOSPF_ON))
		return(0);

	GII_WRITE((gii_ctl, GW_INFO, "Router ID: %A, Flags <%s>",
		sockbuild_in(0, ospf_rtid),
		trace_bits(ospf_rtr_bits, ospf_rtrbits)));

	return(0);
#endif
}

/* List all interfaces and their OSPF parameters
 */
int
gii_ospf_allintf(gii_ctl)
        gii_ctl_t *gii_ctl;
{
#ifndef PROTO_OSPF2
	return(GII_NOAVAILABLE);
#else
	ospf_intf_t *ospf_intf;

        GII_WRITE((gii_ctl, GW_INFO,
"#  name     address         state    area            pri cost  N Designed Rtr    Backup DR        HI  RDI   RI   TD  Options"));
/*2 tunnel00 123.567.901.345 12345678 123.567.901.345 123 12345 1 123.567.901.345 123.567.901.345 123s 123s 123s 123s 1...*/

	OSPF_IF_ADDR(ospf_intf) {
		/* Is ospf running? For some interfaces defined in the config
		 * file, the ifa_link may be null. For virtual interfaces, the
		 * ospf_if_ifap may be null.
		 */
		GII_WRITE((gii_ctl, GW_INFO,
"%-2d %-8s %-15A %-8s %-15A %-3d %-5d %1d %-15A %-15A %3ds %3ds %3ds %3ds <%s>",
			ospf_intf->ospf_if_ifap && ospf_intf->ospf_if_link?
				ospf_intf->ospf_if_ifap->ifa_link->ifl_index:
				-1,
			ospf_intf->ospf_if_ifap && ospf_intf->ospf_if_link?
				ospf_intf->ospf_if_ifap->ifa_link->ifl_name:
				OSPF_ISVIRTUAL(ospf_intf)? "VIR": "---",
			sockbuild_in(0, ospf_intf->ospf_if_addr),
			trace_value(ospf_if_states, ospf_intf->ospf_if_state),
			sockbuild_in(0, ospf_intf->ospf_if_area->area_id),
			ospf_intf->ospf_if_rtpri,
			ospf_intf->ospf_if_cost,
			ospf_count_ngb(ospf_intf),
			sockbuild_in(0, ospf_intf->ospf_if_drID),
			sockbuild_in(0, ospf_intf->ospf_if_backupDrID),
			ospf_intf->ospf_if_helloInt,
			ospf_intf->ospf_if_rtdeadInt,
			ospf_intf->ospf_if_rmxInt,
			ospf_intf->ospf_if_transDel,
			trace_bits(ospf_pck_options,
				ospf_intf->ospf_if_options)));
	} OSPF_IF_ADDR_END(ospf_intf);
	return(0);
#endif
}

/* Show a detailled interface structure. Includes all neighbours
 */
int
gii_ospf_intf(gii_ctl, tokens, nbTok)
        gii_ctl_t *gii_ctl;
        char tokens[GIITOKENLEN][GIINBMAXTOKENS];
        int nbTok;
{
#ifndef PROTO_OSPF2
        return(GII_NOAVAILABLE);
#else
        ospf_intf_t *ospf_intf;
	sockaddr_un *addr, *mask;
	ospf_ngb_t *ngb;

        assert(nbTok);

        /* We care about the first interface name
         */
        if (nbTok > 1 || sockstr(tokens[0], &addr, &mask))
                return(gii_write(gii_ctl, GW_ERR,
                "Syntax error (Syntax: show ospf interface [address])"));

        /* Find out about the interface
         */
	if (!(ospf_intf = ospf_if_find(SOCK32(addr))))
                return(gii_write(gii_ctl, GW_ERR, "%s: No such interface",
                        tokens[0]));

        /* Print the full dump of this interface.
         */
	GII_WRITE((gii_ctl, GW_INFO,
		"Index: %d Name: %s Address: %A State: %s",
		ospf_intf->ospf_if_ifap && ospf_intf->ospf_if_link?
                        ospf_intf->ospf_if_ifap->ifa_link->ifl_index:
			-1,
                ospf_intf->ospf_if_ifap && ospf_intf->ospf_if_link?
                        ospf_intf->ospf_if_ifap->ifa_link->ifl_name:
			OSPF_ISVIRTUAL(ospf_intf)? "VIRTUAL": "---",
                sockbuild_in(0, ospf_intf->ospf_if_addr),
		trace_value(ospf_if_states, ospf_intf->ospf_if_state)));
	if (ospf_intf->ospf_if_state == OSPF_IFS_NOOSPF)
		return(0);

	/* Be carefull, the physical interface may not be defined yet.
	 */
	if (ospf_intf->ospf_if_ifap)
		GII_WRITE((gii_ctl, GW_INFO,
	"MTU: %d DDinPacket: %d LSRinPacket: %d ACKinPacket: %d",
			ospf_intf->ospf_if_mtu,
			ospf_intf->ospf_if_maxDDinPacket,
			ospf_intf->ospf_if_maxLSRinPacket,
			ospf_intf->ospf_if_maxACKinPacket));
	else
		GII_WRITE((gii_ctl, GW_INFO,
			"*** MTU:  No physical interface defined yet ***"));

	GII_WRITE((gii_ctl, GW_INFO,
		"Area: %A Designed router: %A Backup: %A",
		sockbuild_in(0, ospf_intf->ospf_if_area->area_id),
		sockbuild_in(0, ospf_intf->ospf_if_drID),
		sockbuild_in(0, ospf_intf->ospf_if_backupDrID)));
	GII_WRITE((gii_ctl, GW_INFO,
		"Priority: %d Cost: %d Options: %s",
		ospf_intf->ospf_if_rtpri,
		ospf_intf->ospf_if_cost,
		trace_bits(ospf_pck_options,ospf_intf->ospf_if_options)));
	GII_WRITE((gii_ctl, GW_INFO,
	"Intervals: Hello: %ds Dead Router: %ds Retransmission: %d Delay %ds",
		ospf_intf->ospf_if_helloInt,
		ospf_intf->ospf_if_rtdeadInt,
		ospf_intf->ospf_if_rmxInt,
		ospf_intf->ospf_if_transDel));
	GII_WRITE((gii_ctl, GW_INFO, "Neighbours (%d):",
		ospf_count_ngb(ospf_intf)));
	GII_WRITE((gii_ctl, GW_INFO,
"Address         State    RouterID        Pri Dr              Backup DR          DS        LSR       RMX   (Len)   Options",
/*23.567.901.345 12345678 123.567.901.345 123 123.567.901.345 123.567.901.345 ...*/
		ospf_count_ngb(ospf_intf)));

	OSPF_INTF_NGB(ngb, ospf_intf) {
		GII_WRITE((gii_ctl, GW_INFO,
"%-15A %-8.8s %-15A %-3d %-15A %-15A %4d:%-4d %4d:%-4d %4d:%-4d (%d) %s",
			sockbuild_in(0, ngb->ngb_addr),
			trace_value(ospf_ngb_states, ngb->ngb_state),
			sockbuild_in(0, ngb->ngb_rtid),
			ngb->ngb_rtpri,
			sockbuild_in(0, ngb->ngb_dr),
			sockbuild_in(0, ngb->ngb_backupDr),
			ngb->ngb_DSlistSize, ngb->ngb_lastDDsize,
			ngb->ngb_lsReq.pt_used_nodes,
			ngb->ngb_LSRbuffer.pt_used_nodes,
			ngb->ngb_lsRmx.pt_used_nodes,
			ngb->ngb_RMXbuffer.pt_used_nodes,
			ngb->ngb_RMXbuffer_len,
			trace_bits(ospf_pck_flags, ngb->ngb_options)));
	}
	return(0);
#endif
}

/* Dump the list of configured area.
 */
int
gii_ospf_allarea(gii_ctl)
        gii_ctl_t *gii_ctl;
{
#ifndef PROTO_OSPF2
        return(GII_NOAVAILABLE);
#else
	ospf_area_t *area;

	GII_WRITE((gii_ctl, GW_INFO,
"ID              RTR announce    TreeSize Routers  Networks SPFcount flags"));

	OSPF_ALL_AREA(area) {
		GII_WRITE((gii_ctl, GW_INFO,
			"%-15A %-15A %-8d %-8d %-8d %-3d <%s>",
			sockbuild_in(0, area->area_id),
			area->area_rtrls?
				sockbuild_in(0, area->area_rtrls->vtx_id):
				sockbuild_str("---"),
			area->area_tree.pt_used_nodes,
			area->area_graph.gr_nbVtx[LST_RTR],
			area->area_graph.gr_nbVtx[LST_NTW],
			area->area_spfcount,
			trace_bits(ospf_rtr_bits, area->area_rtrbits)));
	} OSPF_ALL_AREA_END(area);
	return(0);
#endif
}

/* Dump an area, including the whole database. Will create a BG job so we can
 * dump a small portion at a time.
 */
int
gii_ospf_area(gii_ctl, tokens, nbTok)
	gii_ctl_t *gii_ctl;
	char tokens[GIITOKENLEN][GIINBMAXTOKENS];
        int nbTok;
{
#ifndef PROTO_OSPF2
        return(GII_NOAVAILABLE);
#else
	sockaddr_un *addr, *mask;
	ospf_area_t *area;
	ospf_intf_t *ospf_intf;

	/* Parse the argument, which should be the area ID.
	 */
	if (nbTok != 1 || (sockstr(tokens[0], &addr, &mask)))
                return(gii_write(gii_ctl, GW_ERR,
                        "Syntax error: show ospf area x.x.x.x"));

	/* Find the area...
	 */
	if (!(area = ospf_find_area(SOCK32(addr))))
		return(gii_write(gii_ctl, GW_ERR,
			"No such area %A", addr));

	GII_WRITE((gii_ctl, GW_INFO,
		"Area: %A, Authentication: %s, Interfaces:",
		sockbuild_in(0, area->area_id),
		trace_state(ospf_auth_types, area->area_auth)));

	GII_WRITE((gii_ctl, GW_INFO, ""));

        GII_WRITE((gii_ctl, GW_INFO,
"#  name     address         state    area            pri cost  N Designed Rtr    Backup DR        HI  RDI   RI   TD  Options"));
/*2 tunnel00 123.567.901.345 12345678 123.567.901.345 123 12345 1 123.567.901.345 123.567.901.345 123s 123s 123s 123s 1...*/

	OSPF_IF_ADDR_AREA(area, ospf_intf) {
                GII_WRITE((gii_ctl, GW_INFO,
"%2d %-8s %-15A %-8s %-15A %-3d %-5d %1d %-15A %-15A %3ds %3ds %3ds %3ds <%s>"
,
                        ospf_intf->ospf_if_ifap && ospf_intf->ospf_if_link?
                                ospf_intf->ospf_if_ifap->ifa_link->ifl_index:
                                -1,
                        ospf_intf->ospf_if_ifap && ospf_intf->ospf_if_link?
                                ospf_intf->ospf_if_ifap->ifa_link->ifl_name:
                                OSPF_ISVIRTUAL(ospf_intf)? "VIR": "---",
                        sockbuild_in(0, ospf_intf->ospf_if_addr),
                        trace_value(ospf_if_states, ospf_intf->ospf_if_state),
                        sockbuild_in(0, ospf_intf->ospf_if_area->area_id),
                        ospf_intf->ospf_if_rtpri,
                        ospf_intf->ospf_if_cost,
                        ospf_count_ngb(ospf_intf),
                        sockbuild_in(0, ospf_intf->ospf_if_drID),
                        sockbuild_in(0, ospf_intf->ospf_if_backupDrID),
                        ospf_intf->ospf_if_helloInt,
                        ospf_intf->ospf_if_rtdeadInt,
                        ospf_intf->ospf_if_rmxInt,
                        ospf_intf->ospf_if_transDel,
                        trace_bits(ospf_pck_options,
                                ospf_intf->ospf_if_options)));
        } OSPF_IF_ADDR_AREA_END(area, ospf_intf);

	GII_WRITE((gii_ctl, GW_INFO, ""));
	GII_WRITE((gii_ctl, GW_INFO, "LS Database:"));

	/* Initialise the walk.
	 */
	gii_ctl->g_ospfwalk = ospf_pt_walkinit(area);

	/* Create the job...
	 */
	gii_ctl->g_job = task_job_create(gii_ctl->g_task, TASK_JOB_PRIO_WORST,
		"GII_OSPF_TREE", gii_job_ospftree, (void_t)gii_ctl);
	gii_ctl->g_state = GIIS_JOB;
	return(0);
#endif
}

#define OSPFTREE_BULK	10

/* The actual job of displaying the OSPF DB.
 */
void
gii_job_ospftree __PF1(tjp, task_job *)
{
#ifndef PROTO_OSPF2
        return;
#else
        gii_ctl_t *gii_ctl;
        ospf_vtxLink_t *vl;
	vertex_t *vtx;
	int ind;

	gii_ctl = (gii_ctl_t *)tjp->task_job_data;

	/* We gonna print a bulk of entries from the tree.
	 */
	for(ind = 0; ind < OSPFTREE_BULK; ind++) {

		/* Are we at the end? If so, delete the job.
		 */
		if (!(gii_ctl->g_ospfwalk->tw_node)) {
			gii_ctl->g_ospfwalk =
				ospf_pt_walkdel(gii_ctl->g_ospfwalk);
			task_job_delete(gii_ctl->g_job);
			gii_ctl->g_job = NULL;
			gii_ctl->g_state = GIIS_SESSION;
			gii_prompt(gii_ctl);
			return;
		}
		vtx = TREENODE_VTX(gii_ctl->g_ospfwalk->tw_node);

		/* Print the vertex, the neighbours and the route.
		 */
		if (gii_write(gii_ctl, GW_INFO,
	"%-37s age %5d seq %08x cksum %04x len %-2d opt %02X <%s>",
				ls_vtx_name(vtx),
				VTX_AGE(vtx),
				vtx->vtx_seq, vtx->vtx_cksum,
				vtx->vtx_len,
				vtx->vtx_options,
				trace_bits(ospf_vtx_flags, vtx->vtx_flags))) {
			gii_terminate(gii_ctl->g_task);
			return;
		}
		if (vtx->vtx_rt && gii_write(gii_ctl, GW_INFO,
			"\t%sRoute %15A/%-2d Cost %-3d NextHop %-15A <%s>",
			(vtx->vtx_rt == ospf_nullrt)? "*** Ignore *** ": "",
			vtx->vtx_rt->rt_dest, mask_bits_32(vtx->vtx_mask),
			vtx->vtx_cost,
			vtx->vtx_nbNextHops?
			  (vtx->vtx_nextHops[0]?
			    vtx->vtx_nextHops[0]:
			sockbuild_str(vtx->vtx_outIntf[0]->ifa_link->ifl_name)):
			  sockbuild_str("---"),
			trace_bits(rt_state_bits, vtx->vtx_rt->rt_state))) {
			gii_terminate(gii_ctl->g_task);
                        return;
                }

		switch(vtx->vtx_type) {
		case LST_RTR:
		case LST_NTW:
			for(vl = vtx->vtx_links; vl; vl = vl->vl_next)
				if (gii_write(gii_ctl, GW_INFO,
				"\t-> %-5s %-18s cost %-3d data %-15A <%s>",
						trace_state(ospf_lslinks_types,
							vl->vl_type),
						ls_vtx_name(vl->vl_vtx),
						vl->vl_cost,
						sockbuild_in(0, vl->vl_data),
						trace_bits(ospf_vtx_flags,
							vl->vl_flags))) {
					gii_terminate(gii_ctl->g_task);
                                       	return;
                        	}
			break;

		case LST_ASIP:
		case LST_ASBR:
			if (gii_write(gii_ctl, GW_INFO, "\tTOS %d Metric %d",
					SLAMETRIC_TOS(vtx->vtx_tosmet),
					SLAMETRIC_METRIC(vtx->vtx_tosmet))) {
				gii_terminate(gii_ctl->g_task);
				return;
			}
			break;

		case LST_ASEX:
			if (gii_write(gii_ctl, GW_INFO,
				"\tTOS %d Type %d Metric %d Tag %X Forward %A",
					EXTMETRIC_TOS(vtx->vtx_tosmet),
					EXTMETRIC_TYPE(vtx->vtx_tosmet),
					EXTMETRIC_METRIC(vtx->vtx_tosmet),
					vtx->vtx_rt?
						vtx->vtx_tag:
						0xffffffff,
					sockbuild_in(0, vtx->vtx_forward))) {
				gii_terminate(gii_ctl->g_task);
				return;
			}
			break;
		}

		/* Get the next one
		 */
		ospf_pt_walk(gii_ctl->g_ospfwalk);
	}
#endif
}









