/*
 * DRET-IPv6
 *
 * Contacts:	
 *              INRIA      <Christophe.Diot@sophia.inria.fr>
 *              MASI       <Eric.Horlait@masi.ibp.fr>
 *              This software has been developped with the financial support
 *              of DRET (French Military Research Agency).
 *
 * Version:    $Id: inet6.c,v 1.13 1996/12/18 10:57:49 eh Exp $ 
 *	
 * Authors:	
 *	        Frank Zago
 *
 * Description:
 *
 *              inet6.c is derived from inet.c
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *              This notice must be copied with the distributed package.
 *
 */				

#define	INCLUDE_IF
#include "include.h"
#include "inet6.h"

int inet6_ipforwarding = FALSE;	/* IPv6 forwarding disabled by default */

task *inet6_task;
int sock_inet6;			/* AF_INET6 socket for several usages */

/* FIXME : both in6addr_any and in6addr_loopback must be in a library */
#ifndef in6addr_any
struct in6_addr in6addr_any=IN6ADDR_ANY_INIT;
#endif
#ifndef in6addr_loopback
struct in6_addr in6addr_loopback=IN6ADDR_LOOPBACK_INIT;
#endif

#define IN6ADDR_DEFAULT IN6ADDR_ANY_INIT;

static const bits inet6_if_bits[] = {
    { 0 }
};

sockaddr_un *inet6_addr_any = 0;
sockaddr_un *inet6_addr_default = 0;
sockaddr_un *inet6_addr_loopback = 0;
sockaddr_un *inet6_addr_reject = 0;		/* Where reject routes need to point */
sockaddr_un *inet6_addr_blackhole = 0;		/* Where unreachable routes need to point */

struct martians_in6 {
    const char *dest;
    flag_t adv_flags;
    flag_t dm_flags;
};

static struct martians_in6 inet6_martians[] = {
  /* Reserved adresses (RFC 1884) */
  { "0000::/8", ADVF_NO, 0},	/* Reserved + IPv4 compatible adresses */
  { "0200::/7", ADVF_NO, 0},	/* NSAP       */
  { "0400::/7", ADVF_NO, 0},	/* IPX        */
  { "0600::/7", ADVF_NO, 0},	/* unassigned */
  { "0800::/5", ADVF_NO, 0},	/* unassigned */
  { "1000::/4", ADVF_NO, 0},	/* unassigned */
  { "2000::/3", ADVF_NO, 0},	/* unassigned */
  { "6000::/3", ADVF_NO, 0},	/* unassigned */
  { "a000::/3", ADVF_NO, 0},	/* unassigned */
  { "c000::/3", ADVF_NO, 0},	/* unassigned */
  { "e000::/4", ADVF_NO, 0},	/* unassigned */
  { "f000::/5", ADVF_NO, 0},	/* unassigned */
  { "f800::/6", ADVF_NO, 0},	/* unassigned */
  { "fc00::/7", ADVF_NO, 0},	/* unassigned */
  { "fe00::/9", ADVF_NO, 0},	/* unassigned */
  { "fe80::/10", ADVF_NO, 0},	/* Link local */
  { "fec0::/10", ADVF_NO, 0},	/* Site local */
  { "ff00::/8", ADVF_NO, 0},	/* Multicast  */

  { NULL, 0, 0},
};

/*
 *	Reinit
 */
static void
inet6_reinit __PF1(tp, task *) 
{
	SI_MARTIANS(AF_INET6) = adv_destmask_finish(SI_MARTIANS(AF_INET6));
}

static void
inet6_dump __PF2(tp, task *,
		 fp, FILE *)
{
	(void) fprintf(fp, "\tIP forwarding: %d\n\n",
		       inet6_ipforwarding);
/* FIXME : Dret-IPv6 does not support reject/blackhole 
	if (!IS_ANYADDR6(sock2in6(inet6_addr_reject)) || 
	    !IS_ANYADDR6(sock2in6(inet6_addr_blackhole))) {
		if (!IS_ANYADDR6(sock2in6(inet6_addr_reject))) {
			(void) fprintf(fp, "\tReject address: %A",
				       inet6_addr_reject);
		}
		if (!IS_ANYADDR6(sock2in6(inet6_addr_blackhole))) {
			(void) fprintf(fp, "\tBlackhole address: %A",
				       inet6_addr_blackhole);
		}
		(void) fprintf(fp, "\n\n");
	}
*/	
	/* Martians */
	(void) fprintf(fp, "\tMartians:\n");
	control_dmlist_dump(fp,
			    2,
			    SI_MARTIANS(AF_INET6),
			    (adv_entry *) 0,
			    (adv_entry *) 0);
	(void) fprintf(fp, "\n");
}

/*
 * Compare the "length" first bits of two IPv6 adresses s1 and s2
 */
int 
sockaddrcmp2_in6 __PF3(s1, sockaddr_un *,
		       s2, sockaddr_un *,
		       length, int)
{
	int pdw, pbi;
	
	if ((sock2plen6(s1)<length) || (sock2plen6(s2)<length))
		return(FALSE);
	
	pdw = length >> 0x05;  /* number of full 32 bits words */
	pbi = length & 0x1f;   /* number of bits in incomplete byte */
	
	/* Compare full bytes in prefix */
	if (pdw) {
		if (memcmp(&sock2ip6(s1), &sock2ip6(s2), pdw << 2))
			return(FALSE);
	}
	
	/* Compare bits in incomplete byte */
	if (pbi) {
		u_int32 w1, w2;
		u_int32 mask;
		
		w1 = sock2in6(s1).s6_addr32[pdw];
		w2 = sock2in6(s2).s6_addr32[pdw];
		
		mask = htonl((0xffffffff) << (0x20 - pbi));
		
		if ((w1 ^ w2) & mask)
			return(FALSE);
	}
	
	return(TRUE);		/* Adresses matches */
}

/*
 * Compare two IPv6 adresses (all 128 bits)
 */
int 
sockaddrcmp_in6 __PF2(s1, sockaddr_un *,
		      s2, sockaddr_un *)
{
	return(SAME_ADDR6(sock2in6(s1), sock2in6(s2)));
}

/*
 *	Cleanup
 */
static void
inet6_cleanup __PF1(tp, task *)
{
    if (SI_MARTIANS(AF_INET6)) {
	adv_free_list(SI_MARTIANS(AF_INET6));

	SI_MARTIANS(AF_INET6) = (adv_entry *) 0;
    }

    trace_freeup(tp->task_trace);
}


void
inet6_var_init __PF0(void) { }

static void
inet6_terminate __PF1(tp, task *)    
{
    inet6_cleanup(tp);
    
    task_delete(tp);
}


void
inet6_init __PF0(void)
{
	sockaddr_un *addr;
	const struct martians_in6 *ptr;

	
	if (inet6_task) {
		inet6_task->task_trace = trace_set_global((bits *) 0, (flag_t) 0);
	} else {
		inet6_task = task_alloc("INET6",
					TASKPRI_FAMILY,
					trace_set_global((bits *) 0, (flag_t) 0));
		task_set_cleanup(inet6_task, inet6_cleanup);
		task_set_reinit(inet6_task, inet6_reinit);
		task_set_terminate(inet6_task, inet6_terminate);
		
		inet6_task->task_rtproto = RTPROTO_INET6;
		inet6_task->task_socket = task_get_socket(inet6_task, PF_INET6, SOCK_DGRAM, 0);
		
		task_set_dump(inet6_task, inet6_dump);
		if (!task_create(inet6_task)) {
			task_quit(EINVAL);
		}
	}
	
	/* Verify the validity of some kernel variables */
	if (!BIT_TEST(task_state, TASKS_TEST)) {
		if (!inet6_ipforwarding) {
			trace_log_tp(inet6_task,
				     0,
				     LOG_WARNING,
				     ("inet6_init: *WARNING* IPv6 forwarding disabled!"));
		}
		
	}
	
	/* Build list of martian networks */
	for (ptr = inet6_martians; ptr->dest; ptr++) {
		if (sockstr_in6(ptr->dest, &addr) == -1) {
			trace_log_tp(inet6_task,
				     0,
				     LOG_WARNING,
				     ("inet6_init: Invalid martian entry %s",
				      ptr->dest));
			continue;
		}
		martian_add(addr,
			    addr,
			    ptr->adv_flags,
			    ptr->dm_flags);
	}
}


/*
 *	Init all kinds of inet6 structures
 */
void
inet6_family_init __PF0(void)
{
	sock_init_family(AF_INET6,
			 (SOCKADDR_IN6_LEN - sizeof sock2in6(inet6_addr_default)),
			 SOCKADDR_IN6_LEN,
			 NULL,
			 0,
			 "sockaddr_un.in6");
	
	/* Build useful addresses */
	/* in6addr_any and in6addr_loopback are defined in BSD API draft v5 */
	inet6_addr_any = sockdup(sockbuild_in6(0, &in6addr_any, 0)); 
	inet6_addr_loopback = sockdup(sockbuild_in6(0, &in6addr_loopback, 0));
	
	rt_table_init_family(AF_INET6);
	rt_static_init_family(AF_INET6);
	
	/* Set up interface bits to be printed */
	int_ps_bits[RTPROTO_INET6] = inet6_if_bits;
}


/* 
 * Return the type of an IPv6 adress, given its prefix and prefix length.
 */
int 
inet6_get_addr_type __PF2(addr, struct in6_addr *, 
			  preflen, unsigned int)
{
	u_int32 st = ntohl(addr->s6_addr32[0]);
	
	if ((preflen >= 96) && 
	    (!(st || addr->s6_addr32[1] || addr->s6_addr32[2]))) {
		switch (ntohl(addr->s6_addr32[3]))
		{
		case 0: return INET6_ADDR_ANY;
		case 1: return INET6_ADDR_LOOPBACK;
		default: return INET6_ADDR_IP4COMP;
		}
	}

	if (preflen<3)
		return INET6_ADDR_UNASSIGNED;      

	switch (st & 0xE0000000) /* 3 first bits */
	{
	case 0x40000000: return INET6_ADDR_UNICAST; /* provider unicast */
	case 0x80000000: return INET6_ADDR_UNICAST; /* geographic unicast */

	case 0xe0000000: {
		if (preflen<8)
			return INET6_ADDR_UNASSIGNED;      

		switch(st & 0xff000000)
		{
		case 0xff000000: return INET6_ADDR_MCAST;
		case 0xfe000000: {
			if (preflen<10)
				return INET6_ADDR_UNASSIGNED;      
			
			switch (st & 0xffc00000)
			{
			case 0xfe800000: return INET6_ADDR_LINKLOCAL;
			case 0xfec00000: return INET6_ADDR_SITE_LOCAL;
			default: return INET6_ADDR_UNASSIGNED;  
			}
		}
		default: return INET6_ADDR_UNASSIGNED;
		}

	}
	default: return INET6_ADDR_UNASSIGNED;  
	}
}

