/*
 * Copyright (c) 1996 The Regents of the University of Michigan
 * All Rights Reserved
 * 
 * License to use, copy, modify, and distribute this software and its
 * documentation can be obtained from Merit at the University of Michigan.
 * 
 * Merit GateDaemon Project
 * 4251 Plymouth Road, Suite C
 * Ann Arbor, MI 48105
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE REGENTS OF THE
 * UNIVERSITY OF MICHIGAN AND MERIT DO NOT WARRANT THAT THE
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the
 * University of Michigan and Merit shall not be liable for
 * any special, indirect, incidental or consequential damages with respect
 * to any claim by Licensee or any third party arising from use of the
 * software. GateDaemon was originated and developed through release 3.0
 * by Cornell University and its collaborators.
 * 
 * Please forward bug fixes, enhancements and questions to the
 * gated mailing list: gated-people@gated.merit.edu.
 * 
 * This copyright has ben automaticly added by the util/addcopyright.pl program.
 * __END_OF_COPYRIGHT__
 */

#define	INCLUDE_TIME
#include "include.h"
#include "isis_includes.h"

#define MAXTIMER	512

struct timerEntry {
	int		ttf;		/* time to fire */
	int 	(*func)();	/* func to call */
	caddr_t	data;		/* data to pass */
} timers[MAXTIMER];

void
initTimers()
{
	int i;
	for (i=0; i<MAXTIMER; i++)
		timers[i].ttf = 0;
}

void
feepTimer()
{
	int i;

	for (i=0; i<MAXTIMER; i++) {
		if (timers[i].ttf > 0) {
			if (--timers[i].ttf == 0)
				(*(timers[i].func))(timers[i].data);
		}
	}
}

void
clearTimer(i)
int     i;
{
        timers[i].ttf = 0;
}

void
resetTimer(i, interval)
int     i;
int     interval;
{
        timers[i].ttf = interval;
}

/* 
 * Set func to be called in interval seconds; return index of timer.
 * STEVE XXX - this doesn't comply with 10589 Section 10.1, which 
 * specifies a range (with jitter) of [0.75*interval, interval].
 */
int
setTimer(interval, func, data)
int	interval;
int (*func)();
caddr_t data;
{
	int i;
	for (i=0; i<MAXTIMER; i++) {
		if (timers[i].ttf <= 0) {
			if (interval > 1) {
				if (random() & 0x01)
					interval++;
				else
					interval--;
			}
			timers[i].ttf = interval;
			timers[i].func = func;
			timers[i].data = (caddr_t) data;
			return(i);
		}
	}
	trace_log_tf(isis_trace_options,
		     0,
		     LOG_ERR,
		     ("out of timers"));
	task_quit(0);
        return(-1);
}

