/*
 * Copyright (c) 1996 The Regents of the University of Michigan
 * All Rights Reserved
 * 
 * License to use, copy, modify, and distribute this software and its
 * documentation can be obtained from Merit at the University of Michigan.
 * 
 * Merit GateDaemon Project
 * 4251 Plymouth Road, Suite C
 * Ann Arbor, MI 48105
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE REGENTS OF THE
 * UNIVERSITY OF MICHIGAN AND MERIT DO NOT WARRANT THAT THE
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the
 * University of Michigan and Merit shall not be liable for
 * any special, indirect, incidental or consequential damages with respect
 * to any claim by Licensee or any third party arising from use of the
 * software. GateDaemon was originated and developed through release 3.0
 * by Cornell University and its collaborators.
 * 
 * Please forward bug fixes, enhancements and questions to the
 * gated mailing list: gated-people@gated.merit.edu.
 * 
 * This copyright has ben automaticly added by the util/addcopyright.pl program.
 * __END_OF_COPYRIGHT__
 */

#define	INCLUDE_ROUTE
#define	INCLUDE_KVM
#include "include.h"
#ifdef	PROTO_INET
#include "inet.h"
#endif	/* PROTO_INET */
#ifdef	PROTO_ISO
#include "iso.h"
#endif	/* PROTO_ISO */
#include "krt.h"
#include "krt_var.h"



 /*  Read the kernel's routing table.			*/
int
krt_rtread __PF1(tp, task *)
{
    int i;
    size_t rtbufsize;
    rt_parms rtparms;
    int num_routes = 0;
    if_addr *ifap;
    int error = 0;

    struct rtlist  rtlist;
    struct rtreq   *rtptr, *base;

    bzero((caddr_t) &rtparms, sizeof (rtparms));
    rtparms.rtp_n_gw = 1;

    trace_only_tp(tp,
		  TRC_NL_BEFORE,
		  ("krt_rtread: Initial routes read from kernel (via ioctl):"));

    if ( ioctl (krt_task->task_socket, SIOCGRTSIZE, &num_routes) < 0) {
        error = errno;
	trace_only_tp(tp,
		      0,
		      ("krt_rtread: ioctl fails on SIOCGRTSIZE: %m"));
         return (error);
    }

    /* set up to read the kernel routing table */

    rtbufsize = (num_routes + 8) * sizeof(struct rtreq);
    rtptr = (struct rtreq *) task_block_malloc(rtbufsize);
    base = rtptr;

    rtlist.rtl_rtreq = rtptr;
    rtlist.rtl_len = rtbufsize;
    rtlist.rtl_cnt = 0;

    if ( ioctl (krt_task->task_socket, SIOCGRTTABLE, (void*) &rtlist) < 0) {
        error = errno;
	trace_only_tp(tp,
		      0,
		      ("krt_rtread: ioctl fails on SIOCGRTTABLE: %m"));
         return (error);
    }

    /* map the kernel routing data to gated structure */
    for (i=0; i < rtlist.rtl_cnt; i++, rtptr++)  {
        rtparms.rtp_dest =  sockbuild_in (0, rtptr->rtr_destaddr);
        rtparms.rtp_router = sockbuild_in (0, rtptr->rtr_gwayaddr) ;

        rtparms.rtp_state = krt_flags_to_state((flag_t) rtptr->rtr_flags);

	/* Determine netmask */
	rtparms.rtp_dest_mask = inet_mask_locate(rtptr->rtr_subnetmask);

	krt_rtread_add(tp,
		       &rtparms,
		       (flag_t) rtptr->rtr_flags,
		       FALSE,
		       "REMNANT");
    }
    task_block_reclaim(rtbufsize, (caddr_t) base);

    return 0;
}
