/*
 * Copyright (c) 1996 The Regents of the University of Michigan
 * All Rights Reserved
 * 
 * License to use, copy, modify, and distribute this software and its
 * documentation can be obtained from Merit at the University of Michigan.
 * 
 * Merit GateDaemon Project
 * 4251 Plymouth Road, Suite C
 * Ann Arbor, MI 48105
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE REGENTS OF THE
 * UNIVERSITY OF MICHIGAN AND MERIT DO NOT WARRANT THAT THE
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the
 * University of Michigan and Merit shall not be liable for
 * any special, indirect, incidental or consequential damages with respect
 * to any claim by Licensee or any third party arising from use of the
 * software. GateDaemon was originated and developed through release 3.0
 * by Cornell University and its collaborators.
 * 
 * Please forward bug fixes, enhancements and questions to the
 * gated mailing list: gated-people@gated.merit.edu.
 * 
 * This copyright has ben automaticly added by the util/addcopyright.pl program.
 * __END_OF_COPYRIGHT__
 */

#define	INCLUDE_NLIST
#define	INCLUDE_ROUTE
#define	INCLUDE_KVM
#include "include.h"
#include "inet6.h"
#include "krt.h"
#include "krt_var.h"
#include <stdio.h>

/*  Read the kernel's routing table for IPv6 */

/* Dret-IPv6 /proc/net6/route6 structure example
Iface   Destination                             Gateway
Flags   RefCnt  Use     Metric  Pref    MTU     Window  IRTT

No default route registered.

lo      00000000:00000000:00000000:01000000     00000000:00000000:00000000:00000000     0005    0       1       0       128     3584    0       0

eth0    00B5065F:003DE384:00003D00:00000000     00000000:00000000:00000000:00000000     0001    1       3       0       80      1500    0       0

eth0    000080FE:00000000:00000000:6B57F3C0     00000000:00000000:00000000:00000000     0004    1       1       1       128     1500    0       0

*/

int
krt_rtread_in6 __PF1(tp, task *)
{
	FILE *DFile;
	char if_name[IFL_NAMELEN + 1];
	char procline[500];	/* to read a line in /proc/net6/route6 */
	if_addr *ifap;
	rt_parms rtparms;
	struct in6_addr addr_dest, addr_gw;
	unsigned int refcnt, use, metric, preflen, mtu, windows, irtt;
	flag_t flags;
	int i;
	
	bzero((caddr_t) & rtparms, sizeof(rtparms));
	rtparms.rtp_n_gw = 1;
	
	DFile = fopen("/proc/net6/route6", "r");
	if (DFile == NULL)
		return EBADF;
	
	trace_only_tp(tp,
		      TRC_NL_BEFORE,
		      ("krt_rtread: Initial routes for IPv6 read from kernel (via /proc/net6/route6):"));
	
	/* read route data from /proc/net6/rout6 */
	
	/* skip over header line */
	fgets(procline, sizeof(procline), DFile); /* Ignore first line */

	while (fgets(procline, sizeof(procline), DFile)) {
		i = sscanf(procline, "%s\t%X:%X:%X:%X\t%X:%X:%X:%X\t%04x\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n",
			   if_name,
			   &addr_dest.s6_addr32[0],
			   &addr_dest.s6_addr32[1],
			   &addr_dest.s6_addr32[2],
			   &addr_dest.s6_addr32[3],
			   &addr_gw.s6_addr32[0],
			   &addr_gw.s6_addr32[1],
			   &addr_gw.s6_addr32[2],
			   &addr_gw.s6_addr32[3],
			   &flags,
			   &refcnt,
			   &use,
			   &metric,
			   &preflen,
			   &mtu,
			   &windows,
			   &irtt);
		
		if (i<17)
			continue; /* invalid line, ignore it */

		rtparms.rtp_dest = rtparms.rtp_dest_mask = sockbuild_in6(0, &addr_dest, preflen);
		rtparms.rtp_state = krt_flags_to_state(flags); 
		
		if (!IS_ANYADDR6(addr_gw)) {
			/* Gateway was specified, use it */
			
			rtparms.rtp_router = sockbuild_in6(0, &addr_gw, 128);
			
			/* Search for the intf (necessary for tunnels) */
			IF_ADDR(ifap) {
			  if (!strcmp(ifap->ifa_info.ifi_link->ifl_name, if_name))
			    rtparms.rtp_ifap = ifap;
			} IF_ADDR_END(ifap);
		} else {
			/* Gateway was not specified - try to figure it out */
			
			if (BIT_TEST(flags, RTF_GATEWAY | RTF_DCACHE)) {
				/* Not an interface route or
				 * a route learned from others
				 * - ignore it */
				continue;
			}
			
			ifap = if_withsubnet(rtparms.rtp_dest);

			if (!ifap
			    || strcmp(ifap->ifa_link->ifl_name, if_name)) {
				/* We could not find the interface */
				continue;
			}

			rtparms.rtp_router = sockbuild_in6(0, &sock2in6(ifap->ifa_addr), 128);
		}
		trace_only_tp(tp,
			      TRC_NL_BEFORE,
			      ("krt_rtread_in6: %s [%A/%d] router %A", if_name, rtparms.rtp_dest, sock2plen6(rtparms.rtp_dest), rtparms.rtp_router));
		krt_rtread_add(tp,
			       &rtparms,
			       flags,
			       FALSE,
			       "REMNANT");
	}
	
	fclose(DFile);
	
	return 0;
}
