/*
 * Copyright (c) 1996 The Regents of the University of Michigan
 * All Rights Reserved
 * 
 * License to use, copy, modify, and distribute this software and its
 * documentation can be obtained from Merit at the University of Michigan.
 * 
 * Merit GateDaemon Project
 * 4251 Plymouth Road, Suite C
 * Ann Arbor, MI 48105
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE REGENTS OF THE
 * UNIVERSITY OF MICHIGAN AND MERIT DO NOT WARRANT THAT THE
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the
 * University of Michigan and Merit shall not be liable for
 * any special, indirect, incidental or consequential damages with respect
 * to any claim by Licensee or any third party arising from use of the
 * software. GateDaemon was originated and developed through release 3.0
 * by Cornell University and its collaborators.
 * 
 * Please forward bug fixes, enhancements and questions to the
 * gated mailing list: gated-people@gated.merit.edu.
 * 
 * This copyright has ben automaticly added by the util/addcopyright.pl program.
 * __END_OF_COPYRIGHT__
 */

/* OSPF by lpj
 */
#include "include.h"
#include "inet.h"
#include "new_ospf.h"
#include "new_ospf_ls.h"

#define _OSPF_DEBUG_TEST
#define OSPF_DEBUG_TEST

/* All varialbles we need
 */
task		*ospf_task = NULL;
flag_t		ospf_flags = 0;
trace 		*ospf_trace_options = NULL;
block_t		ospf_if_blocks = NULL;
block_t		ospf_ngb_blocks = NULL;
block_t		ospf_vtxlink_blocks = NULL;
block_t		ospf_ptw_blocks = NULL;
block_t		ospf_treenodes_blocks = NULL;
block_t		ospf_vtx_blocks = NULL;
block_t		ospf_vtxlist_block = NULL;
block_t		ospf_cdtList_block = NULL;
block_t		ospf_area_blocks = NULL;
ospf_pck_t 	ospf_pck_buffer[OSPF_BUFFER_NBPCK];
int 		ospf_pck_bffct = 0;
adv_entry	*ospf_current_intf = NULL;
sockaddr_un 	*all_ospf_routers = NULL;
sockaddr_un 	*all_ospf_dr_routers = NULL;
sockaddr_un 	*all_ospf_eligible = NULL;
rtid_t		ospf_rtid = 0;
ospf_area_t	*ospf_default_area = NULL;
ospf_area_t	*ospf_current_area = NULL;
ospf_area_t	*ospf_area_list = NULL;
task_timer	*ospf_spf_timer = NULL;
time_t		ospf_spfInt = OSPF_SPF_INT;
metric_t	ospf_lsinfinity = OSPF_LSINFINITY;
cost_t		ospf_if_cost = OSPF_IF_COST;
pref_t		ospf_preference = RTPREF_OSPF;
sockaddr_un	*sock_rtmask = NULL;
gw_entry	*ospf_defaultGwList = NULL;
int		ospf_nbNgbNotReady = 0;
ospf_intf_t	*ospf_cr_vintf = NULL;
adv_entry	*ospf_export_list = NULL;
searchdata_t	__searchdata__;
ospf_ngb_t       *ospf_flush_list = NULL;
task_timer	*ospf_refresh_timer = NULL;
time_t		ospf_lsRefreshTime = OSPF_LSREFRESH_TIME;
int		ospf_nbAreas = 0;
flag_t		ospf_rtrbits = 0;
flag_t		ospf_oldrtrbits = LSRTR_BITS_UNSET;
flag_t		ospf_global_changes = 0;
iflist_t        *conf_iflist = NULL;
if_addr		*ospf_multicast_ifap = NULL;
vtxList_t	*ospf_floodQueue1 = NULL;
vtxList_t	*ospf_floodQueue2 = NULL;
task_timer	*ospf_floodTimer1 = NULL;
task_timer	*ospf_floodTimer2 = NULL;

const bits ospf_trace_types[] = {
        { TR_OSPF_STATE,	"State Transitions" },
        { TR_OSPF_DRELECT,	"DR Election" },
	{ TR_OSPF_DB,		"Link State Database" },
	{ TR_OSPF_SPF,		"Shortest Path First Computations" },
        { TR_OSPF_PCK_ALL,	"All packets" },
	{ TR_OSPF_PCK_HELLO,	"Hello packets" },
	{ TR_OSPF_PCK_DD,	"Database Desciption packets" },
	{ TR_OSPF_PCK_LSR,	"Link State Request packets" },
	{ TR_OSPF_PCK_LSU,	"Link State update packets" },
	{ TR_OSPF_PCK_LSA,	"Link State Acknowledgement packets" },
        { 0, NULL }
};

const bits ospf_pck_options[] = {
	{ OSPF_PCK_OPT_T,	"TOS" },
	{ OSPF_PCK_OPT_E,	"Ext" },
	{ 0, NULL }
};

const bits ospf_pck_flags[] = {
	{ OSPF_DD_FLAGS_MS,	"MS" },
	{ OSPF_DD_FLAGS_M,	"M" },
	{ OSPF_DD_FLAGS_I,	"I" },
	{ 0, NULL }
};

const bits ospf_rtr_bits[] = {
	{ LSRTR_BITS_B,		"AreaBorder" },
	{ LSRTR_BITS_E,		"ASBorder" },
	{ LSRTR_BITS_V,		"VLink" },
	{ 0, NULL }
};

/* All States and Events names
 */
const bits ospf_if_states[] = {
	{ OSPF_IFS_UNDEF,	"Undefined" },
	{ OSPF_IFS_NOOSPF,	"No OSPF" },
	{ OSPF_IFS_DOWN,	"Down" },
	{ OSPF_IFS_LOOPBACK,	"LoopBack" },
	{ OSPF_IFS_WAITING,	"Waiting" },
	{ OSPF_IFS_P2P,		"Pt2Pt" },
	{ OSPF_IFS_DROTHER,	"Other" },
	{ OSPF_IFS_BACKUPDR,	"Back DR" },
	{ OSPF_IFS_DR,		"DR" },
	{ 0, 0 }
};

const bits ospf_if_events[] = {
	{ EV_IF_UP,		"Interface UP" },
	{ EV_IF_WAITTIMER,	"Wait Timer" },
	{ EV_IF_BCKUPSEEN,	"Backup Seen" },
	{ EV_IF_NGBCHG,		"Neighbour Change" },
	{ EV_IF_LOOPON,		"Loop ON" },
	{ EV_IF_LOOPOFF,	"Loop OFF" },
	{ EV_IF_DOWN,		"Interface Down" },
	{ 0, 0 }
};

const bits ospf_ngb_states[] = {
	{ OSPF_NGBS_DOWN,	"Down" },
	{ OSPF_NGBS_ATTEMPT,	"Attempt" },
	{ OSPF_NGBS_INIT,	"Init" },
	{ OSPF_NGBS_2WAYS,	"2 Ways" },
	{ OSPF_NGBS_EXSTART,	"Exch Start" },
	{ OSPF_NGBS_EXCHANGE,	"Exchange" },
	{ OSPF_NGBS_LOADING,	"Loading" },
	{ OSPF_NGBS_FULL,	"Full" },
	{ 0, 0 }
};

const bits ospf_ngb_events[] = {
	{ EV_NBG_HELLORECV,	"Hello Recv" },
	{ EV_NBG_START,		"Start" },
	{ EV_NBG_2WAYS,		"2 Ways" },
	{ EV_NBG_NEGDONE,	"Negotiation Done" },
	{ EV_NBG_EXCHDONE,	"Exchange Done" },
	{ EV_NBG_BADLSREQ,	"Bad LS Sequence" },
	{ EV_NBG_LOADDONE,	"Loading Done" },
	{ EV_NBG_ADJOK,		"Adjacency OK?" },
	{ EV_NBG_SEQNUMIS,	"Sequence Mismatch" },
	{ EV_NBG_1WAY,		"1 Way" },
	{ EV_NBG_KILLNBR,	"Kill Neighbour" },
	{ EV_NBG_INACTIMER,	"Inactivity Timer" },
	{ EV_NBG_LLDOWN,	"LL Down" },
	{ 0, 0 }
};

const bits ospf_pck_types[] = {
	{ OSPF_PCK_HELLO,	"Hello" },
	{ OSPF_PCK_DD,		"Database Desciption" },
	{ OSPF_PCK_LSR,		"LS Request" },
	{ OSPF_PCK_LSU,		"LS Update" },
	{ OSPF_PCK_LSA,		"LS Acknowledgement" },
	{ 0, NULL }
};

const bits ospf_errors_msg[] = {
	{ OSPF_NOERR,		"No error" },
	{ OSPF_ERR_LENGHT,	"Packet too short" },
	{ OSPF_ERR_LENGHT2,	"Invalid header lengh" },
	{ OSPF_ERR_VERSION,	"Invalid version" },
	{ OSPF_ERR_CKSUM,	"Invalid checksum" },
	{ OSPF_ERR_AUTHTYPE,	"Invalid authentication type" },
	{ OSPF_ERR_NOMIF,	"No such multicast interface" },
	{ OSPF_ERR_NOIF,	"No such interface" },
	{ OSPF_ERR_INVIF,	"Invalid interface" },
	{ OSPF_ERR_HMASK,	"Invalid Mask" },
	{ OSPF_ERR_HINT,	"Invalid Hello interval" },
	{ OSPF_ERR_HRTDEAD,	"Invalid Router Dead interval" },
	{ OSPF_ERR_INVNGB,	"Unknown neighbour" },
	{ OSPF_ERR_INVOPTIONS,	"Options mismatch" },
	{ OSPF_ERR_DDMSMIS,	"Master flag mismatch" },
	{ OSPF_ERR_DDIMIS,	"Initialize flag mismatch" },
	{ OSPF_ERR_DDSEQMIS,	"DD sequence number mismatch" },
	{ OSPF_ERR_INVLSTYPE,	"Invalid LS type" },
	{ OSPF_ERR_BADLSR,	"Bad LS Request"},
	{ OSPF_ERR_NOVIF,	"No such virtual interface" },
	{ OSPF_ERR_BADAREA,	"Invalid area ID" },
	{ OSPF_ERR_DDCONF,	"Confusing Master/Initial flags" },
	{ 0, NULL }
};

const bits ospf_auth_types[] = {
	{ OSPF_AUTH_NONE,	"None" },
	{ OSPF_AUTH_SIMPLE,	"Simple" },
	{ 0, NULL }
};

/* Initialisation of the OSPF task. Called after parsing.
 */
void
ospf_init __PF0(void)
{
	int s;

	/* If OSPF is not used, clean up
	 */
	if (!BIT_TEST(ospf_flags, FOSPF_ON) || ospf_check_all()) {
		ospf_terminate(ospf_task);
		return;
	}

	/* Get the trace options from the global trace
         */
        trace_inherit_global(ospf_trace_options, ospf_trace_types, (flag_t) 0);

	/* Set the options to be used with the router LSA.
	 */
	ospf_rtrbits = (ospf_nbAreas > 1)? LSRTR_BITS_B: 0;
        ospf_rtrbits |= ospf_export_list? LSRTR_BITS_E: 0;
        if (ospf_oldrtrbits != LSRTR_BITS_UNSET &&
			ospf_oldrtrbits != ospf_rtrbits)
                ospf_global_changes = ospf_rtrbits ^ ospf_oldrtrbits;
	ospf_oldrtrbits = ospf_rtrbits;

	/* Should we create a task to receive/send the OSPF packets?
	 */
	if (!ospf_task) {
		int ospf_proto;

		/* Alloc the task...
		 */
		ospf_task = task_alloc("OSPF", TASKPRI_PROTO + 1,
			ospf_trace_options);

		/* Get the socket. OSPF is directly on top of IP, so use
		 * a raw IP socket...
		 */
		ospf_proto = task_get_proto(ospf_trace_options, "ospf",
				IPPROTO_OSPF);
		if ((ospf_task->task_socket = task_get_socket(ospf_task,
				AF_INET, SOCK_RAW, ospf_proto)) < 0) {
			s = errno;
			trace_log_tp(ospf_task, 0, LOG_ERR,
				("ospf_init: cannot get socket"));
			task_quit(s);
		}

		/* which must be non blocking of course...
		 */
		if (task_set_option(ospf_task, TASKOPTION_NONBLOCKING,TRUE) < 0)
			task_quit(s);

		/* A.O.B.
		 */
		task_set_terminate(ospf_task, ospf_terminate);
		task_set_ifachange(ospf_task, ospf_ifchange);
		task_set_dump(ospf_task, ospf_dump);
		task_set_cleanup(ospf_task, ospf_cleanup);
		task_set_reinit(ospf_task, ospf_reinit);
		task_set_recv(ospf_task, ospf_recv);
		task_set_flash(ospf_task, ospf_flash);
		task_set_newpolicy(ospf_task, ospf_newpolicy);
		ospf_task->task_rtbit = rtbit_alloc(ospf_task, 0,
			sizeof(vertex_t *), (void_t)ospf_task, NULL);

		/* Start the task...
		 */
		task_create(ospf_task);
	}

	ospf_reconfig();

	/* Initialise the areas.
	 */
	ospf_area_init();

	/* We do not want to recieve our own packets.
	 */
	if (task_set_option(ospf_task, TASKOPTION_MULTI_LOOP, FALSE) < 0)
		/* Well, this is bad. Better get out.
		 */
		 task_quit(errno);

	/* XX I am sure gated does that for me somewhere!
	 */
	sock_rtmask = inet_mask_locate(RTMASK);

	/* The default gateway structure, used when a route as not
	 * particulary gateway.
	 */
	ospf_defaultGateway = gw_locate(&ospf_defaultGwList, RTPROTO_OSPF,
		ospf_task, 0, 0, sockbuild_in(0, ospf_rtid), 0);

	/* The route associated with our own router LSA. Will not be deleted.
	 */
	if (!ospf_nullrt) {
		rt_parms rtp;

		rtp.rtp_dest = sockbuild_in(0, ospf_rtid);
        	rtp.rtp_dest_mask = sock_rtmask;
        	rtp.rtp_n_gw = 0;
        	rtp.rtp_gwp = ospf_defaultGateway;
        	rtp.rtp_metric = ospf_lsinfinity;
        	rtp.rtp_metric2 = ospf_lsinfinity;
        	rtp.rtp_tag = 0;
        	rtp.rtp_state = RTS_INTERIOR;
        	rtp.rtp_preference = -1;
        	rtp.rtp_preference2 = 0;
        	rtp.rtp_rtd = (void_t)NULL;
#ifdef  PROTO_ASPATHS
        	rtp.rtp_asp = NULL;
#endif
		rt_open(ospf_task);
        	ospf_nullrt = rt_add(&rtp);
		OSPF_RT_VTX(ospf_nullrt) = NULL;
		ASSERT(!BIT_TEST(ospf_nullrt->rt_state, RTS_ACTIVE));
		rt_close(ospf_task, NULL, 0, "OSPF Init");
	}

#ifdef OSPF_DEBUG_TEST
        ospf_test_db();
#endif
	ASEXTREE_BUG("ospf_init 1");
}

/* Make sure every parameters are there. And everything makes sence. Called
 * just after the config file is parsed. Return > 0 if error.
 */
int
ospf_check_all __PF0(void)
{
	/* First we need a router ID.
	 */
	if (!inet_routerid_entry) {
		trace_log_tf(ospf_trace_options, 0, LOG_WARNING,
			("OSPF: no router ID defined. OSPF disabled"));
		ospf_flags &= ~FOSPF_ON;
		return(1);
	}

	ospf_rtid = SOCK32(inet_routerid);
	return(0);
}

/* Called before the config file is read.
 */
void
ospf_var_init __PF0(void)
{
        /* Initialize all memory structures... Timers.
         */
        if (!ospf_if_blocks)
                ospf_if_blocks =
                        task_block_init(sizeof(ospf_intf_t), "ospf_intf_t");
        if (!ospf_ngb_blocks)
                ospf_ngb_blocks =
                        task_block_init(sizeof(ospf_ngb_t), "ospf_ngb_t");
        if (!ospf_treenodes_blocks)
                ospf_treenodes_blocks =
                        task_block_init(sizeof(tree_node_t),"ospf_tree_node_t");
        if (!ospf_vtxlink_blocks)
                ospf_vtxlink_blocks =
                       task_block_init(sizeof(ospf_vtxLink_t),"ospf_vtxLink_t");
        if (!ospf_ptw_blocks)
                ospf_ptw_blocks =
                        task_block_init(sizeof(ospf_ptwalk_t), "ospf_ptwalk_t");
        if (!ospf_vtx_blocks)
                ospf_vtx_blocks =
                        task_block_init(sizeof(vertex_t), "vertex_t");
        if (!ospf_vtxlist_block)
                ospf_vtxlist_block =
                        task_block_init(sizeof(vtxList_t), "vtxList_t");
        if (!ospf_cdtList_block)
                ospf_cdtList_block =
                        task_block_init(sizeof(cdtList_t), "cdtList_t");
        if (!ospf_area_blocks)
                ospf_area_blocks =
                        task_block_init(sizeof(ospf_area_t),"ospf_area_t");

	/* Reset all global variable to default.
	 */
	ospf_reset_default();

        /* A.O.B. First register the multicast addresses.
         */
        if (!all_ospf_dr_routers) {
                all_ospf_dr_routers =
                                sockdup(sockbuild_in(0, ALL_OSPF_DR_ROUTERS));
                krt_multicast_add(all_ospf_dr_routers);
        }
        if (!all_ospf_routers) {
                all_ospf_routers =
                                sockdup(sockbuild_in(0, ALL_OSPF_ROUTERS));
                krt_multicast_add(all_ospf_routers);
        }
        if (!all_ospf_eligible)
                all_ospf_eligible = sockdup(sockbuild_in(0, ALL_OSPF_ELIGIBLE));

        /* Initialize SPF scheduler timer.
         */
        if (!ospf_spf_timer)
                ospf_spf_timer = task_timer_create(ospf_task, "SPF Schedule",
                        0, 0, 0, ospf_spf_compute, (void_t) NULL);
 
        /* And the LS refresh timer
         */
        if (!ospf_refresh_timer)
                ospf_refresh_timer = task_timer_create(ospf_task, "LS Refresh",
                        0, ospf_lsRefreshTime, 0, ospf_refresh_schedule, NULL);
	else
		task_timer_set(ospf_refresh_timer, ospf_lsRefreshTime, 0);
 
        /* Those 2 timers are used for delayed transmission of self originated
         * LSA.
         */
        if (!ospf_floodTimer1)
                ospf_floodTimer1 =  task_timer_create(ospf_task, "Flood 1",
                        0, 0, 0, ospf_flood_timer, NULL);
        if (!ospf_floodTimer2)
                ospf_floodTimer2 =  task_timer_create(ospf_task, "Flood 2",
                        0, 0, 0, ospf_flood_timer, NULL);
}

/* Set all undefined interface parameters to a default value. By default all
 * interfaces run OSPF.
 */
void
ospf_init_if __PF2(area, ospf_area_t *, ifap, if_addr *)
{
	/* If this interface has already been configured, that's
	 * fine!
	 */
	if (ifap->ifa_ospf && ifap->ifa_ospf_state)
		return;

	/* The OSPF specific data structure does not exist yet,
	 * we create one now!
	 */
	(void)ospf_new_intf(SOCK32(ifap->ifa_addr), area, 0);
}

/* Create a structure for storing the OSPF interface parameters.
 * Or uses an old existing one,
 */
ospf_intf_t *
ospf_new_intf __PF3(addr, ipaddr_t, area, ospf_area_t *, flag, flag_t)
{
	ospf_intf_t *ospf_intf, **oifp;

	/* Create an initialise an OSPF specific structure. If no real
	 * interface exist (ifap) then this is a virtual link!
	 * Used an old one if it already exist.
	 */
	if (!(ospf_intf = ospf_if_find(addr))) {
		ospf_intf = (ospf_intf_t *)task_block_alloc(ospf_if_blocks);

		/* Add the interface to the area's list.
		 */
		ospf_intf->ospf_if_area         = area;
		ospf_intf->ospf_if_next = area->area_intf;
		area->area_intf = ospf_intf;

		ospf_intf->ospf_if_helloTimer   = NULL;
		ospf_intf->ospf_if_waitTimer    = NULL;
		ospf_intf->ospf_if_ngb          = NULL;
		ospf_intf->ospf_if_drID         = OSPF_NULL_DR;
		ospf_intf->ospf_if_drAddr       = OSPF_NULL_DR;
		ospf_intf->ospf_if_backupDrID   = OSPF_NULL_DR;
		ospf_intf->ospf_if_backupDrAddr = OSPF_NULL_DR;
		ospf_intf->ospf_if_ntls	        = NULL;
		ospf_intf->ospf_if_addr         = addr;
		OSPF_IF_NEWSTATE(ospf_intf, OSPF_IFS_DOWN);
		ospf_intf->ospf_if_helloInt     = OSPF_HELLO_INT;
		ospf_intf->ospf_if_rtdeadInt    = OSPF_RTDEAD_INT;
		ospf_intf->ospf_if_transDel     = OSPF_TRANS_DELAY;
		ospf_intf->ospf_if_rtpri        = OSPF_RTPRI;
		ospf_intf->ospf_if_cost         = ospf_if_cost;
		ospf_intf->ospf_if_rmxInt       = OSPF_IF_RMX;
		ospf_intf->ospf_if_authType     = OSPF_IF_AUTH;
		ospf_intf->ospf_if_authType2    = OSPF_IF_AUTH2;
		ospf_intf->ospf_if_options      = OSPF_PCK_OPT_E;
		ospf_intf->ospf_if_flags	|= flag;
	}
	else {
		ospf_intf->ospf_if_changes |= CFG_OLD|CFG_KEEP;

		if (ospf_intf->ospf_if_area != area) {
			/* If this interface belongs to a new area, change it.
		 	*/
			for(oifp = &ospf_intf->ospf_if_area->area_intf; *oifp;)
				if (*oifp == ospf_intf) {
					*oifp = ospf_intf->ospf_if_next;
					break;
				}
				else
					oifp = &(*oifp)->ospf_if_next;
	
			ospf_intf->ospf_if_area = area;
			ospf_intf->ospf_if_next = area->area_intf;
			area->area_intf = ospf_intf;
			ospf_intf->ospf_if_changes |= CFG_AREA;
		}
	}
	ospf_intf->ospf_if_changes      |= CFG_KEEP;

	return(ospf_intf);
}

/* what to do when an interface changes...
 */
void
ospf_ifchange __PF2(tp, task *, ifap, if_addr *)
{
	ASEXTREE_BUG("ospf_ifchange 1");

	/* If no ifa_ospf structure exist yet, find one. When configuring
	 * an interface from the config file, no ospf_if_ifap is
	 * set. We set it the first time the interface comes up.
	 */
	if (!ifap->ifa_ospf) {
		if (!(ifap->ifa_ospf = ospf_if_find(SOCK32(ifap->ifa_addr))))
			return;
		ASSERT(ifap->ifa_ospf->ospf_if_addr);
		ifap->ifa_ospf->ospf_if_ifap = ifap;
		ifap->ifa_ospf->ospf_if_flags |= ifap->ifa_state;
	}

	/* We don't care about anything else than INET, and if the interface
	 * does not run OSPF.
	 */
	if (socktype(ifap->ifa_addr) != AF_INET || !ifap->ifa_ospf ||
			ifap->ifa_ospf_state == OSPF_IFS_NOOSPF) {
		ASEXTREE_BUG("ospf_ifchange 2");
		return;
	}

	/* We open the routing table because an interface change may
	 * trigger a new LSA for our own router, and therefore a modification
	 * of the routing table.
	 */
	rt_open(ospf_task);

	/* If it's an deletion, get rid of the interface.
	 */
	if (BIT_TEST(ifap->ifa_change, IFC_DELETE)) {
		ospf_if_event(EV_IF_DOWN, ifap->ifa_ospf);
		ASEXTREE_BUG("ospf_ifchange 2.1");
		rt_close(ospf_task, NULL, 0, "OSPF Interface Change");
		return;
	}

	/* We care about the interface going up, going down...
	 */
	if (BIT_TEST(ifap->ifa_change, IFC_UPDOWN)) {
		if (BIT_TEST(ifap->ifa_state, IFS_UP))
			ospf_if_event(EV_IF_UP, ifap->ifa_ospf);
		else
			ospf_if_event(EV_IF_DOWN, ifap->ifa_ospf);
		ASEXTREE_BUG("ospf_ifchange 2.2");
		rt_close(ospf_task, NULL, 0, "OSPF Interface Change");
		return;
	}

	/* ... or something changed but still up
	 */
	if (BIT_TEST(ifap->ifa_change,
			IFC_NETMASK|IFC_METRIC|IFC_BROADCAST|IFC_ADDR) &&
			BIT_TEST(ifap->ifa_state, IFS_UP)) {
		/* Something changed. We add/del the interface?
		 * XXX Carefull, if the address changed, we need to change
		 * the ifa_ospf pointer.
		 */
		ospf_if_event(EV_IF_DOWN, ifap->ifa_ospf);
		ospf_if_event(EV_IF_UP, ifap->ifa_ospf);
		rt_close(ospf_task, NULL, 0, "OSPF Interface Change");
		ASEXTREE_BUG("ospf_ifchange 2.3");
		return;
	}

	/* It seems that during initialisation gated does an ifchange()
	 * with no change set (IFC_NOCHANGE)! In this case, if the
	 * interface is UP and in the DOWN ospf state then we force a
	 * EV_IF_UP event. Also if we get a IFC_ADD, do the same.
	 */
	if ((ifap->ifa_change == IFC_NOCHANGE ||
		BIT_TEST(ifap->ifa_change, IFC_ADD)) &&
			BIT_TEST(ifap->ifa_state, IFS_UP) &&
			ifap->ifa_ospf_state <= OSPF_IFS_DOWN) {
		ospf_if_event(EV_IF_UP, ifap->ifa_ospf);
		ASEXTREE_BUG("ospf_ifchange 2.4");
		rt_close(ospf_task, NULL, 0, "OSPF Interface Change");
		return;
	}

	rt_close(ospf_task, NULL, 0, "OSPF Interface Change");
	ASEXTREE_BUG("ospf_ifchange 3");
}

/* A new interface just came up. Everyting need to be set. EV_IF_UP method.
 */
void
ospf_if_add __PF1(ospf_intf, ospf_intf_t *)
{
	static time_t nextSlot = 0;

	/* Some basic stuff linked to the inderlying physical interface.
	 */
	ASSERT(ospf_intf->ospf_if_ifap && ospf_intf->ospf_if_mtu);
	ospf_intf->ospf_if_maxLSRinPacket = MTU2LSR(ospf_intf->ospf_if_mtu);
	ospf_intf->ospf_if_maxDDinPacket = MIN(NGB_MAXDDSIZE,
			MTU2DD(ospf_intf->ospf_if_mtu));
	ospf_intf->ospf_if_maxACKinPacket = MIN(NGB_MAXACKSIZE,
			MTU2ACK(ospf_intf->ospf_if_mtu));

	/* Everything is reset to the state specified on RFC1583, p37
	 * For p2p interface, go to the P2P state,
	 * For loopback go to OSPF_IFS_LOOPBACK state (and ignore it),
	 * else go to WAITING state unless we are no eligible for DR.
	 */
	if (BIT_TEST(ospf_intf->ospf_if_flags, IFS_LOOPBACK)) {
		OSPF_IF_NEWSTATE(ospf_intf, OSPF_IFS_LOOPBACK);
		return;
	}
	else if (BIT_TEST(ospf_intf->ospf_if_flags,
			IFS_POINTOPOINT|IFS_OSPFVLINK))
		OSPF_IF_NEWSTATE(ospf_intf, OSPF_IFS_P2P);
	else if (ospf_intf->ospf_if_rtpri)
		OSPF_IF_NEWSTATE(ospf_intf, OSPF_IFS_WAITING);
	else
		OSPF_IF_NEWSTATE(ospf_intf, OSPF_IFS_DROTHER);

	/* Create the hello timer. An offset is added if ospf_if_add() is
	 * called more than once during the same second. It is to prevent
	 * from setting up too many interfaces at the same time, and to
	 * desymchronise all interfaces. nextSlot is the time the last intf
	 * is supposed to start sending Hellos
	 */
	if (!ospf_intf->ospf_if_helloTimer) {
		char timerName[OSPF_MAX_TIMER_NAME];

		(void)sprintf(timerName, "%A Hello",
			sockbuild_in(0, ospf_intf->ospf_if_addr));
		ospf_intf->ospf_if_helloTimer = task_timer_create(ospf_task,
			strdup(timerName), TIMERF_HIPRIO,
			ospf_intf->ospf_if_helloInt,
			(time_sec > nextSlot)? 0: (nextSlot - time_sec),
			ospf_hello_timer, (void_t)ospf_intf);
	}
	else
		task_timer_set(ospf_intf->ospf_if_helloTimer,
			ospf_intf->ospf_if_helloInt,
			(time_sec > nextSlot)? 0: (nextSlot - time_sec));

	nextSlot = (time_sec > nextSlot)? time_sec:
		((nextSlot - time_sec + 1) % ospf_intf->ospf_if_helloInt) +
			time_sec;

	/* Create the WAIT timer. It's a one shot timer which will fire
	 * when we have to elect a DR.
	 */
	if (!ospf_intf->ospf_if_waitTimer) {
		char timerName[OSPF_MAX_TIMER_NAME];
 
                (void)sprintf(timerName, "%A Wait",
			sockbuild_in(0, ospf_intf->ospf_if_addr));
		ospf_intf->ospf_if_waitTimer = task_timer_create(ospf_task,
			strdup(timerName), TIMERF_HIPRIO, 0,
			ospf_intf->ospf_if_rtdeadInt,
			ospf_wait_timer, (void_t)ospf_intf);
	}
	else
		task_timer_set(ospf_intf->ospf_if_waitTimer, 0,
			ospf_intf->ospf_if_rtdeadInt);

	/* If the interface is multicast, register as a member of
	 * the ospf multicast groups.
	 */
	if (BIT_TEST(ospf_intf->ospf_if_flags, IFS_MULTICAST))
		OSPF_JOINT_GROUP(ospf_intf, all_ospf_routers);

	/* for a NBMA network, move the state of the configured neighbours
	 * to OSPF_NGBS_ATTEMPT. Really? We'll see...
	 */

	/* finally, originate a LSA for our own router.
	 */
	ospf_gen_rtrlsa(ospf_intf->ospf_if_area);
	/*OSPF_GEN_RTRLSA();*/
}

/* An interface is going down.
 */
void
ospf_if_down __PF1(ospf_intf, ospf_intf_t *)
{
	ospf_ngb_t *ngb, *next;

	/* The state is now down.
	 */
	OSPF_IF_NEWSTATE(ospf_intf, OSPF_IFS_DOWN);

	/* Remove all neighbours.
	 */
	ALL_NGB_TERMINATE(ospf_intf);

	/* Reset back pointer.
         */
        if (ospf_intf->ospf_if_ifap)
                ospf_intf->ospf_if_ifap->ifa_ospf = NULL;

	/* Destroy all the timers
         */
        if (ospf_intf->ospf_if_helloTimer)
                task_timer_delete(ospf_intf->ospf_if_helloTimer);
        if (ospf_intf->ospf_if_waitTimer)
                task_timer_delete(ospf_intf->ospf_if_waitTimer);

	/* Quit the multicast group for OSPF
	 */
	if (BIT_TEST(ospf_intf->ospf_if_flags, IFS_MULTICAST)) {
		OSPF_QUIT_GROUP(ospf_intf, all_ospf_routers);
		OSPF_QUIT_GROUP(ospf_intf, all_ospf_dr_routers);
	}

	/* Generate a new router LSA about ourselves.
	 */
	ospf_gen_rtrlsa(ospf_intf->ospf_if_area);
	/*OSPF_GEN_RTRLSA();*/
}

/* Just generate a EV_IF_WAITTIMER event to the interface, to tell
 * it it's time for election.
 */
void
ospf_wait_timer __PF2(tip, task_timer *, interval, time_t)
{
	ospf_intf_t *ospf_intf;

	/* We need to open the routing table since a DR election will
	 * happen, and therefore a LSA for our router be generated.
	 */
	ASEXTREE_BUG("ospf_wait_timer 1");
	rt_open(ospf_task);

	ospf_intf = (ospf_intf_t *)tip->task_timer_data;
	ospf_if_event(EV_IF_WAITTIMER, ospf_intf);

	rt_close(ospf_task, NULL, 0, "OSPF WAIT TIMER");
	ASEXTREE_BUG("ospf_wait_timer 2");
}

/* The job of sending an hello packet. Periodicaly triggered by a timer.
 * Hello packets have some known fields and the list of known neighbours.
 */
void
ospf_hello_timer __PF2(tip, task_timer *, interval, time_t)
{
	ASEXTREE_BUG("ospf_hello_timer 1");
	ospf_send_hello((ospf_intf_t *)tip->task_timer_data);
	ASEXTREE_BUG("ospf_hello_timer 2");
}

void
ospf_send_hello __PF1(ospf_intf, ospf_intf_t *)
{
	ospf_hello_pck_t *pck;
        ospf_ngb_t *ngb;
        int pcklen, maxlen, count;

	/* Get the packet, type HELLO
	 */
	OSPF_PACKET(pck, OSPF_PCK_HELLO, maxlen, ospf_intf->ospf_if_ifap);

	/* Some known fields
	 */
	pck->pck_hello_mask	= OSPF_IFMASK(ospf_intf);
	pck->pck_hello_int	= htons(ospf_intf->ospf_if_helloInt);
	pck->pck_hello_options	= ospf_intf->ospf_if_options;
	pck->pck_hello_rtpri	= ospf_intf->ospf_if_rtpri;
	pck->pck_hello_rtdead	= htonl(ospf_intf->ospf_if_rtdeadInt);
	pck->pck_hello_dr	= ospf_intf->ospf_if_drAddr;
	pck->pck_hello_backupDr = ospf_intf->ospf_if_backupDrAddr;

	/* Go through all the neighbours... and put their router ID on the
	 * packet.
	 */
	count = 0;
	pcklen = sizeof(ospf_hello_pck_t) - sizeof(u_int32);
	OSPF_INTF_NGB(ngb, ospf_intf) {
		if (ngb->ngb_state >= OSPF_NGBS_INIT) {
			pck->pck_hello_ngb[count++] = ngb->ngb_rtid;
			pcklen += sizeof(ngb->ngb_rtid);
		}
	}

	/* Now send the packet... 
	 */
	ospf_send((ospf_pck_t *)pck, pcklen, ospf_intf,
		ospf_send_addr(ospf_intf, SEND_HELLO,
			ospf_intf->ospf_if_ngb));
}

/* Send a OSPF packet. pck is supposed to be a generic OSPF packet.
 * If dst is null that's a multicast (when possible) address.
 */
void
ospf_send __PF4(pck, ospf_pck_t *, pcklen, int, ospf_intf, ospf_intf_t *, to,
	sockaddr_un *)
{
	ospf_ngb_t *ngb;
	int err;

	/* First fills in the generic field lenght, checksum, etc...
	 */
	pck->pck_rtid		= ospf_rtid;
	pck->pck_area		= ospf_intf->ospf_if_area->area_id;
	pck->pck_authtype	= htons(ospf_intf->ospf_if_authType);
	pck->pck_len 		= htons(pcklen);

	switch(ospf_intf->ospf_if_authType) {
	case OSPF_AUTH_NONE:
		/* No authentication needed
		 */
		pck->pck_checksum = inet_cksum((void_t) pck, pcklen);
		break;
	default: /* Why not! */
		ASSERT(0);
	}

	/* Well, here we put all the cases we do not want to send the
	 * packet!
	 */
	/* [XXX] */

	/* Now lets figure out where to send the packet.
	 * There is 3 kind of destination: non broadcast addr, standard
	 * broadcast addr (AllSPFRouters and AllDR), and one used only for
	 * NBMA network, AllEligibleRouter (fake multicast address).
	 */
	if (BIT_TEST(ospf_intf->ospf_if_flags, IFS_MULTICAST)) {
		/* The interface supports multicast. Set the source address to
		 * use.
		 */
		if (ospf_multicast_ifap != ospf_intf->ospf_if_ifap)
			(void)task_set_option(ospf_task, TASKOPTION_MULTI_IF,
			     (ospf_multicast_ifap = ospf_intf->ospf_if_ifap));

		/* Just send the stuff.
		 */
		err = task_send_packet(ospf_task, (void_t)pck, pcklen, 
			!IN_MULTICAST(ntohl(sock2ip(to)))? MSG_DONTROUTE:0, to);
                ospf_trace_packet(pck, pcklen, err, TR_PACKET_SEND, to,
                        sockbuild_in(0, ospf_intf->ospf_if_addr));
                return;
        }
	/* For non multicast address, we can send directly.
	 */
	if (!IN_MULTICAST(ntohl(sock2ip(to)))) {
		/* Unicast address. Just send.
		 */
                err = task_send_packet(ospf_task, (void_t)pck, pcklen,
			MSG_DONTROUTE, to);
		ospf_trace_packet(pck, pcklen, err, TR_PACKET_SEND, to,
			sockbuild_in(0, ospf_intf->ospf_if_addr));
		return;
	}
	/* We have a NBMA network. Go through all neighbours, and
	 * send the packet if needed.
	 */
        OSPF_INTF_NGB(ngb, ospf_intf) {
		/* A kludge for NBMA network with a neigbour in the ATEMPT
		 * state. The Hello packet are sent at larger interval.
		 */
		if (pck->pck_type == OSPF_PCK_HELLO &&
			ngb->ngb_state == OSPF_NGBS_ATTEMPT &&
			ospf_intf->ospf_if_state > OSPF_IFS_WAITING &&
			ngb->ngb_lastHello + ospf_intf->ospf_if_pollInt <
				time_sec) {
			ngb->ngb_lastHello = time_sec;
			err = task_send_packet(ospf_task, (void_t)pck, pcklen,
                                MSG_DONTROUTE,
				sockbuild_in(0, htonl(ngb->ngb_addr)));
                        ospf_trace_packet(pck, pcklen, err, TR_PACKET_SEND,
                                sockbuild_in(0, htonl(ngb->ngb_addr)),
				sockbuild_in(0, ospf_intf->ospf_if_addr));
			continue;
		}

		/* If sent to all_ospf_routers, then send to every body.
		 * If sent to all_ospf_dr_routers, then send to DR and backup
		 * only. If sent to all_ospf_eligible, then send it to
		 * all eligible routers only.
		 */
		if (to == all_ospf_routers ||
			(to == all_ospf_dr_routers &&
				(NGB_ISDR(ngb) || NGB_ISBACKUP(ngb))) ||
			(to == all_ospf_eligible && ngb->ngb_rtpri)) {

                	err = task_send_packet(ospf_task, (void_t)pck, pcklen,
				MSG_DONTROUTE,
				sockbuild_in(0, htonl(ngb->ngb_addr)));
			ospf_trace_packet(pck, pcklen, err, TR_PACKET_SEND,
				sockbuild_in(0, htonl(ngb->ngb_addr)),
				sockbuild_in(0, ospf_intf->ospf_if_addr));
		}
        }
}

/* Send a Database Desciption packet
 */
void
ospf_send_dd __PF1(ngb, ospf_ngb_t *)
{
	ospf_dd_pck_t *pck;
	int pcklen, maxlen;
	ospf_lsa_hdr_t *lsh;
	int ind;
	vertex_t *vtx;

	/* Reset the retransmit timer.
	 */
	task_timer_set(ngb->ngb_rmxTimer, 0, ngb->ngb_intf->ospf_if_rmxInt);

	/* Get a packet, type DD
	 */
	OSPF_PACKET(pck, OSPF_PCK_DD, maxlen, ngb->ngb_ifap);

	/* Some known fields...
	 */
	pck->pck_dd_options = ngb->ngb_intf->ospf_if_options;
	pck->pck_dd_seq = htonl(ngb->ngb_ddseq);

	switch(ngb->ngb_state) {
	case OSPF_NGBS_EXSTART:
		/* If the state is OSPF_NGBS_EXSTART the rest must be empty
	 	 * (no LSA in the packet).
	 	 */
		pck->pck_dd_flags = ngb->ngb_options;
		pcklen = sizeof(ospf_dd_pck_t);
		ospf_send((ospf_pck_t *)pck, pcklen, ngb->ngb_intf,
			sockbuild_in(0, ngb->ngb_addr));
		return;

	case OSPF_NGBS_EXCHANGE:
		/* We send the next items in the database summary list.
		 * The packet will contain ngb->ngb_lastDDsize database
		 * summary which vertex(nodes?) addresses will be put in the
		 * ngb->ngb_lastDDvtx[] array. Once the packet is acknoledged,
		 * all vertex in the array are removed from the database
		 * summary list. Ouf!
		 * If ngb->ngb_lastDDsize is not null, that means we are
		 * doing a retransmission.
		 * First we fill up ngb->ngb_lastDDvtx[] if it's empty.
		 */
		if (!ngb->ngb_lastDDsize)
			for(ngb->ngb_lastDDsize = 0;
				ngb->ngb_lastDDsize < ngb->ngb_maxDDinPacket;
				ngb->ngb_lastDDsize++) {

				NGB_LIST_NEXT_DS(ngb,
				       ngb->ngb_lastDDvtx[ngb->ngb_lastDDsize]);

				/* Are we done? If so, reset the M flag.
				 */
				if (!ngb->ngb_lastDDvtx[ngb->ngb_lastDDsize]) {
					BIT_RESET(ngb->ngb_options,
						OSPF_DD_FLAGS_M);
					break;
				}
			}

		pck->pck_dd_flags = ngb->ngb_options;
		pcklen = sizeof(ospf_dd_pck_t);

		/* Put all vertex from ngb->ngb_lastDDvtx[] into the packet.
		 */
		(ospf_dd_pck_t *)lsh = pck + 1;
		for(ind = 0; ind < ngb->ngb_lastDDsize; ind++) {
			vtx = ngb->ngb_lastDDvtx[ind];
			LSA_SET((ospf_lsa_t *)lsh, VTX_AGE(vtx),
				vtx->vtx_options, vtx->vtx_type, vtx->vtx_id,
				vtx->vtx_advrt, vtx->vtx_seq, vtx->vtx_cksum,
				vtx->vtx_len);
			lsh++;
			pcklen += sizeof(ospf_lsa_hdr_t);
		}

		/* send it now...
		 */
		ospf_send((ospf_pck_t *)pck, pcklen, ngb->ngb_intf,
                        sockbuild_in(0, ngb->ngb_addr));

	default:
		return;
	}
}

/* Send a OSPF LS request packet. From the queue.
 */
void
ospf_send_lsr __PF1(ngb, ospf_ngb_t *)
{
	ospf_lsr_pck_t *pck;
	int pcklen, maxlen;
	ospf_lsr_data_t *lsr;
        vertex_t *vtx;

	ASSERT(ngb->ngb_state == OSPF_NGBS_EXCHANGE ||
		ngb->ngb_state == OSPF_NGBS_LOADING);

	/* Reset the retransmit timer.
         */
        task_timer_set(ngb->ngb_rmxTimer, 0, ngb->ngb_intf->ospf_if_rmxInt);

	/* Get a packet, type LSR
	 */
	OSPF_PACKET(pck, OSPF_PCK_LSR, maxlen, ngb->ngb_ifap);
	pcklen = sizeof(ospf_lsr_pck_t) - sizeof(ospf_lsr_data_t);

	/* If the current buffer empty, then we fill it up.
	 * (if not, this is a retransmission).
	 */
	while(ngb->ngb_LSRbuffer.pt_used_nodes < ngb->ngb_maxLSRinPacket) {
		NGB_LIST_NEXT_LSR(ngb, vtx);
		ASSERT(!vtx || !db_find(&ngb->ngb_lsReq, vtx->vtx_type,
			&vtx->vtx_key, &__searchdata__));
		ASSERT(vtx || !ngb->ngb_lsReq.pt_used_nodes);

		/* Are we done?
		 */
		if (!vtx)
			break;

		/* Put it in the buffer...
		 */
		NGB_LIST_ADD_LSR_BUFF(ngb, vtx->vtx_type, &vtx->vtx_key, vtx);
	}

	/* Put all vertex from the buffer into the packet.
	 */
	lsr = pck->pck_lsr_data;
	PT_WALK(&ngb->ngb_LSRbuffer, vtx) {
		LSR_SET(lsr, vtx->vtx_type, vtx->vtx_id, vtx->vtx_advrt);
		lsr++;
		pcklen += sizeof(ospf_lsr_data_t);
	} PT_WALK_END(&ngb->ngb_LSRbuffer, vtx);

	/* send it now...
	 */
        ospf_send((ospf_pck_t *)pck, pcklen, ngb->ngb_intf,
                sockbuild_in(0, ngb->ngb_addr));
}

/* Resend a LSA. Triggered by the retransmit timer. Only resent what's in
 * the buffer.
 */
void
ospf_send_rmx __PF1(ngb, ospf_ngb_t *)
{
	ospf_lsu_pck_t *pck;
        vertex_t *vtx;
        int pcklen, nbAdvt, maxlen;
        ospf_lsa_t *lsa;

	/* Reset the retransmit timer.
         */
        task_timer_set(ngb->ngb_rmxTimer, 0, ngb->ngb_intf->ospf_if_rmxInt);

	/* Get a packet, type LSU.
         */
        OSPF_PACKET(pck, OSPF_PCK_LSU, maxlen, ngb->ngb_ifap);
 
        lsa = pck->pck_lsu_lsa;
        nbAdvt = 0;
	pcklen = sizeof(ospf_lsu_pck_t) - sizeof(ospf_lsa_t);

        /* Put all vertex from the buffer into the packet.
         */
	PT_WALK(&ngb->ngb_RMXbuffer, vtx) {

		/* The next LSA should fit in the packet.
                 */
		ASSERT(pcklen + vtx->vtx_len < maxlen);
 
                /* Fill the lsa with the values from the vertex.
                 */
                ospf_lsa_set(lsa, vtx);
                pcklen += vtx->vtx_len;
		nbAdvt++;
 
		/* Get the next one...
                 */
                (int)lsa += vtx->vtx_len;

	} PT_WALK_END(&ngb->ngb_RMXbuffer, vtx);

	/* Send the packet.
         */
        pck->pck_lsu_nbadv = htonl(nbAdvt);
        ospf_send((ospf_pck_t *)pck, pcklen, ngb->ngb_intf,
                sockbuild_in(0, ngb->ngb_addr));
}

/* Receive an OSPF packet.
 */
void
ospf_recv __PF1(tp, task *)
{
	int len, err;
	sockaddr_un *dst;
	struct ip *ip;
	ospf_pck_t *pck;
	ospf_intf_t *ospf_intf;

	rt_open(ospf_task);
	ASEXTREE_BUG("ospf_recv 1");

	/* Read all the packets received.
	 */
	while(!task_receive_packet(tp, (size_t *)&len)) {
		/* Get IP header, OSPF packet, and destination.
		 */
		task_parse_ip(ip, pck, ospf_pck_t *);
		dst = sockbuild_in(0, ip->ip_dst.s_addr);
		len -= task_parse_ip_hl(ip);

		ospf_trace_packet(pck, len, 0, TR_PACKET_RECV,
				task_recv_srcaddr, dst);

		/* Validate the packet
		 */
		if ((err = ospf_validate_hdr(pck, len, dst, task_recv_srcaddr,
				&ospf_intf))) {
			trace_log_tf(ospf_trace_options, 0, LOG_ERR,
				("OSPF: invalid packet: %s (%d)",
				trace_value(ospf_errors_msg, err), err));
			continue;
		}

		/* We are supposed to refresh the interface timer. Bummer!
		 */
		if_rtupdate(ospf_intf->ospf_if_ifap);

		/* Process the packet
		 */
		switch(pck->pck_type) {
		case OSPF_PCK_HELLO:
			if ((err = ospf_hello_recv((ospf_hello_pck_t *)pck, len,
					ospf_intf, task_recv_srcaddr)))
				trace_log_tf(ospf_trace_options, 0, LOG_ERR,
					("OSPF: invalid HELLO packet: %s (%d)",
					trace_value(ospf_errors_msg, err),
					err));;
			break;
		case OSPF_PCK_DD:
			if ((err = ospf_dd_recv((ospf_dd_pck_t *)pck, len,
					ospf_intf, task_recv_srcaddr)))
				trace_log_tf(ospf_trace_options, 0, LOG_ERR,
					("OSPF: invalid DD packet: %s (%d)",
					trace_value(ospf_errors_msg, err),
					err));;
			break;
		case OSPF_PCK_LSR:
			if ((err = ospf_lsr_recv((ospf_lsr_pck_t *)pck, len,
                                        ospf_intf, task_recv_srcaddr)))
                                trace_log_tf(ospf_trace_options, 0, LOG_ERR,
                                        ("OSPF: invalid LSR packet: %s (%d)",
                                        trace_value(ospf_errors_msg, err),
                                        err));;
			break;
		case OSPF_PCK_LSU:
			if ((err = ospf_lsu_recv((ospf_lsu_pck_t *)pck, len,
					ospf_intf, task_recv_srcaddr)))
				trace_log_tf(ospf_trace_options, 0, LOG_ERR,
					("OSPF: invalid LSU packet: %s (%d)",
					trace_value(ospf_errors_msg, err),
					err));;
			break;
		case OSPF_PCK_LSA:
			if ((err = ospf_lsa_recv((ospf_lsa_pck_t *)pck, len,
					ospf_intf, task_recv_srcaddr)))
                                trace_log_tf(ospf_trace_options, 0, LOG_ERR,
                                        ("OSPF: invalid LSA packet: %s (%d)",
                                        trace_value(ospf_errors_msg, err),
                                        err));;
                        break;
		default:
			/* Unknown type. Ignore it!
			 */
			break;
		}
	}

	ospf_ngbrmx_flush();
	rt_close(ospf_task, NULL, 0, "OSPF RECV");
	ASEXTREE_BUG("ospf_recv 2");
}

/* Validate a generic OSPF header. Return 0 if it is OK.
 */
int
ospf_validate_hdr __PF5(pck, ospf_pck_t *, len, int, dst, sockaddr_un *,
	src, sockaddr_un *, ospf_intf, ospf_intf_t **)
{
	if_addr *ifap;

	*ospf_intf = NULL;

	/* Check the lenght, version, checksum, authentication field.
	 */
	if (len < sizeof(ospf_pck_header_t))
		return(OSPF_ERR_LENGHT);

	if (len < ntohs(pck->pck_len))	/* not sure about that one... */
		return(OSPF_ERR_LENGHT2);

	if (pck->pck_version != OSPF_VERSION)
		return(OSPF_ERR_VERSION);

	switch(ntohs(pck->pck_authtype)) {
	case OSPF_AUTH_NONE:
		/* No authentication needed
                 */
		if (inet_cksum((void_t) pck, len))
			return(OSPF_ERR_CKSUM);
		break;
	default: /* Why not! */
		return(OSPF_ERR_AUTHTYPE);
	}

	/* From now, interger fields are in host's order.
	 */
	NTOHS(pck->pck_len);
	NTOHS(pck->pck_authtype);

	/* Try to figure out to which interface the packet is sent
	 * If the dst address is AllSPFRouter or AllDRouters then
	 * we have to check the src address agains the interfaces net/mask
	 */
	if (sockaddrcmp_in(dst, all_ospf_routers) ||
			sockaddrcmp_in(dst, all_ospf_dr_routers)) {
		if (!(ifap = if_withdst(src)))
			return(OSPF_ERR_NOMIF);
	}
	/* Well, so the address must be one of our local IP address
	 */
	else if (!(ifap = if_withlcladdr(dst, 0)))
		return(OSPF_ERR_NOIF);

	/* If we don't do OSPF on the interface, get out.
	 */
	if (!ifap->ifa_ospf || ifap->ifa_ospf_state <= OSPF_IFS_DOWN)
		return(OSPF_ERR_NOIF);

	ASSERT(ifap == ifap->ifa_ospf->ospf_if_ifap);

	/* now check the area in the packet. If it's the same as the interface
	 * one, we got it.
	 */
	if (pck->pck_area == ifap->ifa_ospf_area->area_id)
		*ospf_intf = ifap->ifa_ospf;

	else if (pck->pck_area == OSPF_AREAID_BACKBONE) {
		/* The packet is coming from a virtual interface. Look up
		 * all the virtual interfaces.
		 */
		if (!(*ospf_intf = ospf_if_find(pck->pck_rtid)))
			return(OSPF_ERR_NOVIF);
	} else
		 /* The area number is wrong!
		  */
		return(OSPF_ERR_BADAREA);

	/* If this interface does not run OSPF, error!
	 */
	if ((*ospf_intf)->ospf_if_state <= OSPF_IFS_NOOSPF)
		return(OSPF_ERR_INVIF);

	return(OSPF_NOERR);
}

/* Process incoming HELLO packet
 */
int
ospf_hello_recv __PF4(pck, ospf_hello_pck_t *, len, int,
	ospf_intf, ospf_intf_t *, src, sockaddr_un *)
{
	ospf_ngb_t *ngb;
	int MaxNgb, nbNgb, oldstate;

	/* Everyt interger in the header must be in host's oder.
	 */
	NTOHS(pck->pck_hello_int);
	NTOHL(pck->pck_hello_rtdead);

	/* first, the 3 first field must be the same as the interface's ones
	 * On PtoP and virtual interfaces, the mask is ignored.
	 */
	if (!BIT_TEST(ospf_intf->ospf_if_flags, IFS_POINTOPOINT) &&
	 		!BIT_TEST(ospf_intf->ospf_if_flags, IFS_OSPFVLINK) &&
			pck->pck_hello_mask != OSPF_IFMASK(ospf_intf))
		return(OSPF_ERR_HMASK);
	if (pck->pck_hello_int != ospf_intf->ospf_if_helloInt)
		return(OSPF_ERR_HINT);
	if (pck->pck_hello_rtdead != ospf_intf->ospf_if_rtdeadInt)
		return(OSPF_ERR_HRTDEAD);
	if (pck->pck_hello_options != ospf_intf->ospf_if_options)
		return(OSPF_ERR_INVOPTIONS);

	/* Try to find the neighbour (using the ngb ID in the header) in
	 * the current list of the interface ngb. If none, create one, state
	 * DOWN.
	 */
	if (!(ngb = ospf_find_ngb(ospf_intf, pck->pck_hello_rtid)))
		ngb = ospf_new_ngb(ospf_intf, pck->pck_hello_rtid,
			pck->pck_hello_rtpri, SOCK32(src), pck->pck_hello_dr,
			pck->pck_hello_backupDr, OSPF_NGBS_DOWN);
	else if (ngb->ngb_state <= OSPF_NGBS_INIT) {
		/* If we are on a NBMA network, or the neighbour was down,
		 * we reset all the neighbour
		 * parameters from what's on that packet.
		 */
		ngb->ngb_rtpri = pck->pck_hello_rtpri;
		ngb->ngb_addr = SOCK32(src);
		ngb->ngb_dr = pck->pck_hello_dr;
		ngb->ngb_backupDr = pck->pck_hello_backupDr;
	}

	ASSERT(ospf_intf->ospf_if_ngb != ospf_intf->ospf_if_ngb->ngb_next);

	/* Now we do exactly as rfc1583 says page 50! It's just a state machine
	 * processing...
	 */
	ospf_ngb_event(EV_NBG_HELLORECV, ngb);
	oldstate = ngb->ngb_state;

	/* If we are in the hello packet's ngb it's a 2ways event!....
	 */
	MaxNgb = NB_HELLO_NGB(pck->pck_hello_len);
	for(nbNgb = 0; nbNgb < MaxNgb; nbNgb++)
		if (pck->pck_hello_ngb[nbNgb] == ospf_rtid) {
			ospf_ngb_event(EV_NBG_2WAYS, ngb);
			break;
		}

	/* ... else it's a 1way event, so we stop! And we send an hello
	 * packet to tell the other guy we are here!
	 */
	if (nbNgb == MaxNgb) {
		ospf_ngb_event(EV_NBG_1WAY, ngb);
		ospf_send_hello(ngb->ngb_intf);
		return(OSPF_NOERR);
	}

	/* Router priority changes...
	 */
	if (ngb->ngb_rtpri != pck->pck_hello_rtpri) {
		ngb->ngb_rtpri = pck->pck_hello_rtpri;
		ospf_if_event(EV_IF_NGBCHG, ospf_intf);
	}

	/* Well, something with the DR :-) If there is no DR already, then
	 * the BCKUPSEEN event happens. If there is already a DR for this
	 * interface, and this DR changes, then the EV_HELLO_NGBCHG event
	 * happens. The interface state 'WAIT' means the interface has no
	 * DR yet.
	 */
	if (pck->pck_hello_dr == ngb->ngb_addr &&
			pck->pck_hello_backupDr == 0 &&
			ospf_intf->ospf_if_state == OSPF_IFS_WAITING)
		ospf_if_event(EV_IF_BCKUPSEEN, ospf_intf);
	else if ((pck->pck_hello_dr == ngb->ngb_addr) ^
				(ngb->ngb_dr == ngb->ngb_addr)) {
		ngb->ngb_dr = pck->pck_hello_dr;
		ospf_if_event(EV_IF_NGBCHG, ospf_intf);
	}

	/* Same with the backup DR
	 */
	if (pck->pck_hello_backupDr == ngb->ngb_addr &&
			ospf_intf->ospf_if_state == OSPF_IFS_WAITING)
		ospf_if_event(EV_IF_BCKUPSEEN, ospf_intf);
	else if ((pck->pck_hello_backupDr == ngb->ngb_addr) ^
				(ngb->ngb_backupDr == ngb->ngb_addr)) {
		ngb->ngb_backupDr = pck->pck_hello_backupDr;
		ospf_if_event(EV_IF_NGBCHG, ospf_intf);
	}

	/* XXXX The backup DR may have changed. Example: was 0 before, is
	 * ospf_rtid now. Same with DR?
	 */
	ngb->ngb_backupDr = pck->pck_hello_backupDr;
	ngb->ngb_dr = pck->pck_hello_dr;

	/* If the ngb transitioned from lower than OSPF_NGBS_EXSTART to
	 * OSPF_NGBS_EXSTART, then we need to send a DD packet.
	 */
	if (oldstate < OSPF_NGBS_EXSTART && ngb->ngb_state == OSPF_NGBS_EXSTART)
		ospf_send_dd(ngb);

	return(OSPF_NOERR);
}

/* Process incoming Database Description packet
 */
int
ospf_dd_recv __PF4(pck, ospf_dd_pck_t *, len, int, ospf_intf, ospf_intf_t *,
        src, sockaddr_un *)
{
	ospf_ngb_t *ngb;
	int nblsh;
	ospf_lsa_hdr_t *lsh;

	/* find out about the identity of the neighbour
	 */
	if (!(ngb = ospf_find_ngb(ospf_intf, pck->pck_dd_rtid)))
		return(OSPF_ERR_INVNGB);

	/* The options must be the same
	 */
	if (pck->pck_dd_options != ospf_intf->ospf_if_options)
		return(OSPF_ERR_INVOPTIONS);

	/* Now we do what's on the RFC, p50, sec 10.6!
	 * The action depend on the state of the neighbour
	 */
	NTOHL(pck->pck_dd_seq);

	switch(ngb->ngb_state) {
	case OSPF_NGBS_DOWN:
	case OSPF_NGBS_ATTEMPT:
		/* Ignore it!
		 */
		return(OSPF_NOERR);

	case OSPF_NGBS_2WAYS:
		/* Ignore it!
                 */
                return(OSPF_NOERR);

	case OSPF_NGBS_INIT:
		/* the EV_NBG_2WAYS event is executed. If the next state
		 * is > EV_NBG_2WAYS then we fall through the next case
		 */
		ospf_ngb_event(EV_NBG_2WAYS, ngb);
		if (ngb->ngb_state <= OSPF_NGBS_2WAYS)
			return(OSPF_NOERR);

		/* Fall through!
		 */
		assert(ngb->ngb_state == OSPF_NGBS_EXSTART);

	case OSPF_NGBS_EXSTART:
		/* First case, we are slave
		 */
		if (pck->pck_dd_flags == OSPF_DD_FLAGS_MASTER &&
				pck->pck_dd_len == sizeof(ospf_dd_pck_t) &&
				RTID_SSUP(pck->pck_dd_rtid, ospf_rtid)) {
			ngb->ngb_options &= ~(OSPF_DD_FLAGS_MS|OSPF_DD_FLAGS_I);
			ngb->ngb_ddseq = pck->pck_dd_seq;
		}
		else if (!(pck->pck_dd_flags &
					(OSPF_DD_FLAGS_MS|OSPF_DD_FLAGS_I)) &&
				ngb->ngb_ddseq == pck->pck_dd_seq &&
				RTID_SINF(pck->pck_dd_rtid, ospf_rtid)) {
			/* In this case we are Master!
			 */
			ngb->ngb_options |= OSPF_DD_FLAGS_MS;
			ngb->ngb_options &= ~OSPF_DD_FLAGS_I;
		}
		else
			/* Ignore the packet
			 */
			return(OSPF_ERR_DDCONF);

		/* We have to generate a EV_NBG_NEGDONE event,
		 * and fall through again!
		 */
		ospf_ngb_event(EV_NBG_NEGDONE, ngb);
		assert(ngb->ngb_state == OSPF_NGBS_EXCHANGE);

		/* Reset the I flag so the next section's test won't fail.
		 */
		BIT_RESET(pck->pck_dd_flags, OSPF_DD_FLAGS_I);

	case OSPF_NGBS_EXCHANGE:
		/* Check the MASTER flag. Must be consistant with what's stored
		 * BTW, we store *our* state in ngb->ngb_options.
		 */
		if (BIT_MASK_MATCH(ngb->ngb_options, pck->pck_dd_flags,
				OSPF_DD_FLAGS_MS)) {
			ospf_ngb_event(EV_NBG_SEQNUMIS, ngb);
			return(OSPF_ERR_DDMSMIS);
		}

		/* Check the I flag.
		 */
		if (BIT_TEST(pck->pck_dd_flags, OSPF_DD_FLAGS_I)) {
			ospf_ngb_event(EV_NBG_SEQNUMIS, ngb);
			return(OSPF_ERR_DDIMIS);
		}

		/* We are master, and the seq number is in sequence. Process
		 * the packet.
		 */
		if (DD_MASTER(ngb) && ngb->ngb_ddseq == pck->pck_dd_seq)
			break;

		/* We are slave, and the seq number is in sequence...
		 */
		if (DD_SLAVE(ngb) && ngb->ngb_ddseq + 1 == pck->pck_dd_seq)
			break;

		/* We are slave, and got a duplicate. In this case, resent
		 * the last DD packet.
		 */
		if (DD_SLAVE(ngb) && ngb->ngb_ddseq == pck->pck_dd_seq) {
			ospf_send_dd(ngb);
			return(OSPF_NOERR);
		}

		/* Well, too bad, we got a seq number mismatch!
		 */
		ospf_ngb_event(EV_NBG_SEQNUMIS, ngb);
                return(OSPF_ERR_DDSEQMIS);

	case OSPF_NGBS_LOADING:
	case OSPF_NGBS_FULL:
		/* Same as OSPF_NGBS_EXCHANGE, but only duplicates can be
		 * received. If slave, the last DD packet is resent.
		 *
                 * Check the MASTER flag. Must be consistant with what's stored
		 * BTW, we store *our* state in ngb->ngb_options.
                 */
                if (BIT_MASK_MATCH(ngb->ngb_options, pck->pck_dd_flags,
                                OSPF_DD_FLAGS_MS)) {
                        ospf_ngb_event(EV_NBG_SEQNUMIS, ngb);
                        return(OSPF_ERR_DDMSMIS);
                }

                /* Check the I flag.
                 */
                if (BIT_TEST(pck->pck_dd_flags, OSPF_DD_FLAGS_I)) {
                        ospf_ngb_event(EV_NBG_SEQNUMIS, ngb);
                        return(OSPF_ERR_DDIMIS);
                }

                /* We are slave, and got a duplicate. In this case, resent
                 * the last DD packet.
                 */
                if (DD_SLAVE(ngb) && ngb->ngb_ddseq == pck->pck_dd_seq) {
                        ospf_send_dd(ngb);
                        return(OSPF_NOERR);
		}

		/* Well, too bad, we got a seq number mismatch!
                 */
                ospf_ngb_event(EV_NBG_SEQNUMIS, ngb);
                return(OSPF_ERR_DDSEQMIS);

	default:
		assert(0);
	}

	/* we got a good packet. We can flush out the current DD buffer.
	 */
	ospf_ddack(ngb);

	/* At this point we can process the database summaries in the packet.
	 */
	(ospf_dd_pck_t *)lsh = pck + 1;
	nblsh = (pck->pck_dd_len - sizeof(ospf_dd_pck_t)) /
			sizeof(ospf_lsa_hdr_t);
	for(; nblsh; nblsh--, lsh++) {

		NTOH_LSH(lsh);

		/* Check the LS type. Also no AS external if we are a stub area.
		 */
		if (!LS_BOUND(lsh->lsh_type) ||
			(lsh->lsh_type == LST_ASEX &&
			 !BIT_TEST(ospf_intf->ospf_if_options,OSPF_PCK_OPT_E))){
			ospf_ngb_event(EV_NBG_SEQNUMIS, ngb);
			return(OSPF_ERR_INVLSTYPE);
		}

		/* Check if our copy is less recent. If not, ask for the
		 * value. (i.e. put the lsh into the Link state request list.
		 */
		ospf_ls_needupdate((ospf_lsa_t *)lsh, ngb);
	}

	/* If we have something to request, go for it.
         */
	if (ngb->ngb_lsReq.pt_used_nodes)
                ospf_send_lsr(ngb);

	if (DD_MASTER(ngb)) {
		/* If we are master, increment the seq number.
	 	 */
		ngb->ngb_ddseq++;

		/* if the slave has reset the M bit, and we do not
	 	 * have anything to send either, the exchange is done.
		 * Else, send the next DD packet.
		 */
		if (!BIT_TEST(pck->pck_dd_flags, OSPF_DD_FLAGS_M) &&
				!BIT_TEST(ngb->ngb_options, OSPF_DD_FLAGS_M)) {
			ospf_ngb_event(EV_NBG_EXCHDONE, ngb);
			return(OSPF_NOERR);
		}
		ospf_send_dd(ngb);
	}
	else {
		ngb->ngb_ddseq = pck->pck_dd_seq;

		/* If the M bit is cleared, and nothing to be sent,
		 * the exchange is done. As the slave, we need to
		 * send a DD packet anyway.
		 */
		if (!BIT_TEST(pck->pck_dd_flags, OSPF_DD_FLAGS_M) &&
				!BIT_TEST(ngb->ngb_options, OSPF_DD_FLAGS_M)) {
			ospf_send_dd(ngb);
			ospf_ngb_event(EV_NBG_EXCHDONE, ngb);
			return(OSPF_NOERR);
                }
                ospf_send_dd(ngb);
        }

	return(OSPF_NOERR);
}

/* Process incoming Link State Request packet
 */
int
ospf_lsr_recv __PF4(pck, ospf_lsr_pck_t *, len, int, ospf_intf, ospf_intf_t *,
        src, sockaddr_un *)
{
	int nblsr, ind;
	ospf_lsr_data_t *lsr;
	ospf_ngb_t *ngb;
	vertex_t *lsuQueue[NGB_MAXLSRSIZE];
	searchdata_t searchdata;

	/* find out about the identity of the neighbour, and it's state.
         */
        if (!(ngb = ospf_find_ngb(ospf_intf, pck->pck_lsr_rtid)))
                return(OSPF_ERR_INVNGB);

	if (ngb->ngb_state < OSPF_NGBS_EXCHANGE)
		return(OSPF_NOERR);

	nblsr = MIN(NGB_MAXLSRSIZE, PCK_NBLSR(pck->pck_lsr_len));
	lsr = pck->pck_lsr_data;

	/* Go through the list of requests...
	 */
	for(ind = 0; ind < nblsr; ind++, lsr++) {

		NTOH_LSRD(lsr);

		/* Lookup the lsr in the DB. We just have the type and ID.
		 */
		lsuQueue[ind] = ospf_global_find(&ngb->ngb_tree, lsr->lsrd_type,
				&lsr->lsrd_key, &searchdata, NULL);
		if (LS_NOT_INDB(lsuQueue[ind])) {
			/* Something went wrong. Generate a Bad LS Seq event.
			 */
			ospf_ngb_event(EV_NBG_BADLSREQ, ngb);
			return(OSPF_ERR_BADLSR);
		}
	}

	/* Send the whole thing to the neighbour.
	 */
	ospf_send_lsudirect(ngb, lsuQueue, nblsr);
	return(OSPF_NOERR);
}

/* FInd a neighbour of an interface
 */
ospf_ngb_t *
ospf_find_ngb __PF2(ospf_intf, ospf_intf_t *, id, rtid_t)
{
	ospf_ngb_t *ngb;

	OSPF_INTF_NGB(ngb, ospf_intf)
		if (ngb->ngb_rtid == id)
			return(ngb);
	return(NULL);
}

/* HELLO protocol, Neighbour State Machine Processing.
 */
void
ospf_ngb_event __PF2(event, event_t, ngb, ospf_ngb_t *)
{
	int oldstate;

	/* Let do it simply. For each event, we do something. No state
	 * machine table, that's too obscure! Everything from RFC 1583, p46
	 */

	trace_tf(ospf_trace_options, TR_OSPF_STATE, 0,
		("OSPF STATE: NGB %A EVENT %s",
		sockbuild_in(0, ngb->ngb_rtid),
		trace_value(ospf_ngb_events, event)));
	oldstate = ngb->ngb_state;

	switch(event) {
	case EV_NBG_KILLNBR:
		/* first all degrading events!
		 * Well, get rid of everyting! New state is DOWN!
		 */
		ospf_ngb_down(ngb);
		break;

	case EV_NBG_INACTIMER:
		/* get rid of everyting! New state is DOWN!
		 */
		ospf_ngb_down(ngb);
		break;

	case EV_NBG_1WAY:
		/* Important only for states greater than 2 ways.
		 */
		if (ngb->ngb_state >= OSPF_NGBS_2WAYS) {
			/* get rid of everyting! New state is INIT.
			 */
			ospf_ngb_cleanup(ngb);
			OSPF_NGB_NEWSTATE(ngb, OSPF_NGBS_INIT);

			/* If the neighboour was DR or backup, generate a
			 * EV_IF_NGBCHG to reelect someone else.
			 */
			if (NGB_ISDR(ngb) || NGB_ISBACKUP(ngb))
				ospf_if_event(EV_IF_NGBCHG, ngb->ngb_intf);
			break;
		}
		break;

	case EV_NBG_HELLORECV:
		/* Now the upgrading events:
		 * Interesting only if the state is lower than INIT.
		 * We reset the timer.
		 */
		task_timer_set(ngb->ngb_inactimer, 0,
			ngb->ngb_intf->ospf_if_rtdeadInt);

		if (ngb->ngb_state < OSPF_NGBS_INIT)
			OSPF_NGB_NEWSTATE(ngb, OSPF_NGBS_INIT);
		break;

	case EV_NBG_2WAYS:
		/* If the sate is lower than 2WAYS, then upgrade to
	 	 * 2WAYS, check out if we need an adjacency, and do
		 * the appropriate thing! If the state is greater than 2WAYS,
		 * nothing need to be done.
		 */
		/* Well, that the RFC. What i do is to generate a 
		 * EV_IF_NGBCHG event to force the DR election process,
		 * 'cuz the new ngb may be the DR!
		 */
		if (ngb->ngb_state < OSPF_NGBS_2WAYS) {
			OSPF_NGB_NEWSTATE(ngb, OSPF_NGBS_2WAYS);
			ospf_if_event(EV_IF_NGBCHG, ngb->ngb_intf);
		}
		break;

	case EV_NBG_ADJOK:
		/* We have to re-evaluate if we want an adjacency
		 */
		if (ngb->ngb_state < OSPF_NGBS_2WAYS)
			/* Well, not yet!
			 */
			break;

		else if (ngb->ngb_state == OSPF_NGBS_2WAYS) {
			if (ospf_adjacency(ngb, ngb->ngb_intf))
				ospf_exchangeStart(ngb, ngb->ngb_intf);
			break;
		}

		/* The adjacency already exist. If we do not want it anylonger,
		 * destroy it!
		 */
		if (!ospf_adjacency(ngb, ngb->ngb_intf)) {
			ospf_ngb_cleanup(ngb);
			OSPF_NGB_NEWSTATE(ngb, OSPF_NGBS_2WAYS);
		}
		break;

	case EV_NBG_NEGDONE:
		/* the state is now OSPF_NGBS_EXCHANGE. Start sending DD packets
		 */
		ospf_exchangeAll(ngb, ngb->ngb_intf);
		break;

	case EV_NBG_EXCHDONE:
		/* If there is no LS request left, go to FULL state.
		 */
		if (!NGB_LIST_EMPTY_LSR(ngb)) {
			OSPF_NGB_NEWSTATE(ngb, OSPF_NGBS_LOADING);
			break;
		}

		/* Fall through to the FULL state!
		 */
	case EV_NBG_LOADDONE:
		/* Easy. Go to FULL state.
		 */
		OSPF_NGB_NEWSTATE(ngb, OSPF_NGBS_FULL);

		ospf_nbNgbNotReady--;
		DEBUG_NGBREADY("--");

		/* If we are DR for that interface, generate a new NTW LS.
		 */
		if (INTF_ISDR(ngb->ngb_intf))
			ospf_gen_ntwlsa(ngb->ngb_intf);

		/* Will trigger a SPF computation.
		 */
                ospf_gen_rtrlsa(ngb->ngb_intf->ospf_if_area);
		break;

	case EV_NBG_BADLSREQ:
	case EV_NBG_SEQNUMIS:
		/* This should happen only is Exchange state.
		 */
		ASSERT(ngb->ngb_state >= OSPF_NGBS_EXCHANGE);

		/* Reset (empty) all the lists. And go to EXSTART state.
		 * And send a DD packet.
		 */
		ospf_ngb_cleanLists(ngb);
		ospf_exchangeStart(ngb, ngb->ngb_intf);
		ospf_send_dd(ngb);
		break;

	default:
		ASSERT(0);
	}

	/* If the neighbour transitioned from the FULL state to something
	 * else and we are the DR for the interface, we want to send a new NTW
	 * LS.
	 */
	if (oldstate == OSPF_NGBS_FULL && ngb->ngb_state != OSPF_NGBS_FULL &&
			INTF_ISDR(ngb->ngb_intf))
		ospf_gen_ntwlsa(ngb->ngb_intf);
}

/* Interface State Machine.
 */
int
ospf_if_event __PF2(event, event_t, ospf_intf, ospf_intf_t *)
{
        /* Let do it simply. For each event, we do something. No state
         * machine table, that's too obscure! Everything from RFC 1583, p37
         */

        trace_tf(ospf_trace_options, TR_OSPF_STATE, 0,
                ("OSPF STATE: INTERFACE %A EVENT %s ",
			sockbuild_in(0, ospf_intf->ospf_if_addr),
                trace_value(ospf_if_events, event)));

	switch(event) {
	case EV_IF_BCKUPSEEN:
		/* That only happens when the state if WAIT.
		 */
		assert(ospf_intf->ospf_if_state == OSPF_IFS_WAITING);
		ospf_electDR(ospf_intf);
		break;

	case EV_IF_NGBCHG:
		/* something change. Re-elect the DR
		 */
		ospf_electDR(ospf_intf);
		break;

	case EV_IF_UP:
		ospf_if_add(ospf_intf);
		break;

	case EV_IF_DOWN:
		ospf_if_down(ospf_intf);

		/* Generate a new router LSA about ourselves.
         	 */
		ospf_gen_rtrlsa(ospf_intf->ospf_if_area);
        	/*OSPF_GEN_RTRLSA();*/
		break;

	case EV_IF_LOOPON:
	case EV_IF_LOOPOFF:
		/* We don't care yet.
		 */
		break;

	case EV_IF_WAITTIMER:
		/* We vote for a new DR??!!!?!?
		 */
		ospf_electDR(ospf_intf);
		break;

	default:
		ASSERT(0);
	}

	return(0);
}

/* Election of a DR.
 */
int
ospf_electDR __PF1(ospf_intf, ospf_intf_t *)
{
	rtid_t oldDrID, olbBackupDrID, backupdrID, drID;
	ipaddr_t backupdrAddr, drAddr;
	u_int32 pri;
	int count;
	ospf_ngb_t *ngb;

#define PRIORITY_BU_LOCAL(ifap)	(ospf_intf->ospf_if_rtpri + \
	((ospf_intf->ospf_if_backupDrAddr == ospf_intf->ospf_if_addr) << 8))
#define PRIORITY_BU_NGB(ngb)	(ngb->ngb_rtpri + \
	((ngb->ngb_backupDr == ngb->ngb_addr) << 8))
#define PRIORITY_DR_NGB(ngb)    (ngb->ngb_rtpri + \
	((ngb->ngb_dr == ngb->ngb_addr) << 9))

	/* The wait timer for the interface must be stoped.
	 */
	task_timer_reset(ospf_intf->ospf_if_waitTimer);

	/* Just follow the rfc, page 39. PRIORITY_BU_LOCAL and
	 * PRIORITY_BU_NGB gives the 'real' priority of a backup router,
	 * i.e. if the router thinks it is the backup then its priority
	 * is greater than any other routers which do not think they  are
	 * the backup DR. Same with PRIORITY_DR_XX, except that  PRIORITY_DR_XX
	 * is greater than PRIORITY_BU_XX so only router which designed
	 * themselves as DR can be elected.
	 * oldDr and olbBackupDr are router's ID.
	 */
	oldDrID = ospf_intf->ospf_if_drID;
	olbBackupDrID = ospf_intf->ospf_if_backupDrID;
	count = -1;
	do {
		/* Get the backup router. First we suppose we are that one if
		 * we are not the DR.
		 */
		if (!INTF_ISDR(ospf_intf)) {
			pri = PRIORITY_BU_LOCAL(ospf_intf);
			backupdrID = ospf_rtid;
			backupdrAddr = ospf_intf->ospf_if_addr;
		}
		else {
			pri = 0;
			backupdrID = 0;
			backupdrAddr = 0;
		}

		/* Step 2. Look through all neighbours.
		 */
		OSPF_INTF_NGB(ngb, ospf_intf) {
			/* non valid ones
			 */
			if (ngb->ngb_state < OSPF_NGBS_2WAYS || !ngb->ngb_rtpri
					|| ngb->ngb_addr == ngb->ngb_dr)
				continue;

			/* Is the priority better?
			 */
			if (pri < PRIORITY_BU_NGB(ngb) ||
					(pri == PRIORITY_BU_NGB(ngb) &&
					RTID_SSUP(ngb->ngb_rtid, backupdrID))) {
				pri = PRIORITY_BU_NGB(ngb);
				backupdrID = ngb->ngb_rtid;
				backupdrAddr = ngb->ngb_addr;
			}
		}

		trace_tf(ospf_trace_options, TR_OSPF_DRELECT, 0,
    ("OSPF DR ELECTION: 1 interface %A: DR ID %A (IP %A): Backup ID %A (IP %A)",
                sockbuild_in(0, ospf_intf->ospf_if_addr),
                sockbuild_in(0, drID), sockbuild_in(0, drAddr),
                sockbuild_in(0, backupdrID), sockbuild_in(0, backupdrAddr)));

		/* Get the DR router. Take the last Backup found or ourself
		 * if we are DR.
		 */
		if (INTF_ISDR(ospf_intf)) {
			pri = PRIORITY_BU_LOCAL(ospf_intf);
			drID = ospf_rtid;
			drAddr = ospf_intf->ospf_if_addr;
		}
		else {
			pri = pri;
			drID = backupdrID;
			drAddr = backupdrAddr;
		}
		OSPF_INTF_NGB(ngb, ospf_intf) {
			/* non valid ones
			 */
			if (ngb->ngb_state < OSPF_NGBS_2WAYS || !ngb->ngb_rtpri
					|| ngb->ngb_addr != ngb->ngb_dr)
				continue;
			/* Is the priority better?
                         */
                        if (pri < PRIORITY_DR_NGB(ngb) ||
					(pri == PRIORITY_DR_NGB(ngb) &&
					RTID_SSUP(ngb->ngb_rtid, drID))) {
				pri = PRIORITY_DR_NGB(ngb);
				drID = ngb->ngb_rtid;
				drAddr = ngb->ngb_addr;
			}
		}

		trace_tf(ospf_trace_options, TR_OSPF_DRELECT, 0,
    ("OSPF DR ELECTION: 2 interface %A: DR ID %A (IP %A): Backup ID %A (IP %A)",
                sockbuild_in(0, ospf_intf->ospf_if_addr),
                sockbuild_in(0, drID), sockbuild_in(0, drAddr),
                sockbuild_in(0, backupdrID), sockbuild_in(0, backupdrAddr)));

		/* Now we have a DR, and backup. We are suppose to repeat
		 * the election if we are the new DR or Backup, or if we are
		 * no longer the DR or backup. And we are supposed to do that
		 * once more only!
		 */
		ospf_intf->ospf_if_drID = drID;
		ospf_intf->ospf_if_backupDrID = backupdrID;
		ospf_intf->ospf_if_drAddr = drAddr;
		ospf_intf->ospf_if_backupDrAddr = backupdrAddr;

		if (++count)
			break;

	} while((drID == ospf_rtid && oldDrID != drID) ||
		(backupdrID == ospf_rtid && olbBackupDrID != backupdrID) ||
		(drID != ospf_rtid && oldDrID == ospf_rtid) ||
		(backupdrID != ospf_rtid && olbBackupDrID == ospf_rtid));

	trace_tf(ospf_trace_options, TR_OSPF_DRELECT, 0,
      ("OSPF DR ELECTION: interface %A: DR ID %A (IP %A): Backup ID %A (IP %A)",
		sockbuild_in(0, ospf_intf->ospf_if_addr),
		sockbuild_in(0, drID), sockbuild_in(0, drAddr),
		sockbuild_in(0, backupdrID), sockbuild_in(0, backupdrAddr)));

	/* A router should not elect itseft as DR and BDR. (But it seems
	 * that a router can elect another one as DR *and* BDR, C.f. note (5)
	 * p36, RFC1583)
	 */
	ASSERT(ospf_intf->ospf_if_drID &&
		(ospf_intf->ospf_if_drID != ospf_rtid ||
		ospf_intf->ospf_if_drID != ospf_intf->ospf_if_backupDrID));

	/* If we are the new DR or Backup, then change the state,
	 * register for the multicast DR address.
	 * Then send a EV_NBG_ADJOK to the neighbours if something changed.
	 */
	if (INTF_ISDR(ospf_intf)) {
		OSPF_IF_NEWSTATE(ospf_intf, OSPF_IFS_DR);
		if (BIT_TEST(ospf_intf->ospf_if_flags, IFS_MULTICAST))
                	OSPF_JOINT_GROUP(ospf_intf, all_ospf_dr_routers);

		OSPF_INTF_NGB(ngb, ospf_intf)
			ospf_ngb_event(EV_NBG_ADJOK, ngb);

		/* Generate a network LS for that network.
		 */
		ospf_gen_ntwlsa(ospf_intf);
	}
	else {
		/* We are not DR. If We where announcing something before,
		 * flush it out.
		 */
		if (ospf_intf->ospf_if_ntls) {
			VTX_AGE_SET(ospf_intf->ospf_if_ntls, ospf_maxAge);
			BIT_SET(ospf_intf->ospf_if_ntls->vtx_flags, LF_DELETE);
			ospf_flood_selforiginate(ospf_area_list,
				ospf_intf->ospf_if_ntls);
		}

		/* If we are now the backup, joint the multicast group,
		 * and check the adjacencies.
		 */
		if (INTF_ISBACKUP(ospf_intf)) {
			OSPF_IF_NEWSTATE(ospf_intf, OSPF_IFS_BACKUPDR);
			if (BIT_TEST(ospf_intf->ospf_if_flags, IFS_MULTICAST))
                        	OSPF_JOINT_GROUP(ospf_intf,all_ospf_dr_routers);

			OSPF_INTF_NGB(ngb, ospf_intf)
				ospf_ngb_event(EV_NBG_ADJOK, ngb);
		}
		else {
			OSPF_IF_NEWSTATE(ospf_intf, OSPF_IFS_DROTHER);
			if (BIT_TEST(ospf_intf->ospf_if_flags, IFS_MULTICAST))
				OSPF_QUIT_GROUP(ospf_intf, all_ospf_dr_routers);
	
			OSPF_INTF_NGB(ngb, ospf_intf)
                        	ospf_ngb_event(EV_NBG_ADJOK, ngb);
		}
	}

	/* Finally, we generate a new router LSA about our own router.
	 */
	ospf_gen_rtrlsa(ospf_intf->ospf_if_area);
	/*OSPF_GEN_RTRLSA();*/

	return(0);
}

/* Return 1 if we should create an adjacency between an interface and a
 * neighbour.
 */
int
ospf_adjacency __PF2(ngb, ospf_ngb_t *, ospf_intf, ospf_intf_t *)
{
	/* From rfc, p49:
	 * The interface is P2P or virtual link
	 */
	if (BIT_TEST(ospf_intf->ospf_if_flags, IFS_POINTOPOINT|IFS_OSPFVLINK))
		return(1);

	/* The router is the DR or backup
	 */
	if (INTF_ISDR(ospf_intf) || INTF_ISBACKUP(ospf_intf))
		return(1);

	/* The neighbour is the DR or backup
	 */
	if (NGB_ISDR(ngb) || NGB_ISBACKUP(ngb))
		return(1);

	return(0);
}

/* Start a database exchange with a neighbour. Called when a new adjacency
 * is set (after Hello or DD packets), or when a seq mismatch happens.
 */
void
ospf_exchangeStart __PF2(ngb, ospf_ngb_t *, ospf_intf, ospf_intf_t *)
{
	/* All retransmision list of any kind must be empty.
	 */
	ASSERT(!ngb->ngb_lsReq.pt_used_nodes);
	ASSERT(!ngb->ngb_LSRbuffer.pt_used_nodes);
	ASSERT(!ngb->ngb_lsRmx.pt_used_nodes);
	ASSERT(!ngb->ngb_LSRbuffer.pt_used_nodes);
	ASSERT(!ngb->ngb_lastAckIndex);
	ASSERT(!ngb->ngb_lastDDsize);


	/* Set the number of 'busy' neighbours.
	 */
	if (ngb->ngb_state != OSPF_NGBS_EXSTART &&
                        ngb->ngb_state != OSPF_NGBS_EXCHANGE &&
                        ngb->ngb_state != OSPF_NGBS_LOADING) {
                ospf_nbNgbNotReady++;
		DEBUG_NGBREADY("++");
	}

	/* First we go to the OSPF_NGBS_EXSTART state.
	 */
	OSPF_NGB_NEWSTATE(ngb, OSPF_NGBS_EXSTART);

	/* Set the Master/slave options. and initialize the sequence number.
	 */
	ngb->ngb_ddseq = time_sec;
	ngb->ngb_options = OSPF_DD_FLAGS_MASTER;

	/* then we are suposed to send empty DD packets to negotiate
	 * the Master/Slave relationship. That's done every RxmtInterval
	 * until the state changes (RRFC p46).
	 */
	if (!ngb->ngb_rmxTimer) {
		char timerName[OSPF_MAX_TIMER_NAME];
 
                (void)sprintf(timerName, "%A Rmx",
			sockbuild_in(0, ngb->ngb_addr));
		ngb->ngb_rmxTimer = task_timer_create(ospf_task,
			strdup(timerName), TIMERF_HIPRIO, 0,
			ospf_intf->ospf_if_rmxInt, ospf_rmx_timer, (void_t)ngb);
	}
	else
		task_timer_set(ngb->ngb_rmxTimer, 0, ospf_intf->ospf_if_rmxInt);
}

/* It's time to send a DD packet, and whatever is waiting...
 */
void
ospf_rmx_timer __PF2(tip, task_timer *, interval, time_t)
{
	ospf_ngb_t *ngb;

	ngb = (ospf_ngb_t *)tip->task_timer_data;

	/* We retransmit whatever is in the buffer.
	 * XXX maybe do not retransmit what has not timed out?
	 * When in OSPF_NGBS_EXSTART state, retransmit the empty DD packet
	 * always, since nothing is buffered.
	 */
	if (ngb->ngb_lastDDsize || ngb->ngb_state == OSPF_NGBS_EXSTART)
		ospf_send_dd(ngb);

	if (ngb->ngb_LSRbuffer.pt_used_nodes)
		ospf_send_lsr(ngb);

	if (ngb->ngb_RMXbuffer.pt_used_nodes)
		ospf_send_rmx(ngb);

	/* Reset the timer for the next shot.
 	 */
	task_timer_set(tip, 0, ngb->ngb_intf->ospf_if_rmxInt);
}

/* Now that the Master/Slave relationship is defined, we start exchanging
 * the DB.
 */
void
ospf_exchangeAll __PF2(ngb, ospf_ngb_t *, ospf_intf, ospf_intf_t *)
{
	/* The state is now OSPF_NGBS_EXCHANGE
	 */
	OSPF_NGB_NEWSTATE(ngb, OSPF_NGBS_EXCHANGE);

	/* The transmisson timer is stoped
	 */
	task_timer_reset(ngb->ngb_rmxTimer);

	/* All list and list pointer are reseted.
	 */
	ospf_ngb_cleanLists(ngb);

	/* first put all the database object into the Database summary list.
	 */
	ospf_buildDataSum(ngb);

	/* Set the transmisson timer if we are master. (C.f. RFC p53).
	 */
	if (BIT_TEST(ngb->ngb_options, OSPF_DD_FLAGS_MS)) {
		task_timer_set(ngb->ngb_rmxTimer, 0, ospf_intf->ospf_if_rmxInt);
	}
}

/* The neighbour in going to the DOWN (or ATTEMPT) state.
 */
void
ospf_ngb_down __PF1(ngb, ospf_ngb_t *)
{
	/* Reset the number of busy neighbours
         */
        if (ngb->ngb_state == OSPF_NGBS_EXSTART ||
                        ngb->ngb_state == OSPF_NGBS_EXCHANGE ||
                        ngb->ngb_state == OSPF_NGBS_LOADING) {
                ospf_nbNgbNotReady--;
                DEBUG_NGBREADY("--");
        }
 
        /* If we are not NBMA network, the router is dead. Down. Else
         * we still want to try to connect to the guy.
         */
        if (BIT_TEST(ngb->ngb_intf->ospf_if_flags, IFS_BROADCAST))
                OSPF_NGB_NEWSTATE(ngb, OSPF_NGBS_DOWN);
        else {
		/* Set the hello timer to fire every PoolInt sec. On NBMA
		 * network we are still trying to talk to the ngb event when
		 * dead.
		 */
		ngb->ngb_lastHello = time_sec;
                OSPF_NGB_NEWSTATE(ngb, OSPF_NGBS_ATTEMPT);
	}
 
	/* Empty all lists, timers.
         */
        ospf_ngb_cleanLists(ngb);

	/* If the neighboour was DR or backup, generate a
	 * EV_IF_NGBCHG to reelect someone else.
	 */
	if (NGB_ISDR(ngb) || NGB_ISBACKUP(ngb))
		ospf_if_event(EV_IF_NGBCHG, ngb->ngb_intf);
 
        task_timer_reset(ngb->ngb_inactimer);
}

/* The neighbour is no longer part of an adjacency. Clean up some lists.
 */
void
ospf_ngb_cleanup __PF1(ngb, ospf_ngb_t *)
{
	/* Reset the number of busy neighbours
	 */
	if (ngb->ngb_state == OSPF_NGBS_EXSTART ||
			ngb->ngb_state == OSPF_NGBS_EXCHANGE ||
			ngb->ngb_state == OSPF_NGBS_LOADING) {
		ospf_nbNgbNotReady--;
		DEBUG_NGBREADY("--");
	}

	/* Empty all lists, timers.
	 */
	ospf_ngb_cleanLists(ngb);

	task_timer_reset(ngb->ngb_inactimer);
}

/* cleanup all transmission lists and associated timers.
 */
void
ospf_ngb_cleanLists __PF1(ngb, ospf_ngb_t *)
{
	/* XXX The vtx_qref count should be decreased when deleting the RMX
	 * list.
	 */
	ngb->ngb_lastAckIndex = 0;
	NGB_LIST_RESET_DS(ngb);
	NGB_LIST_RESET_LSR(ngb);
	NGB_LIST_RESET_RMX(ngb);

	if (ngb->ngb_rmxTimer)
		task_timer_reset(ngb->ngb_rmxTimer);
	if (ngb->ngb_ackTimer)
		task_timer_reset(ngb->ngb_ackTimer);

	ASSERT(!ngb->ngb_lsReq.pt_used_nodes);
        ASSERT(!ngb->ngb_LSRbuffer.pt_used_nodes);
        ASSERT(!ngb->ngb_lsRmx.pt_used_nodes);
        ASSERT(!ngb->ngb_LSRbuffer.pt_used_nodes);
        ASSERT(!ngb->ngb_lastAckIndex);
        ASSERT(!ngb->ngb_lastDDsize);
}

/* Run when the one shot inactivity timer expires.
 */
void
ospf_ngb_inactimer __PF2(tip, task_timer *, interval, time_t)
{
	ospf_ngb_t *ngb;

	rt_open(ospf_task);
	ngb = (ospf_ngb_t *)tip->task_timer_data;
	ospf_ngb_event(EV_NBG_INACTIMER, ngb);
	rt_close(ospf_task, NULL, 0, "OSPF InacTimer");
}

/* Stuff to do before the config file is re-read.
 */
void
ospf_cleanup __PF1(tp, task *)
{
}

/* Dump the OSPF module
 */
void
ospf_dump __PF2(tp, task *, fd, FILE *)
{
}

/* Called before normal processing begins
 */
void
ospf_reinit __PF1(tp, task *)
{
}

/* Called when gated exits or when OSPF is turned off in the config file.
 */
void
ospf_terminate __PF1(tp, task *)
{
	ospf_area_t *area, *next;

	/* We remove all areas.
	 */
	for(area = ospf_area_list; area; area = next) {
		next = area->area_next;
		ospf_area_remove(area);
	} 

	/* Delete the task.
	 */
	task_delete(tp);

	/* Clean up (reset) some global variables...
	 */
	ospf_default_area = NULL;
	ospf_area_list = NULL;
	ospf_spf_timer = NULL;
	ospf_defaultGwList = NULL;
	ospf_nbNgbNotReady = 0;
	ospf_flush_list = NULL;
	ospf_refresh_timer = NULL;
	ospf_defaultGateway = NULL;
	ospf_refresh_start = 0;
	ospf_refresh_job = NULL;
	ospf_refresh_walk = NULL;
	ospf_refresh_area = NULL;
}

/* Used to add a new interface in an area during configuration.
 */
void
ospf_conf_newintf __PF1(iflist, iflist_t *)
{
	if_addr *ifap;

	/* Go through the current list of interfaces.
	 */
	IFAP_LIST(iflist, ifap) {
		(void)ospf_new_intf(SOCK32(ifap->ifa_addr), ospf_current_area,
			0);
	} IFAP_LIST_END(iflist, ifap);
}
#define OSPF_SET_IFPARAM(type, data, oif, changed) { \
	static ospf_ngb_t *__ngb; \
        switch(type) { \
        case CFG_ENABLE: \
		(changed) = ((oif)->ospf_if_state == OSPF_IFS_DOWN)? 0: 1; \
                OSPF_IF_NEWSTATE((oif), OSPF_IFS_DOWN); \
                break; \
        case CFG_DISABLE: \
		(changed) = ((oif)->ospf_if_state == OSPF_IFS_NOOSPF)? 0: 1; \
                OSPF_IF_NEWSTATE((oif), OSPF_IFS_NOOSPF); \
                break; \
        case CFG_RXMINT: \
		(changed) = ((oif)->ospf_if_rmxInt == *(time_t *)data)? 0: 1; \
                (oif)->ospf_if_rmxInt = *(time_t *)data; \
                break; \
        case CFG_TRANSDEL: \
		(changed) = ((oif)->ospf_if_rmxInt == *(time_t *)data)? 0: 1; \
                (oif)->ospf_if_transDel = *(time_t *)data; \
                break; \
        case CFG_RTPRI: \
		(changed) = ((oif)->ospf_if_rtpri == *(int *)data)? 0: 1; \
                (oif)->ospf_if_rtpri = *(int *)data; \
                break; \
        case CFG_HELLOINT: \
		(changed) = ((oif)->ospf_if_helloInt == *(time_t *)data)? 0: 1;\
                (oif)->ospf_if_helloInt = *(time_t *)data; \
                break; \
        case CFG_RTDEAD: \
		(changed) = ((oif)->ospf_if_rtdeadInt ==*(time_t *)data)? 0: 1;\
                (oif)->ospf_if_rtdeadInt = *(time_t *)data; \
                break; \
        case CFG_NOMULTI: \
                /* XXX */ \
                break; \
        case CFG_AUTH: \
		(changed) = ((oif)->ospf_if_authType == *(auth_t *)data)? 0: 1;\
                (oif)->ospf_if_authType = *(auth_t *)data; \
                break; \
        case CFG_AUTH2: \
		(changed) = ((oif)->ospf_if_authType2 == *(auth_t *)data)? 0:1;\
                (oif)->ospf_if_authType2 = *(auth_t *)data; \
                break; \
        case CFG_POLINT: \
		(changed) = ((oif)->ospf_if_pollInt == *(time_t *)data)? 0: 1;\
                (oif)->ospf_if_pollInt = *(time_t *)data; \
                break; \
        case CFG_RTLIST: \
		(changed) = 0; \
		if (!(__ngb = ospf_find_ngb((oif), \
				SOCK32((sockaddr_un *)data))))\
			__ngb = ospf_new_ngb((oif), \
				SOCK32((sockaddr_un *)data), \
				0, 0, 0, 0, OSPF_NGBS_DOWN); \
		else \
			__ngb->ngb_changes |= CFG_OLD; \
		__ngb->ngb_changes |= CFG_KEEP; \
                break; \
	case CFG_RTELIG: \
		(changed) = 0; \
		__ngb->ngb_rtpri = *(int *)data; \
		break; \
        case CFG_COST: \
		(changed) = ((oif)->ospf_if_cost == *(int *)data)? 0: 1; \
                (oif)->ospf_if_cost = *(int *)data; \
                break; \
        default: \
                ASSERT(0); \
        } \
}

/* Used to configure an interface during the config file reading.
 * iflist is the list of interfaces which 
 * need to be initialized with the correct parameter.
 */
void
ospf_conf_setintf __PF4(iflist, iflist_t *, vlink, ospf_intf_t *,
		 type, int, data, void_t)
{
        if_addr *ifap;
	int changed;
	ospf_intf_t *ospf_intf;

	/* Go through all Interfaces in the list. If the list is null, that
	 * means that we are setting the options of a virtual link.
	 */
	if (iflist) {
		IFAP_LIST(iflist, ifap) {

			ospf_intf = ospf_if_find(SOCK32(ifap->ifa_addr));
			ASSERT(ospf_intf);
			OSPF_SET_IFPARAM(type, data, ospf_intf, changed);

			/* For reconfiguration only. If the intf is marked
			 * as CFG_OLD, we set the change flag.
			 */
			if (changed &&
				BIT_TEST(ospf_intf->ospf_if_changes, CFG_OLD))
				BIT_SET(ospf_intf->ospf_if_changes, type);
	
		} IFAP_LIST_END(iflist, ifap);
	}
	else {
		OSPF_SET_IFPARAM(type, data, vlink, changed);

		/* For reconfiguration only. If the intf is marked
                 * as CFG_OLD, we set the change flag.
                 */
                BIT_SET(vlink->ospf_if_changes, type);
	}
}

/* Count the number of neighbour of a given interface
 */
int
ospf_count_ngb __PF1(ospf_intf, ospf_intf_t *)
{
	int count;
	ospf_ngb_t *ngb;

	count = 0;
	OSPF_INTF_NGB(ngb, ospf_intf)
		count++;

	return(count);
}

/* find an OSPF interface based on its IP address.
 */
ospf_intf_t *
ospf_if_find __PF1(addr, ipaddr_t)
{
	ospf_intf_t *ospf_intf;

	OSPF_IF_ADDR(ospf_intf) {
		if (ospf_intf->ospf_if_addr == addr)
			return(ospf_intf);
	} OSPF_IF_ADDR_END(ospf_intf);
	return(NULL);
}

/* Initialise the areas.
 */
void
ospf_area_init __PF0(void)
{
	ospf_area_t *area;
	if_addr *ifap;

	/* If none area defined, create a default one.
	 */
	if (!ospf_area_list)
		ospf_default_area = ospf_area_create(OSPF_AREAID_BACKBONE,
			AREA_DEFAULT_OPT, 0, OSPF_AUTH_NONE);

        /* If we have only one area without any interface, by default
         * we add all interfaces to that area.
         */
        if (!ospf_area_list->area_next && !ospf_area_list->area_intf) {
                IF_ADDR(ifap) {
                        ospf_init_if(ospf_area_list, ifap);
                } IF_ADDR_END(ifap);
        }

	/* Set the router LSA options (the External and Area Bondary flags)
	 */
	OSPF_ALL_AREA(area) {
		area->area_rtrbits = ospf_rtrbits;
	} OSPF_ALL_AREA_END(area)
}

/* Create an area. If the area already exist, keep it.
 */
ospf_area_t *
ospf_area_create __PF4(id, areaid_t, options, flag_t, rtrbits, flag_t,
		auth, auth_t)
{
	ospf_area_t *area;

	/* Area ID are stores in Network Byte order (Big Endian)
	 */
	HTONL(id);

	/* Find out if the area already exist (when doing reconfig)
	 */
	if (!(area = ospf_find_area(id))) {
		area = (ospf_area_t *)task_block_alloc(ospf_area_blocks);
		(void)bzero((char *)area, sizeof(ospf_area_t));
		area->area_next = ospf_area_list;
		ospf_area_list = area;
		area->area_id = id;
		area->area_tree.pt_area = area;
	}
	else
		BIT_SET(area->area_changes, CFG_OLD);

	/* We'll keep it.
	 */
	BIT_SET(area->area_changes, CFG_KEEP);
	area->area_rtrbits = rtrbits;
	area->area_options = options;

	/* Each area has a default gw_entry structure to be used by the
	 * vertices.
	 */
	area->area_defaultgw = gw_locate(&area->area_gwlist, RTPROTO_OSPF,
		ospf_task, 0, 0, sockbuild_in(0, area->area_id), 0);

	return(area);
}

/* Find an area given its ID.
 */
ospf_area_t *
ospf_find_area __PF1(id, areaid_t)
{
	ospf_area_t *area;

	OSPF_ALL_AREA(area) {
		if (area->area_id == id)
			return(area);
	}
	return(NULL);
}

/* Create a new neighbour.
 */
ospf_ngb_t *
ospf_new_ngb __PF7(ospf_intf, ospf_intf_t *, rtid, rtid_t, rtpri, rtpri_t,
		addr, ipaddr_t, dr, ipaddr_t,  bdr, ipaddr_t, state, int)
{
	ospf_ngb_t *ngb;
	char timerName[OSPF_MAX_TIMER_NAME];

	ngb = (ospf_ngb_t *)task_block_alloc(ospf_ngb_blocks);
        (void)bzero((void_t)ngb, sizeof(ngb));

        ngb->ngb_rtid     = rtid;
        ngb->ngb_rtpri    = rtpri;
        ngb->ngb_addr     = addr;
        ngb->ngb_dr       = dr;
        ngb->ngb_backupDr = bdr;
        OSPF_NGB_NEWSTATE(ngb, state);
        OSPF_IF_ADDNGB(ospf_intf, ngb);
        ngb->ngb_ackDelay = ngb->ngb_intf->ospf_if_rtdeadInt / 2;
 
        (void)sprintf(timerName, "%A Inact",
		sockbuild_in(0, ospf_intf->ospf_if_addr));
        ngb->ngb_inactimer = task_timer_create(ospf_task, strdup(timerName),
                0, 0, 0, ospf_ngb_inactimer, (void_t) ngb);
	(void)sprintf(timerName, "%A Ack",
		sockbuild_in(0, ospf_intf->ospf_if_addr));
        ngb->ngb_ackTimer = task_timer_create(ospf_task, strdup(timerName),
                0, 0, 0, ospf_ngb_acktimer, (void_t) ngb);

	return(ngb);
}

/* Find out where to send the packet.
 */
sockaddr_un *
ospf_send_addr __PF3(ospf_intf, ospf_intf_t *, pcktype, int, ngb, ospf_ngb_t *)
{
	switch(pcktype) {
	case SEND_DD:
	case SEND_LSR:
		/* for those guy, just sent it to the neigbour.
		 * ospf_send_addr() should not be called for this kind of
		 * packets.
		 */
		return(sockbuild_in(0, ngb->ngb_addr));
	case SEND_HELLO:
		/* For virtual Interface, send to the ngb. For p2p and
		 * breadcast, send to AllSPF. For NBMA net, if DR send to
		 * AllSPF, if eligible, send to AllEligible, else send to
		 * DR.
		 */
		if (OSPF_ISVIRTUAL(ospf_intf))
			return(sockbuild_in(0, ngb->ngb_addr));
		else if (BIT_TEST(ospf_intf->ospf_if_flags, IFS_MULTICAST))
			return(all_ospf_routers);
		else if (INTF_ISDR(ospf_intf) || INTF_ISBACKUP(ospf_intf))
			return(all_ospf_routers);
		else if (ospf_intf->ospf_if_rtpri)
			return(all_ospf_eligible);
		else
			return(all_ospf_dr_routers);
	case SEND_LSU_DIRECT:
		/* Looks like it this case we send a unicast packet to
		 * the neighbour. I am not sure. XXXXX
		 */
		return(sockbuild_in(0, ngb->ngb_addr));
	case SEND_LSU_FLOODING:
	case SEND_LSA_DELAYED:
		/* For broadcast nework, send to AllSPF if DR, or AllDR
		 * if not. For NBMA, the same. Other, send to ngb.
		 */
		if (BIT_TEST(ospf_intf->ospf_if_flags, IFS_MULTICAST) ||
			!BIT_TEST(ospf_intf->ospf_if_flags,
                        	IFS_POINTOPOINT|IFS_OSPFVLINK))
			if (INTF_ISDR(ospf_intf) || INTF_ISBACKUP(ospf_intf))
				return(all_ospf_routers);
			else
				return(all_ospf_dr_routers);

		return(sockbuild_in(0, ngb->ngb_addr));
	case SEND_LSU_RMX:
	case SEND_LSA_DIRECT:
		/* Retransmisison and direct ACK are always sent to the
		 * neighbour only.
		 */
		return(sockbuild_in(0, ngb->ngb_addr));
	default:
		ASSERT(0);
	}

	return(NULL);
}

/* Called after a reconfig. Clean up all removed or changed
 * interfaces/neighbours/areas.
 */
void
ospf_reconfig __PF0(void)
{
	ospf_ngb_t **ngb, *lostngb;
        ospf_area_t **area, *lostarea;
        ospf_intf_t **ospf_intf, *lostoif;

	rt_open(ospf_task);

	ASEXTREE_BUG("ospf_reconfig 1");

	/* Go through every thing. If CFG_KEEP is set, keep the object.
	 */
        for(area = &ospf_area_list; *area;) {
                if (!BIT_TEST((*area)->area_changes, CFG_KEEP)) {
			lostarea = *area;
			*area = (*area)->area_next;
			ospf_area_remove(lostarea);
			ASEXTREE_BUG("ospf_reconfig 2");
			continue;
		}

		/* If the  router status has changed (no more border or external
		 * router), we may want to delete some stuff?
		 * First case, the router is no more a area border router.
		 */
		if (BIT_TEST(ospf_global_changes, LSRTR_BITS_B) &&
				!BIT_TEST(ospf_rtrbits, LSRTR_BITS_B)) {
			/* XXXXX */
		}
		/* If the router is no more a external AS router, well,
		 * ospf_newpolicy() should remove all the ASEX LS from the DB.
		 */
 
		/* check for any deleted/changed interfaces.
		 */
		ASEXTREE_BUG("ospf_reconfig 2");
                for(ospf_intf = &(*area)->area_intf; *ospf_intf;) {

			if (!BIT_TEST((*ospf_intf)->ospf_if_changes,
					CFG_KEEP)) {
				lostoif = *ospf_intf;
				*ospf_intf = (*ospf_intf)->ospf_if_next;
				ospf_intf_remove(lostoif);
				ASEXTREE_BUG("ospf_reconfig 3");
				continue;
			}
			/* If some of the timing parameters changed, we
			 * need to reset the adjacencies. Also it the
			 * area options changed. Or just the area.
			 */
			if ((BIT_TEST((*ospf_intf)->ospf_if_changes, CFG_OLD) &&
				BIT_TEST((*ospf_intf)->ospf_if_changes,
				CFG_AREA|CFG_RTPRI|CFG_HELLOINT|CFG_RTDEAD)) ||
				(BIT_TEST((*area)->area_changes, CFG_OLD) &&
				BIT_TEST((*area)->area_changes,
					CFG_OPT_E))) {

				ALL_NGB_TERMINATE(*ospf_intf);
				(*ospf_intf)->ospf_if_changes = 0;
				ospf_intf = &(*ospf_intf)->ospf_if_next;
				ASEXTREE_BUG("ospf_reconfig 4");
				continue;
			}

			/* Reset all changes
			 */
			(*ospf_intf)->ospf_if_changes = 0;
 
                        /* For NBMA interfaces only.Everything else, nothing to 			 * do.
                         */
                        if (BIT_TEST((*ospf_intf)->ospf_if_flags,
                                IFS_MULTICAST|IFS_POINTOPOINT|IFS_OSPFVLINK)) {
				ospf_intf = &(*ospf_intf)->ospf_if_next;
                                continue;
			}
 
			for(ngb = &(*ospf_intf)->ospf_if_ngb; *ngb;) {
				if (!BIT_TEST((*ngb)->ngb_changes, CFG_KEEP)) {
					lostngb = *ngb;
					*ngb = (*ngb)->ngb_next;
					ospf_ngb_terminate(lostngb);
					continue;
				}
				(*ngb)->ngb_changes = 0;
				ngb = &(*ngb)->ngb_next;
			}
			ospf_intf = &(*ospf_intf)->ospf_if_next;
			ASEXTREE_BUG("ospf_reconfig 5");
 
                }
		ASEXTREE_BUG("ospf_reconfig 6");

		/* Reset all changes
		 */
		(*area)->area_changes = 0;
		area = &(*area)->area_next;
 
        }

	ASEXTREE_BUG("ospf_reconfig 7");
	rt_close(ospf_task, NULL, 0, "OSPF Interface Change");
}

/* Remove (and free) and area from the list.
 */
void
ospf_area_remove __PF1(area, ospf_area_t *)
{
	ospf_intf_t *ospf_intf, *next;
	ospf_ptwalk_t *ptw;

	trace_tf(ospf_trace_options, TR_OSPF_STATE, 0,
               ("OSPF STATE: AREA %A REMOVED", sockbuild_in(0, area->area_id)));

	ASEXTREE_BUG("ospf_area_remove 1");

	/* Remove all interfaces
	 */
	for(ospf_intf = area->area_intf; ospf_intf; ospf_intf = next) {
		next = ospf_intf->ospf_if_next;
		ospf_intf_remove(ospf_intf);
	}

	ASEXTREE_BUG("ospf_area_remove 2");

	/* Cancel all ospf_ptwalk_t structures for thst area.
	 */
	for(ptw = area->area_ptwalk; ptw; ptw = ptw->tw_next) {
		ptw->tw_node = NULL;
		ptw->tw_area = NULL;
	}

	/* Destroy the whole database.
	 */
	db_tree_kill(&area->area_tree, TK_REMOVE);

	/* Destroy the list of gateways.
	 */
	gw_freelist(area->area_gwlist);

	/* If there is a current refresh on that area, terminate the job.
	 */
	if (ospf_refresh_job && ospf_refresh_area == area) {
		(void)ospf_pt_walkdel(ospf_refresh_walk);
		task_job_delete(ospf_refresh_job);
		ospf_refresh_job = NULL;
	}

	task_block_free(ospf_area_blocks, area);
}

/* Remove an interface.
 */
void
ospf_intf_remove __PF1(ospf_intf, ospf_intf_t *)
{
	ospf_ngb_t *ngb, *next;

	trace_tf(ospf_trace_options, TR_OSPF_STATE, 0,
                ("OSPF STATE: INTERFACE %A REMOVED",
			sockbuild_in(0, ospf_intf->ospf_if_addr)));

	/* Shut down the interface.
	 */
	ospf_if_down(ospf_intf);

	/* And delete it.
	 */
	task_block_free(ospf_if_blocks, ospf_intf);
}

/* Remove a neighbour.
 */
void
ospf_ngb_terminate __PF1(ngb, ospf_ngb_t *)
{
	/* I really like to send something to the ngb to tell it i am going
	 * down, but there is not such a thing in the RFC :-)
	 */

	trace_tf(ospf_trace_options, TR_OSPF_STATE, 0,
                ("OSPF STATE: NGB %A TERMINATED",
                        sockbuild_in(0, ngb->ngb_rtid)));

	/* cleanup the list.
	 */
	ospf_ngb_cleanup(ngb);

	/* Remove the timers
	 */
	if (ngb->ngb_inactimer)
		task_timer_delete(ngb->ngb_inactimer);
	if (ngb->ngb_rmxTimer)
                task_timer_delete(ngb->ngb_rmxTimer);
        if (ngb->ngb_ackTimer)
                task_timer_delete(ngb->ngb_ackTimer);

	/* Remove the ngb from the flush list.
	 */
	NGB_FLUSH_DEL(ngb);

	task_block_free(ospf_ngb_blocks, ngb);
}

/* Reset all global variable to default value.
 */
void
ospf_reset_default __PF0(void)
{
	ospf_spfInt = OSPF_SPF_INT;
	ospf_lsinfinity = OSPF_LSINFINITY;
	ospf_if_cost = OSPF_IF_COST;
	ospf_preference = RTPREF_OSPF;
	ospf_lsRefreshTime = OSPF_LSREFRESH_TIME;
	ospf_maxAge = OSPF_MAXAGE;
	ospf_maxAgeDiff = OSPF_MAXAGEDIFF;
	ospf_minLSInt = OSPF_MINLSINT;
	ospf_refreshChunkSize = OSPF_REFRESH_CHUNK;
	ospf_nbAreas = 0;
}

/* Debug.
 */
int
ospf_test_intf()
{
	ASEXTREE_BUG("ospf_test_intf");
	return(0);
}
