/*
 * Copyright (c) 1996 The Regents of the University of Michigan
 * All Rights Reserved
 * 
 * License to use, copy, modify, and distribute this software and its
 * documentation can be obtained from Merit at the University of Michigan.
 * 
 * Merit GateDaemon Project
 * 4251 Plymouth Road, Suite C
 * Ann Arbor, MI 48105
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE REGENTS OF THE
 * UNIVERSITY OF MICHIGAN AND MERIT DO NOT WARRANT THAT THE
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the
 * University of Michigan and Merit shall not be liable for
 * any special, indirect, incidental or consequential damages with respect
 * to any claim by Licensee or any third party arising from use of the
 * software. GateDaemon was originated and developed through release 3.0
 * by Cornell University and its collaborators.
 * 
 * Please forward bug fixes, enhancements and questions to the
 * gated mailing list: gated-people@gated.merit.edu.
 * 
 * This copyright has ben automaticly added by the util/addcopyright.pl program.
 * __END_OF_COPYRIGHT__
 */

/* Local debug flags. Allow you to print various stuff that cannot be
 * logged on a normal run.
 */

/* Check the patricia tree when add/change/delete, etc... (very time consuming)
 * If OSPF_CHECK_DB is set the tree is check, not logged. If OSPF_DEBUG_TREE is
 * also defined, the full tree is always logged (dumped).
 */
/*#define OSPF_DEBUG_TREE	1 	*/
/*#define OSPF_DEBUG_FIND 	1	*/
#define OSPF_CHECK_DB

#ifdef OSPF_CHECK_DB
#ifdef OSPF_DEBUG_TREE
#define OSPF_TREE_CHECK(tree)	ospf_check_tree(tree, 1)
#else 	/* OSPF_DEBUG_TREE */
#define OSPF_TREE_CHECK(tree)   ospf_check_tree(tree, 0)
#endif	/* OSPF_DEBUG_TREE */
#else
#define OSPF_TREE_CHECK(tree)
#endif

/* This macro is called in a lot of functions, at the begining and return
 * points. Usefull is something break during the OSPF processing, but we don't
 * know where. All kind of crap can be tested.
 */
#define ASEXTREE_BUG(msg) do { \
	ospf_area_t *__area; \
	if_addr        *__ifap = NULL; \
/*	trace_only_tf(ospf_trace_options,0,("OSPF:ASEXTREE_BUG: %s",(msg)));*/ \
	OSPF_ALL_AREA(__area) { \
		ASSERT((int)(__area->area_tree.pt_root[LST_ASEX]) != 1); \
	} \
	IF_ADDR(__ifap) { \
		ASSERT(socksize(__ifap->ifa_addr) <= 20); \
	} IF_ADDR_END(__ifap); \
} while(0)

/* Just to keep an eye on the number of neighbour in the full state.
 */
#define DEBUG_NGBREADY(str) \
        trace_tf(ospf_trace_options, TR_OSPF_STATE, 0, \
		("OSPF STATE: Number of ready neighbours: %s: Set to %d", \
                (str), ospf_nbNgbNotReady))

/* IP addresses comparation. Since IP addresses are stored in network byte
 * order (BIG_ENDIAN), comparations are byte order dependent.
 */
#define RTID_SSUP(a, b)		(ntohl((a)) > ntohl((b)))
#define RTID_SINF(a, b)         (ntohl((a)) < ntohl((b)))

/* Used by tree_delete_rec() and ospf_ls_delete()
 */
#define TK_NOREF	0
#define TK_REF		1
#define TK_REMOVE	2
#define LSDEL_NORM	0
#define LSDEL_KILL	1

/* Some OSPF definitions.
 */
typedef u_int32 areaid_t;
typedef u_int8	rtpri_t;
typedef u_int32	rtid_t;
typedef u_int16 cost_t;
typedef u_int32 seq_t;
typedef int 	event_t;
typedef u_int8	options_t;
typedef u_int16 auth_t;
typedef u_int32	ipaddr_t;

typedef struct _vtxList_t {
	struct _vtxList_t *l_next;
	struct _vertex_t *l_vtx;
} vtxList_t;

/* The key used for database lookup.
 */
typedef struct _lskey_t {
	u_int32		lsk_id;
	u_int32		lsk_advrt;
} lskey_t;

/* An OSPF packet header. Generic.
 */
typedef struct _ospf_pck_header_t {
        u_int8  pck_hdr_version;        /* Version */
        u_int8  pck_hdr_type;           /* Type of packet. C.f. OSPF_PCK_XXX */
        u_int16 pck_hdr_len;            /* Total lengh of packet */
        u_int32 pck_hdr_rtid;           /* The router ID */
        u_int32 pck_hdr_area;           /* The area */
        u_int16 pck_hdr_checksum;       /* Checksum */
        u_int16 pck_hdr_authtype;       /* Authentication type: OSPF_AUTH_XXX */
        u_int32 pck_hdr_authdata[2];    /* Authentication Data */
} ospf_pck_header_t;
 
/* OSPF LSA packets formats
 * First the header.
 */
typedef struct _ospf_lsa_hdr_t {
        u_int16         lsh_age;
        u_int8          lsh_options;
        u_int8          lsh_type;               /* C.f LST_XXX */
	lskey_t		lsh_key;
        u_int32         lsh_seq;
        u_int16         lsh_cksum;
        u_int16         lsh_len;
} ospf_lsa_hdr_t;

#define lsh_id		lsh_key.lsk_id
#define lsh_advrt	lsh_key.lsk_advrt

typedef struct _ospf_tosmetric_t {
        u_int8          sm_tos;
        u_int8          sm_null;
        cost_t          sm_metric;
} ospf_tosmetric_t;

/* The router LS. first one link, then the whole router.
 */
typedef struct _ospf_lsa_rtlink_t {
        u_int32         lsrl_id;
        u_int32         lsrl_data;
        u_int8          lsrl_type;              /* C.f. LSRLT_XXX */
        u_int8          lsrl_nbtos;
        u_int16         lsrl_metric;
        ospf_tosmetric_t lsrl_tosmetric[1];
} ospf_lsa_rtlink_t;

typedef struct _ospf_lsa_rtr_t {
        ospf_lsa_hdr_t  lsrtr_hdr;
        u_int8          lsrtr_bits;
	u_int8		lsrtr_null;
        u_int16         lsrtr_nblinks;
        ospf_lsa_rtlink_t lsrtr_links[1];
} ospf_lsa_rtr_t;

#define LSRLT_P2P       	1
#define LSRLT_TRANS     	2
#define LSRLT_STUB      	3
#define LSRLT_VLINK    	 	4
#define LSRLT_RTR       	5       /* Fake link type, from net to rtr */

#define LSRTR_BITS_B		1
#define LSRTR_BITS_E		2
#define LSRTR_BITS_V		4
#define LSRTR_BITS_UNSET	0xff

/* The network LS.
 */
typedef struct _ospf_lsa_ntw_t {
        ospf_lsa_hdr_t  lsntw_hdr;
        u_int32         lsntw_mask;
        u_int32         lsntw_rtid[1];
} ospf_lsa_ntw_t;

/* Links Summary LSA
 */
typedef struct _ospf_lsa_sla_t {
	ospf_lsa_hdr_t	lssla_hdr;
	u_int32		lssla_mask;
	union {
		struct {
			u_int32		_lssla_tos: 8;
			u_int32		_lssla_metric: 24;
		} lssla_u_det;
		u_int32 lssla_u_tosmet32;
	} _lssla_u;
} ospf_lsa_sla_t;

/*
#define lssla_tos	_lssla_u.lssla_u_det._lssla_tos
#define lssla_metric	_lssla_u.lssla_u_det._lssla_metric
*/
#define lssla_tosmet32	_lssla_u.lssla_u_tosmet32

#define SLAMETRIC_TOS(tosmet)		(((tosmet) >> 24) & 0xff)
#define SLAMETRIC_METRIC(tosmet)	((tosmet) & 0xffffff)

/* External Link advertisement.
 */
typedef struct _ospf_lsext_tosmet_t {
	union {
		struct {
			u_int32	tosmet_bitE: 1;
			u_int32 tosmet_tos: 7;
			u_int32 tosmet_metric: 24;
		} tosmet_u_det;
		u_int32 tosmet_u_int32;
	} tosmet_u;
	u_int32	tosmet_forw;
	u_int32	tosmet_tag;
} ospf_lsext_tosmet_t;

typedef struct _ospf_lsa_ext_t {
	ospf_lsa_hdr_t 	lsext_hdr;
	u_int32		lsext_mask;
	ospf_lsext_tosmet_t lsext_tosmet[1];
} ospf_lsa_ext_t;

/*
#define lsext_bitE	lsext_tosmet[0].tosmet_u.tosmet_u_det.tosmet_bitE
#define lsext_tos	lsext_tosmet[0].tosmet_u.tosmet_u_det.tosmet_tos
#define lsext_metric	lsext_tosmet[0].tosmet_u.tosmet_u_det.tosmet_metric
*/
#define lsext_tosmet32	lsext_tosmet[0].tosmet_u.tosmet_u_int32
#define	lsext_forw	lsext_tosmet[0].tosmet_forw
#define lsext_tag	lsext_tosmet[0].tosmet_tag

#define EXTMETRIC_TOS(tosmet)		(((tosmet) >> 24) & 0x7f)
#define EXTMETRIC_TYPE(tosmet)		((((tosmet) >> 31) & 0x1) + 1)
#define EXTMETRIC_METRIC(tosmet)	((tosmet) & 0xffffff)

/* The biggest/smaller MTU we expect to see. Bigger MTU will result in
 * packets not completly filled up, smallest ones will result in IP
 * fragmentation.
 */
#define MTU_BIGGEST	4000
#define MTU_SMALLEST	256

/* given a MTU, compute the size of an OSFP packet, and different number of
 * LS which can fit in one packet.
 */
#define MTU2PACKET(mtu)		((mtu) - 2 * sizeof(struct ip))
#define MTU2LSA(mtu)		(MTU2PACKET(mtu) - sizeof(ospf_pck_header_t))
#define MTU2DD(mtu) \
	((MTU2PACKET(mtu) - sizeof(ospf_dd_pck_t)) / sizeof(ospf_lsa_hdr_t))
#define MTU2LSR(mtu) \
	((MTU2PACKET(mtu) - sizeof(ospf_pck_header_t)) /sizeof(ospf_lsr_data_t))
#define MTU2ACK(mtu) \
	((MTU2PACKET(mtu) - sizeof(ospf_pck_header_t)) / sizeof(ospf_lsa_hdr_t))

#define OSPF_PCK_MAXLEN		MTU2PACKET(MTU_BIGGEST)	/* The bigest MTU */
#define MAXLSALENGHT    	MTU2LSA(MTU_SMALLEST)	/* Why not? */

/* A generic LSA
 */
typedef struct _ospf_lsa_t {
        union {
                ospf_lsa_hdr_t  lu_dhr;
                ospf_lsa_rtr_t  lu_rt;
                ospf_lsa_ntw_t  lu_nt;
		ospf_lsa_sla_t	lu_sla;
		ospf_lsa_ext_t	lu_ext;
                u_int32         lu_buffer[(MAXLSALENGHT-sizeof(ospf_lsa_hdr_t))
                                         / sizeof(u_int32)];
        }               lsa_data;
} ospf_lsa_t;

#define lsa_hdr		lsa_data.lu_dhr
#define lsa_age         lsa_data.lu_dhr.lsh_age
#define lsa_type        lsa_data.lu_dhr.lsh_type
#define lsa_key		lsa_data.lu_dhr.lsh_key
#define lsa_id          lsa_data.lu_dhr.lsh_key.lsk_id
#define lsa_advrt       lsa_data.lu_dhr.lsh_key.lsk_advrt
#define lsa_rt          lsa_data.lu_rt
#define lsa_nt          lsa_data.lu_nt
#define lsa_options     lsa_data.lu_dhr.lsh_options
#define lsa_len         lsa_data.lu_dhr.lsh_len
#define lsa_seq         lsa_data.lu_dhr.lsh_seq
#define lsa_cksum       lsa_data.lu_dhr.lsh_cksum
#define lsa_rtr_nblinks lsa_data.lu_rt.lsrtr_nblinks
#define lsa_rtr_bits    lsa_data.lu_rt.lsrtr_bits
#define lsa_rtr_links   lsa_data.lu_rt.lsrtr_links
#define lsa_ntw_rtid    lsa_data.lu_nt.lsntw_rtid
#define lsa_ntw_mask    lsa_data.lu_nt.lsntw_mask
#define lsa_sla_mask	lsa_data.lu_sla.lssla_mask
#define lsa_sla_metric	lsa_data.lu_sla.lssla_metric
#define lsa_sla_tos	lsa_data.lu_sla.lssla_tos
#define lsa_sla_tosmet32 lsa_data.lu_sla.lssla_tosmet32
#define lsa_ext_bitE	lsa_data.lu_ext.lsext_bitE
#define lsa_ext_mask	lsa_data.lu_ext.lsext_mask
#define lsa_ext_metric	lsa_data.lu_ext.lsext_metric
#define lsa_ext_tosmet32 lsa_data.lu_ext.lsext_tosmet32
#define lsa_ext_forw	lsa_data.lu_ext.lsext_forw
#define lsa_ext_tag	lsa_data.lu_ext.lsext_tag
#define lsa_ext_tos	lsa_data.lu_ext.lsext_tos

#define HOSTMASK        0xffffffff
#define LSA_MASK(lsa)   ((lsa->lsa_type == LST_NTW)? \
                                (lsa)->lsa_data.lu_nt.lsntw_mask : \
                                HOSTMASK)

#define LST_NULL        0               /* For empty tree nodes */
#define LST_RTR         1
#define LST_NTW         2
#define LST_ASIP        3
#define LST_ASBR        4
#define LST_ASEX        5
#define LST_MAXTYPE     6
#define LS_BOUND(type)  ((type) > LST_NULL && (type) < LST_MAXTYPE)

/* The tree root, with associated parameters.
 */
typedef struct _db_tree_t {
        union _tree_node_t *pt_root[LST_MAXTYPE];/* The root of the trees */
        int             pt_null_nodes;          /* Number of null nodes */
        int             pt_used_nodes;          /* number of used nodes */
	struct _ospf_area_t *pt_area;		/* Associated area */
} db_tree_t;

/* The OSPF graph.
 */
typedef struct _ospf_graph_t {
        struct _vertex_t *gr_root;               /* The root of the graph. */
        int             gr_nbVtx[LST_MAXTYPE];  /* Number of vertices, by type*/
} ospf_graph_t;

/* Interface states. Cf RFC 1583, p35 (+ some personal additions)
 */
#define OSPF_IFS_UNDEF		0	/* Not yet defined yet */
#define OSPF_IFS_NOOSPF		1	/* No OSPF run on that interface */
#define OSPF_IFS_DOWN		2	/* If is down. Dead. */
#define OSPF_IFS_LOOPBACK	3	/* Loopback. Not used yet. */
#define OSPF_IFS_WAITING	4	/* Waiting to find the DR */
#define OSPF_IFS_P2P		5	/* P2P interface. Not used yet */
#define OSPF_IFS_DROTHER	6	/* We are not the DR! */
#define OSPF_IFS_BACKUPDR	7	/* We are the backup DR */
#define OSPF_IFS_DR		8	/* We are the DR! */

/* Kind of events that happen to the interface. RFC p 36.
 */
#define EV_IF_UP		0	/* We are up now! */
#define EV_IF_WAITTIMER		1	/* Wait timer rang! */
#define EV_IF_BCKUPSEEN		2	/* New backup seen */
#define EV_IF_NGBCHG		3	/* Some change in the neighbour */
#define EV_IF_LOOPON		4	/* Not used yet */
#define EV_IF_LOOPOFF		5	/* Not used yet */
#define EV_IF_DOWN		6	/* We are dead now! */

/* States of a neighbour. RFC 1583 p44
 */
#define OSPF_NGBS_DOWN		0	/* it si dead! */
#define OSPF_NGBS_ATTEMPT	1	/* Almost dead! */
#define OSPF_NGBS_INIT		2	/* We got something from that guy... */
#define OSPF_NGBS_2WAYS		3	/* ...and he got something from us */
#define OSPF_NGBS_EXSTART	4	/* We start creating an adjacency */
#define OSPF_NGBS_EXCHANGE	5	/* We are exchanging databases */
#define OSPF_NGBS_LOADING	6	/* LSRequest exchanged */
#define OSPF_NGBS_FULL		7	/* Routers are adjacent */

/* Kind of events that happen to the neighbours. RFC p 44
 */
#define EV_NBG_HELLORECV	0	/* Hello received */
#define EV_NBG_START		1	/* Send hello packets */
#define EV_NBG_2WAYS		2	/* a 2 ways communication is up */
#define EV_NBG_NEGDONE		3	/* Master/Slave negotiation done */
#define EV_NBG_EXCHDONE		4	/* Exchange of DB done */
#define EV_NBG_BADLSREQ		5	/* Bad LS Request */
#define EV_NBG_LOADDONE		6	/* DB loading done */
#define EV_NBG_ADJOK		7	/* Reconsidere adjacency */
#define EV_NBG_SEQNUMIS		8	/* */
#define EV_NBG_1WAY		9	/* Back to one way communication */
#define EV_NBG_KILLNBR		10	/* Go back to DOWN state! */
#define EV_NBG_INACTIMER	11	/* No hello packet. Back to DOWN */
#define EV_NBG_LLDOWN		12	/* Link is down */

/* OSPF packet. Any kind.
 */
typedef struct _ospf_pck_t {
	ospf_pck_header_t pck_header;	/* OSPF header */
	u_int32	pck_data[OSPF_PCK_MAXLEN / sizeof(u_int32) -
		sizeof(ospf_pck_header_t)];
} ospf_pck_t;
#define pck_version 	pck_header.pck_hdr_version
#define pck_type	pck_header.pck_hdr_type
#define pck_len		pck_header.pck_hdr_len
#define pck_rtid	pck_header.pck_hdr_rtid
#define pck_area	pck_header.pck_hdr_area
#define pck_checksum	pck_header.pck_hdr_checksum
#define pck_authtype	pck_header.pck_hdr_authtype
#define pck_authdata	pck_header.pck_hdr_authdata

/* OSPF packet types
 */
#define OSPF_PCK_HELLO		1
#define OSPF_PCK_DD		2
#define OSPF_PCK_LSR		3
#define OSPF_PCK_LSU		4
#define OSPF_PCK_LSA		5

/* Used by ospf_send_addr() to find out where to send a packet.
 */
#define SEND_DD			1	/* Send a DD packet */
#define SEND_LSR		2	/* Send a LSR packet */
#define SEND_HELLO		3	/* Send a HELLO packet */
#define SEND_LSU_DIRECT		4	/* Send a LSU pckt after LSR received */
#define SEND_LSU_FLOODING	5	/* Send a LSU packet when flooding */
#define SEND_LSU_RMX		6	/* Re-Send a LSU packet */
#define SEND_LSA_DELAYED	7	/* Send a delayed ACK packet */
#define SEND_LSA_DIRECT		8	/* Send a direct ACK packet */

/* OSPF authentication types
 */
#define OSPF_AUTH_NONE		0
#define OSPF_AUTH_SIMPLE	1

/* OSPF version, default value for a bunch of parameters.
 */
#define OSPF_VERSION		2			/* */
#define OSPF_HELLO_INT		10			/* seconds */
#define OSPF_RTDEAD_INT		(4*OSPF_HELLO_INT)	/* seconds */
#define OSPF_TRANS_DELAY	1			/* seconds */
#define OSPF_NULL_DR		0			/* No DR yet */
#define OSPF_IF_COST		1
#define OSPF_IF_COST_MAX	65535
#define OSPF_IF_RMX		5			/* seconds */
#define OSPF_RTPRI		1
#define OSPF_IF_AUTH		OSPF_AUTH_NONE
#define OSPF_IF_AUTH2		OSPF_AUTH_NONE
#define OSPF_LSINFINITY		0xffffff
#define OSPF_SPF_INT		5			/* seconds */
#define OSPF_LSREFRESH_TIME	60			/* seconds XXX */
#define OSPF_REFRESH_CHUNK	100
#define OSPF_MAX_TIMER_NAME	80

/* OSPF Trace flags
 */
#define	TR_OSPF_STATE		TR_USER_1	/* Intf, ngb, area changes */
#define TR_OSPF_DRELECT		TR_USER_2
#define TR_OSPF_DB		TR_USER_3
#define TR_OSPF_SPF		TR_USER_4
#define TR_OSPF_FLOOD		TR_USER_5
#define TR_OSPF_INDEX_ALL	0
#define TR_OSPF_INDEX_HELLO	1
#define TR_OSPF_INDEX_DD	2
#define TR_OSPF_INDEX_LSR	3
#define TR_OSPF_INDEX_LSU	4
#define TR_OSPF_INDEX_LSA	5

/* This is kind of redundant with the previous one, but i do not know how
 * gated work and this trace_parse_packet() function is very confusing!
 */
#define TR_OSPF_PCK_ALL		TR_PACKET
#define TR_OSPF_PCK_HELLO	TR_PACKET_1
#define TR_OSPF_PCK_DD		TR_PACKET_2
#define TR_OSPF_PCK_LSR		TR_PACKET_3
#define TR_OSPF_PCK_LSU		TR_PACKET_4
#define TR_OSPF_PCK_LSA		TR_PACKET_5

/* An OSPF Hello packet.
 */
typedef struct _ospf_hello_pck_t {
	ospf_pck_header_t pck_hello_hdr;	/* normal header */
	u_int32		pck_hello_mask;		/* Newtwork mask */
	u_int16		pck_hello_int;		/* Hello interval */
	u_int8		pck_hello_options;	/* Options */
	u_int8		pck_hello_rtpri;	/* Router priority */
	u_int32		pck_hello_rtdead;	/* Rourter dead interval */
	u_int32		pck_hello_dr;		/* Router's idea of the DR */
	u_int32		pck_hello_backupDr;	/* Same for the backup */
	u_int32		pck_hello_ngb[1];	/* All neighbours */
} ospf_hello_pck_t;

#define pck_hello_rtid	pck_hello_hdr.pck_hdr_rtid
#define pck_hello_len	pck_hello_hdr.pck_hdr_len

/* An OSPF Database Description packet (DD)
 */
typedef struct _ospf_dd_pck_t {
	ospf_pck_header_t pck_dd_hdr;		/* normal header */
	u_int16		pck_dd_junk;		/* Must be 0 */
	u_int8		pck_dd_options;		/* Router capability options */
	u_int8		pck_dd_flags;		/* Master/slave/init/more... */
	u_int32		pck_dd_seq;		/* Sequence number */
} ospf_dd_pck_t;

#define pck_dd_len	pck_dd_hdr.pck_hdr_len
#define pck_dd_rtid	pck_dd_hdr.pck_hdr_rtid

/* An OSPF Link State Request packet.
 */
typedef struct _ospf_lsr_data_t {
	u_int32		lsrd_type;
	lskey_t		lsrd_key;
} ospf_lsr_data_t;

#define lsrd_id		lsrd_key.lsk_id
#define lsrd_advrt	lsrd_key.lsk_advrt

typedef struct _ospf_lsr_pck_t {
	ospf_pck_header_t pck_lsr_hdr;		/* normal header */
	ospf_lsr_data_t	pck_lsr_data[1];	/* The request */
} ospf_lsr_pck_t;

#define pck_lsr_rtid	pck_lsr_hdr.pck_hdr_rtid
#define pck_lsr_len	pck_lsr_hdr.pck_hdr_len

typedef struct _ospf_lsu_pck_t {
	ospf_pck_header_t pck_lsu_hdr;          /* normal header */
	u_int32		pck_lsu_nbadv;		/* # of LSA */
	ospf_lsa_t	pck_lsu_lsa[1];
} ospf_lsu_pck_t;

#define pck_lsu_rtid	pck_lsu_hdr.pck_hdr_rtid
#define pck_lsu_len	pck_lsu_hdr.pck_hdr_len

/* OSPF LS acknowledgement packet.
 */
typedef struct _ospf_lsa_pck_t {
	ospf_pck_header_t pck_lsa_hdr;		/* normal header */
	ospf_lsa_hdr_t	 pck_lsa_lsh[1];
} ospf_lsa_pck_t;

#define pck_lsa_rtid	pck_lsa_hdr.pck_hdr_rtid
#define pck_lsa_len     pck_lsa_hdr.pck_hdr_len

#define PCK_LSA_NBLSH(len)	\
	(((int)(len) - sizeof(ospf_lsa_pck_t)) / sizeof(ospf_lsa_hdr_t) + 1)

/* OSPF Options bits. 
 */
#define OSPF_PCK_OPT_T		0x001
#define OSPF_PCK_OPT_E		0x002

/* DD packets flags. Used in ngb_options.
 */
#define OSPF_DD_FLAGS_MS	0x1
#define OSPF_DD_FLAGS_M		0x2
#define OSPF_DD_FLAGS_I		0x4
#define OSPF_DD_FLAGS_MASTER  (OSPF_DD_FLAGS_MS|OSPF_DD_FLAGS_M|OSPF_DD_FLAGS_I)
#define OSPF_DD_FLAGS_SLAVE	(OSPF_DD_FLAGS_M)
#define DD_MASTER(ngb)		BIT_TEST(ngb->ngb_options, OSPF_DD_FLAGS_MS)
#define DD_SLAVE(ngb)		(!DD_MASTER(ngb))

/* OSPF error codes. C.f. table ospf_errors[] for explaination.
 */
#define OSPF_NOERR		0
#define OSPF_ERR_LENGHT		1
#define OSPF_ERR_LENGHT2	2
#define OSPF_ERR_VERSION	3
#define OSPF_ERR_CKSUM		4
#define OSPF_ERR_AUTHTYPE	5
#define OSPF_ERR_NOMIF		6
#define OSPF_ERR_NOIF		7
#define OSPF_ERR_INVIF		8
#define OSPF_ERR_HMASK		9
#define OSPF_ERR_HINT		10
#define OSPF_ERR_HRTDEAD	11
#define OSPF_ERR_INVNGB		12
#define OSPF_ERR_INVOPTIONS	13
#define OSPF_ERR_DDMSMIS	14
#define OSPF_ERR_DDIMIS		15
#define OSPF_ERR_DDSEQMIS	16
#define OSPF_ERR_INVLSTYPE	17
#define OSPF_ERR_BADLSR		18
#define OSPF_ERR_NOVIF		19
#define OSPF_ERR_BADAREA	20
#define OSPF_ERR_DDCONF		21

/* Used when (re)configuring an area/neighbor/interface.
 * C.f. ospf_conf_setintf()
 */
#define	CFG_ENABLE	0x1
#define	CFG_DISABLE	0x2
#define	CFG_RXMINT	0x4
#define	CFG_TRANSDEL	0x8
#define	CFG_RTPRI	0x10
#define	CFG_HELLOINT	0x20
#define	CFG_RTDEAD	0x40
#define	CFG_NOMULTI	0x80
#define	CFG_AUTH	0x100
#define	CFG_AUTH2	0x200
#define	CFG_POLINT	0x400
#define	CFG_RTLIST	0x800
#define	CFG_COST	0x1000
#define CFG_RTELIG	0x2000
#define CFG_KEEP	0x4000
#define CFG_RTRBITS	0x8000
#define CFG_OPT_E	0x10000
#define CFG_OPT_T	0x20000
#define CFG_OLD		0x40000
#define CFG_AREA	0x80000

/* Interface data structure. RFC p33. Some field are already in the
 * _if_addr structures.
 */
typedef struct _ospf_intf_t {
	struct _ospf_intf_t *ospf_if_next;/* Stored in list */
	struct _ospf_area_t *ospf_if_area;/* Area of the interface */
	int ospf_if_state;		/* State of the interface */
	ipaddr_t ospf_if_addr;		/* Interface IP address */
	flag_t  ospf_if_flags;		/* Same as ifa_state */
	time_t	ospf_if_helloInt;	/* Interval for Hello packets */
	time_t	ospf_if_rtdeadInt;	/* Down router time out */
	time_t 	ospf_if_transDel;	/* Delay on the intf */
	time_t 	ospf_if_rmxInt;		/* LSA retransmit interval */
	time_t  ospf_if_pollInt;	/* NBMA network polling time */
	rtpri_t	ospf_if_rtpri;		/* Router priority */
	task_timer *ospf_if_helloTimer;	/* Hello timer */
	task_timer *ospf_if_waitTimer; 	/* Wait timer */
	struct _ospf_ngb_t *ospf_if_ngb;/* List of neighbours */
	rtid_t ospf_if_drID;		/* The ID of the DR */
	rtid_t ospf_if_backupDrID;	/* The ID of the backup DR */
	ipaddr_t ospf_if_drAddr;	/* The IP address of the DR */
	ipaddr_t ospf_if_backupDrAddr;	/* The IP address of the backup DR */
	cost_t	ospf_if_cost;		/* Cost of the intf */
	auth_t	ospf_if_authType;	/* authentification type */
	auth_t	ospf_if_authType2;	/* Second authentification type */
	struct _vertex_t *ospf_if_ntls;	/* The LS for the network, if DR */
	if_addr *ospf_if_ifap;		/* Associated physical interface */
	int ospf_if_maxLSRinPacket;	/* Max LSR in one packet */
	int ospf_if_maxDDinPacket;	/* Max DD in one packet */
	int ospf_if_maxACKinPacket;	/* Max ACK in one packet */
	flag_t ospf_if_changes;		/* changes when reconfiguring */
} ospf_intf_t;

#define ospf_if_options		ospf_if_area->area_options
#define ospf_if_tree		ospf_if_area->area_tree
#define ospf_if_mtu		ospf_if_ifap->ifa_mtu
#define ospf_if_sockaddr	ospf_if_ifap->ifa_info.ifi_addr
#define ospf_if_mask	ospf_if_ifap->ifa_info.ifi_netmask->in.gin_addr.s_addr
#define ospf_if_socklocaddr	ospf_if_ifap->ifa_info.ifi_addr_local
#define ospf_if_link		ospf_if_ifap->ifa_link
#define ospf_if_locaddr	\
	ospf_if_ifap->ifa_info.ifi_addr_local->in.gin_addr.s_addr
#define	ifa_ospf_state		ifa_ospf->ospf_if_state
#define	ifa_ospf_area		ifa_ospf->ospf_if_area
#define ifa_ospf_ifap		ifa_ospf->ospf_if_ifap

#define OSPF_ISVIRTUAL(ospf_intf) \
	BIT_TEST((ospf_intf)->ospf_if_flags, IFS_OSPFVLINK)

#define OSPF_IFMASK(ospf_intf)	((ospf_intf)->ospf_if_ifap? \
		(ospf_intf)->ospf_if_mask: RTMASK)
#define OSPF_INTF_NGB(ngb, ospf_intf) \
	for((ngb) = (ospf_intf)->ospf_if_ngb; (ngb); (ngb) = (ngb)->ngb_next)

#define OSPF_IF_ADDR_AREA(area, oif) \
	for((oif) = (area)->area_intf; (oif); (oif) = (oif)->ospf_if_next)

#define OSPF_IF_ADDR_AREA_END(area, oif)

#define OSPF_IF_ADDR(oif) { \
	ospf_area_t *__area; \
	for(__area = ospf_area_list; __area; __area = __area->area_next) { \
		OSPF_IF_ADDR_AREA(__area, oif)

#define OSPF_IF_ADDR_END(oif) \
		OSPF_IF_ADDR_AREA_END(__area, oif); \
	} \
}

/* Max value for the number of DD, LSR, ACK in the biggest packet (bigest MTU).
 */
#define NGB_MAXDDSIZE		MTU2DD(MTU_BIGGEST)
#define NGB_MAXACKSIZE		MTU2ACK(MTU_BIGGEST)
#define NGB_MAXLSRSIZE		MTU2LSR(MTU_BIGGEST)

/* Neighbour data structure
 */
typedef struct _ospf_ngb_t {
	struct _ospf_ngb_t *ngb_next; /* Interface's list of neighbours */
	struct _ospf_ngb_t *ngb_flushnext; /* List of ngb to the flushed */
	struct _ospf_ngb_t *ngb_flushprev; /* List of ngb to the flushed */
	u_int	ngb_state;	/* Its state. C.f. OSPF_NBGS_XXX */
	task_timer *ngb_inactimer;/* Inactivity timer */
	task_timer *ngb_rmxTimer;/* DD transmission timer */
	task_timer *ngb_ackTimer; /* Delayed ACK timer */
	options_t ngb_options;	/* Options for the neighbour. C.f. OSPF_DD_XXX*/
	seq_t	ngb_ddseq;	/* DD sequence number */
	rtid_t	ngb_rtid;	/* ngb's router ID */
	rtpri_t	ngb_rtpri;	/* ngb's router priority */
	ipaddr_t ngb_addr;	/* ngb's IP address */
	ipaddr_t ngb_dr;	/* ngb's idea of the DR IP address */
	ipaddr_t ngb_backupDr;	/* ngb's idea of the backup DR IP address */
	ospf_intf_t *ngb_intf;	/* Back link to interface */
	time_t ngb_lastHello;	/* Last hello sent (used for NBMA net only */
	flag_t ngb_changes;	/* changes when reconfiguring */

	vtxList_t *ngb_dataSum;	/* Database summary list */
	int ngb_DSlistSize;	/* Number of DS in the list */
	int ngb_lastDDsize;	/* Size of ngb_lastDDvtx */
	struct _vertex_t *ngb_lastDDvtx[NGB_MAXDDSIZE]; /* List of nodes sent
				 * in the last DD packet */

	db_tree_t ngb_lsReq;	/* Link State Request list */
	db_tree_t ngb_LSRbuffer;/* what's in the last packet sent */

	db_tree_t ngb_lsRmx;	/* Link State Retransmission list. */
	db_tree_t ngb_RMXbuffer;/* what's in the last packet sent */
	int ngb_RMXbuffer_len;	/* Size of packet in the buffer */

	struct _vertex_t *ngb_ackList[NGB_MAXACKSIZE]; /* List of pending ACK */
	int ngb_lastAckIndex;	/* Number of ACK in ngb_ackList[] */
	time_t ngb_ackDelay;	/* How long to delay ACK? */
} ospf_ngb_t;

#define ngb_tree		ngb_intf->ospf_if_area->area_tree
#define ngb_area		ngb_intf->ospf_if_area
#define ngb_ifap		ngb_intf->ospf_if_ifap
#define ngb_maxLSRinPacket	ngb_intf->ospf_if_maxLSRinPacket
#define ngb_maxDDinPacket	ngb_intf->ospf_if_maxDDinPacket
#define ngb_maxACKinPacket	ngb_intf->ospf_if_maxACKinPacket

/* Area structure.
 */
typedef struct _ospf_area_t {
	struct _ospf_area_t *area_next;		/* List of areas... */
	struct _vertex_t *area_rtrls;		/* Router announcement */
	areaid_t area_id;			/* The ID */
	flag_t	area_rtrbits;			/* The bits field of the RT LS*/
	db_tree_t area_tree;			/* The area's LS database */
	ospf_graph_t area_graph;		/* The area's LS graph */
	flag_t  area_options;			/* Bit E, what else? */
	gw_entry *area_gwlist;			/* List of border routers */
	u_int area_spfcount;			/* Used for SPF computation */
	auth_t area_auth;			/* Default authentication. */
	ospf_intf_t *area_intf;			/* List of interfaces in area */
	flag_t area_changes;			/* Changes when reconfiguring */
	struct _ospf_ptwalk_t *area_ptwalk;	/* Those walking the tree */
	gw_entry *area_defaultgw;		/* Default gateway associated
						 * with a route */
} ospf_area_t;

#define OSPF_ALL_AREA(area) \
	for((area) = ospf_area_list; (area); (area) = (area)->area_next)
#define OSPF_ALL_AREA_END(area)
#define OSPF_AREAID_BACKBONE	0
#define AREA_DEFAULT_OPT	OSPF_PCK_OPT_E
#define AREA_DEF_RTRBITS	0

/* Some flags used during config
 */
#define FOSPF_ON	1	/* OSPF is ON */

/* Get an OSPF packet buffer. The max size is thew MTU minus ip header size and
 * some extrat space (ip header size).
 */
#define OSPF_BUFFER_NBPCK	1	/* one packet in OSPF buffer */
#define OSPF_PACKET(pck, type, maxlen, ifa) do { \
	(ospf_pck_t *)(pck) = &ospf_pck_buffer[ospf_pck_bffct]; \
	ospf_pck_bffct = (ospf_pck_bffct + 1) % OSPF_BUFFER_NBPCK; \
	(maxlen) = MIN(sizeof(ospf_pck_t), MTU2PACKET((ifa)->ifa_mtu)); \
	(void)bzero((void_t)(pck), (maxlen)); \
	((ospf_pck_t *)(pck))->pck_type = (type); \
	((ospf_pck_t *)(pck))->pck_version = OSPF_VERSION; \
} while(0)

/* Some socket stuff which do not seen to be somewhere else
 */
#define SOCKCMP(a, b)	    ((a)->in.gin_addr.s_addr == (b)->in.gin_addr.s_addr)
#define SOCK32(s)	    ((s)->in.gin_addr.s_addr)
#define ALL_OSPF_ROUTERS	htonl(0xe0000005)  /* 224.0.0.5 */
#define ALL_OSPF_DR_ROUTERS	htonl(0xe0000006)  /* 224.0.0.6 */
#define ALL_OSPF_ELIGIBLE	htonl(0xe0000007)  /* A fake address for NBMA */
#define OSPF_IF_NEWSTATE(ospf_intf, state)  do { \
	if (TRACE_TF(ospf_trace_options, TR_OSPF_STATE)) \
		tracef("OSPF STATE: Interface %A change from <%s> ", \
			sockbuild_in(0, (ospf_intf)->ospf_if_addr), \
			trace_value(ospf_if_states, \
				(ospf_intf)->ospf_if_state)); \
	(ospf_intf)->ospf_if_state = (state); \
	trace_tf(ospf_trace_options, TR_OSPF_STATE, 0, \
		("to <%s>", trace_value(ospf_if_states, \
		 (ospf_intf)->ospf_if_state)));\
} while(0)

#define OSPF_NGB_NEWSTATE(ngb, state) do { \
	if (TRACE_TF(ospf_trace_options, TR_OSPF_STATE)) \
		tracef("OSPF STATE: NGB %A change from <%s> ", \
			sockbuild_in(0, (ngb)->ngb_addr), \
			trace_value(ospf_ngb_states, (ngb)->ngb_state)); \
	(ngb)->ngb_state = (state); \
	trace_tf(ospf_trace_options, TR_OSPF_STATE, 0, \
		("to <%s>", trace_value(ospf_ngb_states, (ngb)->ngb_state))); \
} while(0)

#ifdef IP_MULTICAST
#define OSPF_JOINT_GROUP(ospf_intf, grp)  do { \
	if (task_set_option(ospf_task, TASKOPTION_GROUP_ADD, \
			(ospf_intf)->ospf_if_ifap, (grp)) < 0) { \
		/* If the address is alredy there, we don't care. else \
                 * we disable the interface. \
                 */ \
                if (errno != EADDRINUSE) \
                        ospf_if_down(ospf_intf); \
        } \
} while(0)
#define OSPF_QUIT_GROUP(ospf_intf, grp)  do { \
	(void)task_set_option(ospf_task, TASKOPTION_GROUP_DROP, \
		(ospf_intf)->ospf_if_ifap, (grp)); \
} while(0)
#else	/* IP_MULTICAST */
#define OSPF_JOINT_GROUP(ospf_intf, grp)
#define OSPF_QUIT_GROUP(ospf_intf, grp)
#endif	/* IP_MULTICAST */
#define OSPF_IF_ADDNGB(ospf_intf, ngb) { \
	(ngb)->ngb_next = (ospf_intf)->ospf_if_ngb; \
	(ospf_intf)->ospf_if_ngb = (ngb); \
	(ngb)->ngb_intf = (ospf_intf); \
}
#define NB_HELLO_NGB(len)	(((int)(len) - sizeof(ospf_hello_pck_t) + \
		sizeof(u_int32)) / sizeof(u_int32))

#define PCK_NBLSR(len) 	(((int)len - \
		sizeof(ospf_lsr_pck_t)) / sizeof(ospf_lsr_data_t) + 1)

/* Reset a list. I.e set all pointers/variable to known values.
 * Anything on the list will be deleted.
 */
#define NGB_LIST_RESET_DS(ngb)	{ \
	vtxList_t *vl; \
	while((ngb)->ngb_dataSum) { \
		vl = (ngb)->ngb_dataSum->l_next; \
		task_block_free(ospf_vtxlist_block, (ngb)->ngb_dataSum); \
		(ngb)->ngb_dataSum = vl; \
	} \
	(ngb)->ngb_lastDDsize = 0; \
	(ngb)->ngb_DSlistSize = 0; \
}
#define NGB_LIST_RESET_LSR(ngb) { \
	db_tree_kill(&(ngb)->ngb_lsReq, TK_NOREF); \
	db_tree_kill(&(ngb)->ngb_LSRbuffer, TK_NOREF); \
}
#define NGB_LIST_RESET_RMX(ngb) { \
	db_tree_kill(&(ngb)->ngb_lsRmx, TK_REF); \
	db_tree_kill(&(ngb)->ngb_RMXbuffer, TK_REF); \
}

#define OSPF_NGBLIST_TRACE

/* when there is a LSA to send, the nighbour is put on the ospf_flush_list
 * and at the end of OSPF processing all LSA in the ngb's queues in the list
 * are flushed.
 */
#define NGB_FLUSH_ADD(ngb) do { \
	(ngb)->ngb_flushnext = ospf_flush_list; \
	(ngb)->ngb_flushprev = NULL; \
	ospf_flush_list = (ngb); \
} while(0)
#define NGB_FLUSH_DEL(ngb) do { \
	if (!(ngb)->ngb_flushprev) \
		ospf_flush_list = (ngb)->ngb_flushnext; \
	else \
		(ngb)->ngb_flushprev->ngb_flushnext = (ngb)->ngb_flushnext; \
	if ((ngb)->ngb_flushnext) \
		(ngb)->ngb_flushnext->ngb_flushprev = (ngb)->ngb_flushprev; \
	(ngb)->ngb_flushnext = NULL; \
} while(0)
#define NGB_FLUSH_ALL(ngb) \
	for((ngb) = ospf_flush_list; (ngb); (ngb) = (ngb)->ngb_flushnext)

/* List operation trace.
 */
#ifdef OSPF_NGBLIST_TRACE
#define TRACE_NGB_LIST(ngb, type, key, msg, age) \
	tracef("OSPF FLOOD: NGB %A LIST %s type %s key %A:%A age %d ", \
		sockbuild_in(0, (ngb)->ngb_rtid), \
		(msg), \
		trace_state((type), ospf_lsa_types), \
		sockbuild_in(0, (key)->lsk_id), \
		sockbuild_in(0, (key)->lsk_advrt), \
		(age))
#define TRACE_NGB_LIST_END(cpt) \
	trace_tf(ospf_trace_options, TR_OSPF_FLOOD, 0, ("size %d", (cpt)))
#else
#define TRACE_NGB_LIST(ngb, type, id, msg, age)
#define TRACE_NGB_LIST_END(cpt)
#endif

/* Add a vertex into the list of LS to be sent
 */
#define NGB_LIST_ADD_DS(ngb, type, key, vtx) do { \
	vtxList_t *vl; \
	vl = ( vtxList_t *)task_block_alloc(ospf_vtxlist_block); \
	vl->l_next = (ngb)->ngb_dataSum; \
	vl->l_vtx = (vtx); \
	(ngb)->ngb_dataSum = vl; \
	(ngb)->ngb_DSlistSize++; \
} while(0)
#define NGB_LIST_ADD_LSR(ngb, type, key, vtx) do { \
	db_add_direct(&(ngb)->ngb_lsReq, type, key, vtx); \
	TRACE_NGB_LIST(ngb, type, key, "ADD LSR queue", VTX_AGE((vtx))); \
	TRACE_NGB_LIST_END((ngb)->ngb_lsReq.pt_used_nodes); \
} while(0)
#define NGB_LIST_ADD_LSR_BUFF(ngb, type, key, vtx) do { \
	db_add_direct(&(ngb)->ngb_LSRbuffer, type, key, vtx); \
	TRACE_NGB_LIST(ngb, type, key, "ADD LSR buffer", VTX_AGE((vtx))); \
	TRACE_NGB_LIST_END((ngb)->ngb_LSRbuffer.pt_used_nodes); \
} while(0)
#define NGB_LIST_ADD_RMX(ngb, type, key, vtx) do { \
	db_add_direct(&(ngb)->ngb_lsRmx, type, key, vtx); \
	TRACE_NGB_LIST(ngb, type, key, "ADD RMX queue", VTX_AGE((vtx))); \
	TRACE_NGB_LIST_END((ngb)->ngb_lsRmx.pt_used_nodes); \
} while(0)
#define NGB_LIST_ADD_RMX_BUFF(ngb, type, key, vtx) do { \
	db_add_direct(&(ngb)->ngb_RMXbuffer, type, key, vtx); \
	(ngb)->ngb_RMXbuffer_len += (vtx)->vtx_len; \
	TRACE_NGB_LIST(ngb, type, key, "ADD RMX buffer", VTX_AGE((vtx))); \
	TRACE_NGB_LIST_END((ngb)->ngb_RMXbuffer.pt_used_nodes); \
} while(0)

/* true if the list is empty (the queue and the buffer!).
 */
#define NGB_LIST_EMPTY_LSR(ngb)	\
	(!(ngb)->ngb_lsReq.pt_used_nodes && !(ngb)->ngb_LSRbuffer.pt_used_nodes)

/* Get the next vertex of the list. Free the list structure.
 */
#define NGB_LIST_NEXT_DS(ngb, vtx) { \
	vtxList_t *vl; \
	if ((ngb)->ngb_dataSum) { \
		vtx = (ngb)->ngb_dataSum->l_vtx; \
		vl = (ngb)->ngb_dataSum->l_next; \
		task_block_free(ospf_vtxlist_block, (ngb)->ngb_dataSum); \
		(ngb)->ngb_dataSum = vl; \
		(ngb)->ngb_DSlistSize--; \
	} \
	else \
		vtx = NULL; \
}
	
#define NGB_LIST_NEXT_LSR(ngb, vtx) do {	\
	((vtx) = ngb_list_next(&(ngb)->ngb_lsReq)); \
	if (vtx) { \
		TRACE_NGB_LIST(ngb,vtx->vtx_type, &vtx->vtx_key, \
			"NEXT LSR queue", VTX_AGE((vtx)));\
		TRACE_NGB_LIST_END((ngb)->ngb_lsReq.pt_used_nodes); \
	} \
} while(0);

#define NGB_LIST_NEXT_RMX(ngb, vtx) do { \
	((vtx) = ngb_list_next(&(ngb)->ngb_lsRmx)); \
	if (vtx) { \
		TRACE_NGB_LIST(ngb,vtx->vtx_type, &vtx->vtx_key, \
			"NEXT RMX queue", VTX_AGE((vtx)));\
	 	TRACE_NGB_LIST_END((ngb)->ngb_lsRmx.pt_used_nodes); \
	} \
} while(0);

#define NGB_LIST_FIND_LSR(req, ngb, vtx) \
	((req) = db_find(&(ngb)->ngb_lsReq, (vtx)->vtx_type, &(vtx)->vtx_key, \
		&__searchdata__)? (req): \
		(req) = db_find(&(ngb)->ngb_LSRbuffer, (vtx)->vtx_type, \
			&(vtx)->vtx_key, &__searchdata__))

#define NGB_LIST_DEL_LSR(ngb, vtx) do { \
	if (!db_del_fast(&(ngb)->ngb_lsReq, (vtx))) {\
		TRACE_NGB_LIST(ngb,vtx->vtx_type, &vtx->vtx_key, \
				"DEL LSR buffer", VTX_AGE((vtx))); \
		(vtx) = db_del_fast(&(ngb)->ngb_LSRbuffer, (vtx)); \
		TRACE_NGB_LIST_END((ngb)->ngb_LSRbuffer.pt_used_nodes); \
	} \
	else { \
		TRACE_NGB_LIST(ngb, vtx->vtx_type, &vtx->vtx_key, \
			"DEL LSR queue", VTX_AGE((vtx)));\
		TRACE_NGB_LIST_END((ngb)->ngb_lsReq.pt_used_nodes); \
	} \
} while(0);

#define LSR_SET(lsr, type, id, advrt) { \
	(lsr)->lsrd_type = htonl((type)); \
	(lsr)->lsrd_id = (id); \
	(lsr)->lsrd_advrt = (advrt); \
}

/* check the checksum of the LSA. The checksum is computed when the LSA
 * is in this final stage, i.e. network byte order.
 */
#define ospf_check_lsacksum(lsa) \
	(iso_cksum((void_t)&((lsa)->lsa_options), \
		(size_t)ntohs((lsa)->lsa_len) - sizeof(u_int16), NULL))

#define NGB_ALL(ngb)	{ \
	ospf_intf_t *__ospf_intf; \
	OSPF_IF_ADDR(__ospf_intf) { \
		for((ngb) = __ospf_intf->ospf_if_ngb; (ngb); \
			(ngb) = (ngb)->ngb_next) {
#define NGB_ALL_END(ngb)	\
		} \
	} OSPF_IF_ADDR_END(__ospf_intf); \
}

#define ALL_NGB_TERMINATE(oif) do { \
	ospf_ngb_t *__next, *__ngb; \
	for(__ngb = (oif)->ospf_if_ngb; __ngb; __ngb = __next) { \
		__next = __ngb->ngb_next; \
		ospf_ngb_terminate(__ngb); \
	(oif)->ospf_if_ngb = NULL; \
	} \
} while(0)

#define NGB_ISDR(ngb) \
	((ngb)->ngb_intf->ospf_if_drAddr == (ngb)->ngb_addr)
#define NGB_ISBACKUP(ngb) \
	((ngb)->ngb_intf->ospf_if_backupDrAddr == (ngb)->ngb_addr)
#define INTF_ISBACKUP(ospf_intf) \
	((ospf_intf)->ospf_if_backupDrAddr == (ospf_intf)->ospf_if_addr)
#define INTF_ISDR(ospf_intf) \
	((ospf_intf)->ospf_if_drAddr == (ospf_intf)->ospf_if_addr)

#define OSPF_TSI_GET(rth, rtbit, vtx) rttsi_get((rth), (rtbit), (byte *) &(vtx))
#define OSPF_TSI_SET(rth, rtbit, vtx) rttsi_set((rth), (rtbit), (byte *) &(vtx))
#define OSPF_TSI_CLEAR(rth, rtbit, vtx) rttsi_reset((rth), (rtbit))

/* God know what those are for.... (There're from the old ospf export code).
 */
#define OSPF_EXPORT_TAG         0x04       /* Tag is present */
#define OSPF_EXPORT_TAG_METRIC2 0x08       /* Tag is in metric2 vs metric */
#define OSPF_EXPORT_TYPE1       0x01
#define OSPF_EXPORT_TYPE2       0x02
#define OSPF_LIMIT_EXPORTTYPE   OSPF_EXPORT_TYPE1, OSPF_EXPORT_TYPE2
#define OSPF_LIMIT_METRIC	0, OSPF_LSINFINITY

extern task            	*ospf_task;
extern flag_t          	ospf_flags;
extern trace           	*ospf_trace_options;
extern block_t         	ospf_if_blocks;
extern block_t         	ospf_ngb_blocks;
extern block_t          ospf_ptw_blocks;
extern block_t          ospf_vtx_blocks;
extern block_t          ospf_vtxlink_blocks;
extern block_t          ospf_vtxlist_block;
extern block_t          ospf_cdtList_block;
extern block_t          ospf_treenodes_blocks;
extern ospf_pck_t      	ospf_pck_buffer[OSPF_BUFFER_NBPCK];
extern int             	ospf_pck_bffct;
extern adv_entry       	*ospf_current_intf;
extern const bits	ospf_trace_types[];
extern const bits	ospf_pck_flags[];
extern const bits	ospf_pck_options[];
extern const bits	ospf_if_states[];
extern const bits	ospf_if_events[];
extern const bits	ospf_ngb_states[];
extern const bits	ospf_ngb_events[];
extern const bits	ospf_rtr_bits[];
extern const bits	ospf_auth_types[];
extern const bits	ospf_pck_types[];
extern rtid_t		ospf_rtid;
extern sockaddr_un	*all_ospf_routers;
extern sockaddr_un	*all_ospf_dr_routers;
extern sockaddr_un	*all_ospf_eligible;
extern ospf_area_t	*ospf_area_list;
extern ospf_area_t     	*ospf_current_area;
extern task_timer	*ospf_spf_timer;
extern time_t		ospf_spfInt;
extern pref_t		ospf_preference;
extern cost_t          	ospf_if_cost;
extern metric_t		ospf_lsinfinity;
extern sockaddr_un     	*sock_rtmask;
extern int		ospf_nbNgbNotReady;
extern iflist_t        	*conf_iflist;
extern ospf_intf_t     	*ospf_cr_vintf;
extern adv_entry      	*ospf_export_list;
extern ospf_ngb_t	*ospf_flush_list;
extern time_t           ospf_lsRefreshTime;
extern int		ospf_nbAreas;
extern rt_entry		*ospf_nullrt;
extern flag_t           ospf_rtrbits;
vtxList_t       	*ospf_floodQueue1;
vtxList_t       	*ospf_floodQueue2;
task_timer      	*ospf_floodTimer1;
task_timer      	*ospf_floodTimer2;

void	ospf_init __A((void));
void	ospf_var_init __A((void));
void	ospf_init_if __A((ospf_area_t *, if_addr *));
ospf_intf_t *ospf_new_intf __A((ipaddr_t, ospf_area_t *, flag_t));
int	ospf_check_all __A((void));
void	ospf_ifchange __A((task *, if_addr *));
void	ospf_if_add __A((ospf_intf_t *));
void	ospf_if_down __A((ospf_intf_t *));
void	ospf_wait_timer  __A((task_timer *, time_t));
void	ospf_hello_timer __A((task_timer *, time_t));
void	ospf_send_hello __A((ospf_intf_t *));
void	ospf_send __A((ospf_pck_t *, int, ospf_intf_t *, sockaddr_un *));
void	ospf_send_dd __A((ospf_ngb_t *));
void	ospf_send_lsr __A((ospf_ngb_t *));
void	ospf_send_rmx __A((ospf_ngb_t *));
void	ospf_ngb_acktimer __A((task_timer *, time_t));
void	ospf_send_ack __A((ospf_ngb_t *, struct _ospf_lsa_t *,
		struct _vertex_t *, flag_t));
void	ospf_send_lsudirect __A((ospf_ngb_t *, struct _vertex_t **, int));
void	ospf_recv __A((task *));
int	ospf_validate_hdr __A((ospf_pck_t *, int, sockaddr_un *, sockaddr_un *,
		ospf_intf_t **));
int	ospf_hello_recv __A((ospf_hello_pck_t *, int, ospf_intf_t *,
		sockaddr_un *));
int	ospf_dd_recv __A((ospf_dd_pck_t *, int, ospf_intf_t *,sockaddr_un *));
int	ospf_lsr_recv __A((ospf_lsr_pck_t *, int, ospf_intf_t *,sockaddr_un *));
int	ospf_lsu_recv __A((ospf_lsu_pck_t *, int, ospf_intf_t *,sockaddr_un *));
ospf_ngb_t *ospf_find_ngb __A((ospf_intf_t *, rtid_t));
void	ospf_ngb_event __A((event_t, ospf_ngb_t *));
int	ospf_if_event __A((event_t, ospf_intf_t *));
int	ospf_electDR __A((ospf_intf_t *));
int	ospf_adjacency __A((ospf_ngb_t *, ospf_intf_t *));
void	ospf_exchangeStart __A((ospf_ngb_t *, ospf_intf_t *));
void	ospf_rmx_timer __A((task_timer *, time_t));
void	ospf_exchangeAll __A((ospf_ngb_t *, ospf_intf_t *));
void	ospf_ngb_down __A((ospf_ngb_t *));
void	ospf_ngb_cleanup __A((ospf_ngb_t *));
void	ospf_ngb_cleanLists __A((ospf_ngb_t *));
void	ospf_ngb_inactimer __A((task_timer *, time_t));
void	ospf_cleanup __A((task *));
void	ospf_dump __A((task *, FILE *));
void	ospf_reinit __A((task *));
void	ospf_terminate __A((task *));
void	ospf_conf_newintf __A((iflist_t *));
void	ospf_conf_setintf __A((iflist_t *, ospf_intf_t *, int, void_t));
int	ospf_count_ngb __A((ospf_intf_t *));
ospf_intf_t *ospf_if_find __A((ipaddr_t));
void	ospf_area_init __A((void));
ospf_area_t *ospf_area_create __A((areaid_t, flag_t, flag_t, auth_t));
ospf_area_t *ospf_find_area  __A((areaid_t));
ospf_ngb_t *ospf_new_ngb __A((ospf_intf_t *, rtid_t, rtpri_t,
		ipaddr_t, ipaddr_t, ipaddr_t, int));
sockaddr_un *ospf_send_addr __A((ospf_intf_t *, int, ospf_ngb_t *));
void	ospf_reconfig __A((void));
void	ospf_area_remove __A((ospf_area_t *));
void	ospf_intf_remove __A((ospf_intf_t *));
void	ospf_ngb_terminate __A((ospf_ngb_t *));
void	ospf_reset_default __A((void));

/* In new_ospf_trace.c
 */
void	ospf_trace_packet __A((ospf_pck_t *, int, int, flag_t, sockaddr_un *,
		sockaddr_un *));
void	ospf_trace_lsh __A((ospf_lsa_hdr_t *, char *));
void	ospf_trace_vtx __A((struct _vertex_t *, char *));
void	ospf_trace_lsa __A((ospf_lsa_t *, char *));

