/*
 * Copyright (c) 1996 The Regents of the University of Michigan
 * All Rights Reserved
 * 
 * License to use, copy, modify, and distribute this software and its
 * documentation can be obtained from Merit at the University of Michigan.
 * 
 * Merit GateDaemon Project
 * 4251 Plymouth Road, Suite C
 * Ann Arbor, MI 48105
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE REGENTS OF THE
 * UNIVERSITY OF MICHIGAN AND MERIT DO NOT WARRANT THAT THE
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the
 * University of Michigan and Merit shall not be liable for
 * any special, indirect, incidental or consequential damages with respect
 * to any claim by Licensee or any third party arising from use of the
 * software. GateDaemon was originated and developed through release 3.0
 * by Cornell University and its collaborators.
 * 
 * Please forward bug fixes, enhancements and questions to the
 * gated mailing list: gated-people@gated.merit.edu.
 * 
 * This copyright has ben automaticly added by the util/addcopyright.pl program.
 * __END_OF_COPYRIGHT__
 */

/* OSPF Link State Database Structures.
 */

typedef u_int8 linktype_t;
typedef u_int8 lstype_t;
typedef u_int32 lsid_t;
typedef s_int32 lsseq_t;
typedef s_int16 lsage_t;
typedef u_int16 lslenght_t;
typedef u_int16 lscksum_t;
typedef u_int8	qref_t;	     /* Its size defines the max number of neighbour */

#define OSPF_MAX_NGB	(8 * sizeof(qref_t))

/* This is machine dependent. The two bytes bis_byte:bis_bits will
 * be compared as 16 bits numbers.
 */
typedef union _bitindex_t {
	u_int16	bi_fullindex;
	struct {
#if BYTE_ORDER == BIG_ENDIAN
		u_int8 bis_byte;
		u_int8 bis_bits;
#else
		u_int8 bis_bits;
		u_int8 bis_byte;
#endif
	} bi_struct;
} bitindex_t;

#define bi_byte		bi_struct.bis_byte
#define bi_bits		bi_struct.bis_bits

typedef struct _ospf_vtxLink_t {
	struct _ospf_vtxLink_t 	*vl_next;	/* 4 */
	struct _vertex_t 	*vl_vtx;	/* 4 */
	linktype_t		vl_type;	/* 1 */
	u_int8 			vl_flags;	/* 1 */
	cost_t 			vl_cost;	/* 2 */
	u_int32 		vl_data;	/* 4 */
} ospf_vtxLink_t;

#define vl_id		vl_vtx->vtx_id

#define LINK_MATCH(new, old) do { \
	(new)->vl_cost = (old)->vl_cost; \
	(new)->vl_flags &= ~LF_ASYM; \
	(old)->vl_flags &= ~LF_ASYM; \
} while(0)

/* a generic vertex on the graph, also used to store any kind of LS in the DB.
 * The begining must match the tree_node_t
 * structure.
 */
typedef struct _vertex_t {			/* size 56 */
	union _tree_node_t *vtx_node_right;	/* 4 */
	union _tree_node_t *vtx_node_left;	/* 4 */
	u_int16		vtx_flags;	/* 2 */
	bitindex_t	vtx_bitindex;	/* 2 */
	lskey_t		vtx_key;	/* 8 */	/* End of tree_node_t */
	time_t		vtx_nextRecv;	/* 4 */
	u_int8		vtx_rtr_bits;	/* 1 */
        options_t       vtx_options;	/* 1 */
	lscksum_t	vtx_cksum;	/* 2 */
        lsage_t         vtx_install;	/* 2 */
	lslenght_t	vtx_len;	/* 2 */
	lsseq_t		vtx_seq;	/* 4 */
	union {				/* 4 */
		ospf_vtxLink_t *_u_vtx_links;	/* For LST_RTR/NTW */
		u_int32	        _u_vtx_tosmet;	/* For LST_ASEX */
	} _vtx_u;
	rt_entry	*vtx_rt;	/* 4 */
	gw_entry	*vtx_gw;	/* 4 */
	ipaddr_t	vtx_forward;	/* 4 */
	u_int8		vtx_changes;	/* 1 */
	u_int8		vtx_spfcount;	/* 1 */
	qref_t		vtx_qref;	/* 1 */
	lstype_t	vtx_type;	/* 1 */
} vertex_t;

#define vtx_id		vtx_key.lsk_id
#define vtx_advrt	vtx_key.lsk_advrt
#define vtx_cost	vtx_rt->rt_metric
#define	vtx_mask	vtx_rt->rt_dest_mask->in.gin_addr.s_addr
#define vtx_nextHops	vtx_rt->rt_routers
#define vtx_nbNextHops	vtx_rt->rt_n_gw
#define vtx_outIntf	vtx_rt->rt_ifaps
#define vtx_tag		vtx_rt->rt_tag
#define vtx_rt_gw	vtx_rt->rt_gwp
#define vtx_preference	vtx_rt->rt_preference
#define vtx_preference2	vtx_rt->rt_preference2
#define vtx_links	_vtx_u._u_vtx_links
#define vtx_tosmet	_vtx_u._u_vtx_tosmet

/* Must be the same as RTCF_xxx in rt_table.h
 */
#define VC_TAG		((u_char)RTCF_TAG)
#define VC_COST		((u_char)RTCF_METRIC)
#define VC_NEXTHOP	((u_char)RTCF_NEXTHOP)
#define VC_STATE	((u_char)RTCF_METRIC)

#define OSPF_RT_VTX(rt)	((vertex_t *)((rt)->rt_data))
#define VTX_AREA(vtx)	((ospf_area_t *)((vtx)->vtx_gw->gw_data))
#define VTX_TREE(vtx)	(&(VTX_AREA(vtx)->area_tree))

/* flags for a graph object. LF_HIDDEN means that the object is not announced
 * in a LSA but obviously must exist. LF_FOUND is for graph lookups.
 * LF_REQUEST means that the LS has been sent in a DD packet, but was not
 * in the DB yet. So we created one, and wait for a complete desciption from
 * the neighbour. LF_ASYM means that the reverse link has not been
 * announced (i.e. is marked as LF_HIDDEN)
 * LF_PREM is used for section 14.1 of the RFC, i.e. premature aging +
 * wrap up of sequence number.
 * LF_ROUTE is set when we force a route to an LS (usualy a host) to
 * be created.
 * !!! NOTE !!! No more than 16 flags can be defined.
 * Also the tree_node_t flags are shared with those flags.
 */
#define LF_FOUND	0x0001
#define LF_HIDDEN	0x0002
#define LF_FLUSH	0x0004
#define LF_REQUEST	0x0008
#define LF_ASYM		0x0010
#define TNF_EXTERN	0x0020		/* Only used by nodes */
#define LF_SEQWRAP	0x0040
#define LF_DELETE	0x0080
#define LF_ROUTE	0x0100

#define RES_LS_SEQ	0x80000000
#define MIN_LS_SEQ	0x80000001
#define MAX_LS_SEQ	0x7fffffff
#define OSPF_MAXAGE	3600		/* seconds (= 1h) */
#define OSPF_MAXAGEDIFF	900		/* seconds (= 15min) */
#define OSPF_MINLSINT	5		/* seconds */
#define AGEDIFF(age1, age2) \
		((age2) > (age1)? (age2) - (age1): (age1) - (age2))

/* Database management. Patricia tree. A node can contain the whole
 * data (in this case a node and a vertex is the same), or point to
 * an external data (the vertex). Should start like the vertex_t structure.
 */
typedef union _tree_node_t {		/* size 20 */
	vertex_t	tnu_vtx;
	struct {			/* Must match the begining of vtx*/
		union _tree_node_t 	*tnu_right;
		union _tree_node_t 	*tnu_left;
		u_int8		    	tnu_flags;
		lstype_t		tnu_type;
		bitindex_t	    	tnu_bitindex;
		u_int32			tnu_key[2];
		void_t		    	tnu_data;
	} 		tnu_extnode;
} tree_node_t;

#define tn_flags	tnu_extnode.tnu_flags
#define tn_data		tnu_extnode.tnu_data
#define tn_left		tnu_extnode.tnu_left
#define tn_right	tnu_extnode.tnu_right
#define tn_key		tnu_extnode.tnu_key
#define tn_type		tnu_extnode.tnu_type
#define tn_bitindex	tnu_extnode.tnu_bitindex

/* Used when adding/search a node. Makes the code faster.
 * Currently the last node seen when searching for a node (unsuccessfully)
 */
typedef struct _searchdata_t {
	tree_node_t *sd_last;
} searchdata_t;

/* Deepth of the patricia tree. keys are 32 (u_int32) bits longs).
 */
#define PT_DEEPTH	(sizeof(u_int32) * 8)

/* Used to walk along the tree.
 */
typedef struct _ospf_ptwalk_t {
	struct _ospf_ptwalk_t *tw_next;		/* We keep them chained */
	tree_node_t	*tw_node;		/* The current node */
	tree_node_t	**tw_start;		/* From where we started */
	tree_node_t	*tw_stack[PT_DEEPTH];	/* The stack of previous nodes*/
	int 		tw_stkind;		/* Index in the stack */
	db_tree_t	*tw_tree;		/* The tree we are using */
	int		tw_type;		/* The current type of tree */
	ospf_area_t	*tw_area;		/* Current area */
} ospf_ptwalk_t;

#define tw_vtx		tw_node->tn_vtx
#define tw_mask		tw_node->tn_mask
#define tw_id		tw_node->tn_vtx->vtx_id

#define DB_VTXLINK_ALLOC() \
		((ospf_vtxLink_t *)task_block_alloc(ospf_vtxlink_blocks))
#define PT_WALK_ALLOC()	  ((ospf_ptwalk_t *)task_block_alloc(ospf_ptw_blocks))
#define DB_VTX_ALLOC()    ((vertex_t *)task_block_alloc(ospf_vtx_blocks))
#define DB_VTXLINK_FREE(vl) task_block_free(ospf_vtxlink_blocks, vl)
#define DB_VTX_FREE(vtx)   task_block_free(ospf_vtx_blocks, vtx);
#define PT_WALK_FREE(ptw)   task_block_free(ospf_ptw_blocks, ptw)

#define RTMASK		0xffffffff

/* Patricia tree partitioning
 * Return the vertex associared with the node
 */
#define TREENODE_VTX(node) \
	((vertex_t *)(BIT_TEST((node)->tn_flags, TNF_EXTERN)? \
		(node)->tn_data: (node)))
/* True is the branch is downstream from the node.
 */
#define TREENODE_DOWN(node, branch) \
	BITINDEX_STRICTLOWER((node)->tn_bitindex, (branch)->tn_bitindex)
/* Allocation, free.
 */
#define TREENODE_ALLOC(new, data, key, keylen) do { \
	(new) = (tree_node_t *)task_block_alloc(ospf_treenodes_blocks); \
	(new)->tn_data = (void_t)data; \
	(void)bcopy((key), &(new)->tn_key, (keylen)); \
	(new)->tn_flags = TNF_EXTERN; \
} while(0);
#define TREENODE_FREE(node) do { \
	if (BIT_TEST((node)->tn_flags, TNF_EXTERN)) \
		task_block_free(ospf_treenodes_blocks, (node)); \
	else \
		task_block_free(ospf_vtx_blocks, (node)); \
} while(0);

/* The lowest bit index possible. The right bit is number 0, the left one is
 * 31 (or 63).  The bit index goes down as we go down the trye. 
 * This bit index will only be used by initial lieaf. XXXXXX
 */
#define BITINDEX_RESET(bitindex)     ((bitindex).bi_fullindex = 0x0801)
#define BITINDEX_ISRESET(bitindex)   ((bitindex).bi_fullindex == 0x0801)
/* Initiale the bit index to the smallest bit index before starting
 * a search.
 */
#define BITINDEX_INIT(bitindex)		((bitindex).bi_fullindex = 0x0)
/* Copy
 */
#define BITINDEX_SET(new, old)	((new).bi_fullindex = (old).bi_fullindex)
/* Compare. Return 0 if same.
 */
#define BITINDEX_CMP(bi1, bi2)	((bi1).bi_fullindex - (bi2).bi_fullindex)
/* True if bi1 is stricly smaller that bi2
 */
#define BITINDEX_STRICTLOWER(bi1, bi2) ((bi1).bi_fullindex < (bi2).bi_fullindex)
/* True if bi1 is smaller that bi2
 */
#define BITINDEX_LOWER(bi1, bi2) ((bi1).bi_fullindex <= (bi2).bi_fullindex)
/* Test the bit of a given key.
 */
#if BYTE_ORDER == BIG_ENDIAN
#define TREEBIT_TEST(bitindex, key, keylen) \
	((key)[keylen - (bitindex).bi_byte] & (bitindex).bi_bits)
#else
#define TREEBIT_TEST(bitindex, key, keylen) \
	((key)[(bitindex).bi_byte - 1] & (bitindex).bi_bits)
#endif
/* Compare two keys. True if the same. Uses the fact that keylen = 4 or 8.
 */
#define TREEKEY_CMP(key1, keylen, key2) 				\
	((keylen == 4)? *((u_int32 *)(key1)) == *((u_int32 *)(key2)): 	\
		((u_int32 *)(key1))[0] == ((u_int32 *)(key2))[0] &&	\
		((u_int32 *)(key1))[1] == ((u_int32 *)(key2))[1])

/* Stack manipulation
 */
#define PTW_PUSH(ptw, node)	((ptw)->tw_stack[(ptw)->tw_stkind++] = node)
#define PTW_POP(ptw)		(((ptw)->tw_stkind)? \
					(ptw)->tw_stack[--(ptw)->tw_stkind]: \
					(tree_node_t *)NULL)

/* Whatever we need to initialize the LSA structures. LSA_SET will use
 * network byte order. LSA_SET_H uses host byte order.
 */
#define LSA_SET(lsa, age, opt, type, id, advrt, seq, cksum, len) { \
	(lsa)->lsa_age = htons((age)); \
	(lsa)->lsa_options = (opt); \
	(lsa)->lsa_type = (type); \
	(lsa)->lsa_id = (id); \
	(lsa)->lsa_advrt = (advrt); \
	(lsa)->lsa_seq = htonl((seq)); \
	(lsa)->lsa_len = htons((len)); \
	if (!(cksum)) { \
		(void)iso_cksum((void_t)&(lsa)->lsa_options, \
			(size_t)(len) - sizeof(u_int16), \
			(byte *)&(lsa)->lsa_cksum);\
		ASSERT(!ospf_check_lsacksum(lsa)); \
	} \
	else \
		(lsa)->lsa_cksum = (cksum); \
	/*HTONS((lsa)->lsa_cksum);*/ \
}
#define LSA_SET_H(lsa, age, opt, type, id, advrt, seq, cksum, len) { \
	(lsa)->lsa_age = (age); \
	(lsa)->lsa_options = (opt); \
	(lsa)->lsa_type = (type); \
	(lsa)->lsa_id = (id); \
	(lsa)->lsa_advrt = (advrt); \
	(lsa)->lsa_seq = (seq); \
	(lsa)->lsa_cksum = (cksum); \
	(lsa)->lsa_len = (len); \
}
#define LSA_EXT_SET(lsext, mask, bitE, tos, metric, nexthop, tag) { \
	(lsext)->lsext_mask = (mask); \
	(lsext)->lsext_tosmet32 = \
		htons((((bitE) & 0x1) << 31) | \
			(((tos) & 0x7f) << 24) | ((metric) & 0xffffff)); \
	(lsext)->lsext_forw = (nexthop); \
	(lsext)->lsext_tag = htonl((tag)); \
}
#define LSA_SL_SET(sla, mask, tos, met) { \
	(sla)->lssla_mask = (mask); \
	(sla)->lssla_tosmet32 = ntohl(((tos) << 24) | ((met) & 0xffffff)); \
}
#define LSA_RT_SET(lsa, bits, nblinks)	{ \
	(lsa)->lsrtr_bits = (bits); \
	(lsa)->lsrtr_null = 0; \
	(lsa)->lsrtr_nblinks = htons((nblinks)); \
}
#define LSA_RL_SET(rtlink, id, data, type, nbtos, metric) { \
	(rtlink)->lsrl_id = (id); \
	(rtlink)->lsrl_data = (data); \
	(rtlink)->lsrl_type = (type); \
	(rtlink)->lsrl_nbtos = (nbtos); \
	(rtlink)->lsrl_metric = htons((metric)); \
}
#define LSA_NT_SET(lsnt, mask)	{ (lsnt)->lsntw_mask = (mask); }
#define LSA_NL_SET(lsnl, id)	{ *(lsnl) = (id); }
#define LSA_NTW_NBRTR(lsa, len) \
	(((len) - sizeof(ospf_lsa_ntw_t)) / sizeof(u_int32) + 1)
#define NTOH_LSH(lsh) { \
	NTOHS((lsh)->lsh_age); \
	NTOHL((lsh)->lsh_seq); \
	(lsh)->lsh_cksum; \
	NTOHS((lsh)->lsh_len); \
}
#define NTOH_LSRD(lsrd) { \
	NTOHL((lsrd)->lsrd_type); \
}
#define NTOH_RTLINK(link) { \
	NTOHS((link)->lsrl_metric); \
}

#define mask_bits_32(m)	(mask_bits(sockbuild_in(0, (m))))
#define IFNET(ospf_intf) ((ospf_intf)->ospf_if_addr & (ospf_intf)->ospf_if_mask)

#define OSPF_GEN_RTRLSA()	{ \
	ospf_area_t *area = ospf_area_list; \
	for(; area; area = area->area_next) \
		ospf_gen_rtrlsa(area); \
}

#define PT_WALK(tree, vtx) { \
	tree_node_t *__node; \
	PT_WALK_NODE(tree, __node) { \
		(vtx) = TREENODE_VTX(__node); \
		if ((vtx)) do {

#define PT_WALK_END(tree, vtx) \
		} while(0); \
	} PT_WALK_NODE_END(tree, __node); \
}

#define PT_WALK_TYPE(tree, type, vtx) do { 				      \
	tree_node_t *__node;				      		      \
	tree_node_t *__next;                                                  \
        tree_node_t *__Stack[PT_DEEPTH];				      \
	int __ind = 0;					      		      \
	__node = (tree)->pt_root[(type)];				      \
	while(__node) {					         	      \
		(vtx) = TREENODE_VTX(__node);				      \
		if (vtx) do {
			/* Do something */
#define PT_WALK_TYPE_END(tree, type, vtx) 				      \
		} while(0);                                                   \
                /* Get the next node (c.f. ospf_pt_walk()) */                 \
                if (TREENODE_DOWN(__node, __node->tn_left)) {                 \
                        __Stack[__ind++] = __node;                            \
                        __node = __node->tn_left;                             \
                        continue;                                             \
                }                                                             \
                else if (TREENODE_DOWN(__node, __node->tn_right)) {           \
                        __Stack[__ind++] = __node;                            \
                        __node = __node->tn_right;                             \
                        continue;                                             \
                }                                                             \
                while((__next = (__ind > 0)? __Stack[--__ind]: NULL)) {       \
                        if (TREENODE_DOWN(__next, __next->tn_right) &&        \
                                        __next->tn_right != __node) {         \
                                __Stack[__ind++] = __next;                    \
                                __node = __next->tn_right;                    \
                                break;                                        \
                        }                                                     \
                        __node = __next;                                      \
                }                                                             \
                if (!__next)                                                  \
                        break;                                                \
        }								      \
} while(0)

#define PT_WALK_NODE(tree, node) {					      \
	int __type = 0, __ind = 0;					      \
	tree_node_t *__next;						      \
	tree_node_t *__Stack[PT_DEEPTH];				      \
	for(__type = 0; __type < LST_MAXTYPE; __type++) {		      \
		(node) = (tree)->pt_root[__type];			      \
		while((node)) {						      \
			do {
				/* Do something... */			      \
#define PT_WALK_NODE_END(tree, node) 					      \
			} while(0);					      \
			/* Get the next node (c.f. ospf_pt_walk()) */	      \
			if (TREENODE_DOWN((node), (node)->tn_left)) {	      \
				__Stack[__ind++] = (node);		      \
				(node) = (node)->tn_left;		      \
				continue;				      \
			}						      \
			else if (TREENODE_DOWN((node), (node)->tn_right)) {   \
				__Stack[__ind++] = (node);		      \
				(node) = (node)->tn_right;		      \
				continue;				      \
			}						      \
			while((__next = (__ind > 0)? __Stack[--__ind]: NULL)) {\
				if (TREENODE_DOWN(__next, __next->tn_right) && \
						__next->tn_right != (node)) { \
					__Stack[__ind++] = __next;	      \
					(node) = __next->tn_right;	      \
					break;				      \
				}					      \
				(node) = __next;			      \
			}						      \
			if (!__next)					      \
				break;					      \
		}							      \
	}								      \
}

/* A vertex is not inthe DB is db_find returns null, or if the LF_REQUEST
 * flag is set.
 */
#define LS_NOT_INDB(vtx) (!(vtx) || BIT_TEST((vtx)->vtx_flags, LF_REQUEST))

/* The 'vtx_install' time is hte time that the first instance has
 * been create in the original router, i.e. the time when lsa_age == 0.
 */
/*
#define VTX_AGE_SET(vtx, age)	((vtx)->vtx_install = (time_sec & 0xffff) - age)
#define VTX_AGE(vtx)		((time_sec & 0xffff) - (vtx)->vtx_install)
#define VTX_AGE_MAX(vtx)	\
	((time_sec & 0xffff) - (vtx)->vtx_install > ospf_maxAge)
 */
#define AGESLOT		10000
#define VTX_AGE_SET(vtx, age)	\
	((vtx)->vtx_install = (time_sec - (age)) % AGESLOT)
#define VTX_AGE(vtx)		\
	((time_sec - (vtx)->vtx_install) % AGESLOT)
#define VTX_AGE_MAX(vtx)	\
	(VTX_AGE(vtx) > ospf_maxAge)

#define LSADD_OLDER	1
#define LSADD_OK	0

#define CKL_DEAD	1
#define CKL_OK		2
#define CKL_MISSING	3

/* What type of acknowlegment to send. ACKF_NOW means sent it NOW,
 * ACKF_FBACK means we send the LSA back to the receiving interface,
 * ACKF_IMPLIED means we got an implied acknowlegment. ACKF_DELAY means
 * that the ACK timer timed out, and we need to send a delayed ACK.
 * ACKF_NEW is used when the LSA is new, but not flooded back.
 */
#define ACKF_IMPLIED	1
#define ACKF_NOW	2
#define ACKF_FBACK	3
#define ACKF_DELAY	4
#define ACKF_NEW	5

/* The candidate list must be a sorted-by-cost ordered list of vertices.
 * We will use a radix tree indexed with the cost. All leaves will
 * point to a list of vertices os equal cost.
 * The operations on the list are: get the next lowest cost vertex, remove it
 * from the list; put a new vertex on the list; update a given vertex (cost)
 * on the list (remove + add).
 * The code will be the same as db_XXX(), except the lists management at
 * the leaves.
 * If the candidate list is big, that justifies all those complex structures.
 * If the list stays small, well, i guess i wasted my time!
 *
 * Ok, first i try with a simple ordered double linked list.
 * A vertex is in the candidate list if the LF_FOUND flag is set.
 * l_vl is the best link to the vertex.
 */
typedef struct _cdtList_t {
        struct _cdtList_t *l_prev;
        struct _cdtList_t *l_next;
        vertex_t *l_vtx;
        ospf_vtxLink_t *l_vl;
        rt_changes *l_rtchg;
} cdtList_t;

extern const bits      ospf_vtx_flags[];
extern const bits      ospf_lslinks_types[];
extern const bits      ospf_lsa_types[];
extern lsage_t         ospf_maxAge;
extern lsage_t         ospf_minLSInt;
extern gw_entry        *ospf_defaultGateway;
extern searchdata_t    __searchdata__;
extern ospf_ptwalk_t   *ospf_ptwalk;
extern time_t          ospf_refresh_start;
extern task_job        *ospf_refresh_job;
extern ospf_ptwalk_t   *ospf_refresh_walk;
extern ospf_area_t     *ospf_refresh_area;
extern int	       ospf_refreshChunkSize;
extern lsage_t	       ospf_maxAgeDiff;
extern int	       ospf_key_len[];

vertex_t	*ospf_ls_add __A((db_tree_t *, ospf_lsa_t *));
void		ospf_ls_del __A((db_tree_t *, ospf_lsa_t *));
vertex_t 	*ospf_ls_create __A((db_tree_t *, ospf_lsa_t *, rt_entry *,
			searchdata_t *));
vertex_t	*ls_vtx_create __A((db_tree_t *, lskey_t *, options_t, int,
			lsage_t, lsseq_t, lscksum_t, int,
			lslenght_t, searchdata_t *));
void		ospf_ls_update __A((db_tree_t *, ospf_lsa_t *, vertex_t *));
ospf_vtxLink_t *ls_vtxlink_create __A((ospf_vtxLink_t *, vertex_t *,
			linktype_t, u_int32, flag_t, cost_t));
vertex_t	*ls_reverse_vtx __A((db_tree_t *, u_int32, linktype_t,
			u_int32, cost_t, vertex_t *, ospf_vtxLink_t *));
gw_entry	*vtx_get_gw __A((ospf_area_t *, vertex_t *));
rt_entry	*vtx_get_route __A((ipaddr_t, ipaddr_t, gw_entry *, metric_t,
			u_int32, vertex_t *));
u_int32		pt_shortmask32 __A((u_int32, u_int32, u_int32, u_int32));
vertex_t	*db_find __A((db_tree_t *, int, lskey_t *, searchdata_t *));
void		db_add __A((db_tree_t *, int, lskey_t *, searchdata_t *,
			vertex_t *, flag_t));
void		db_add_direct __A((db_tree_t *, int, lskey_t *, vertex_t *));
vertex_t 	*db_del_fast __A((db_tree_t *, vertex_t *));
void		db_delete __A((vertex_t *));
ospf_vtxLink_t **ls_vtxlink_delete __A((ospf_vtxLink_t **, vertex_t *));
ospf_vtxLink_t *vtxlink_find __A((vertex_t *, vertex_t *));
ospf_vtxLink_t *vtxlink_findwithid __A((linktype_t, lsid_t, vertex_t *));
ospf_vtxLink_t *vtxlink_findwithid_prev __A((ospf_vtxLink_t ***, linktype_t,
			lsid_t, ospf_vtxLink_t **));
int		ospf_check_db __A((ospf_area_t *));
int		rec_check_vtx __A((vertex_t *));
int		check_link __A((vertex_t *, vertex_t *, flag_t));
char		*ls_vtx_name __A((vertex_t *));
ospf_ptwalk_t	*ospf_pt_walkdel __A((ospf_ptwalk_t *));
ospf_ptwalk_t	*ospf_pt_walkinit __A((ospf_area_t *));
void		ospf_pt_walk __A((ospf_ptwalk_t *));
void		ospf_ptwalk_delnode __A((ospf_area_t *, tree_node_t *));
void		ospf_gen_rtrlsa __A((ospf_area_t *));
void		ospf_gen_ntwlsa __A((ospf_intf_t *));
int		ls_cmpage  __A((lsage_t, lsseq_t, lscksum_t,
					lsage_t, lsseq_t, lscksum_t));
void		ospf_check_tree __A((db_tree_t *, int));
void		ospf_test_db __A((void));
void		ospf_spf_schedule __A((void));
void		ospf_spf_compute __A((task_timer *, time_t));
void 		ospf_ngbrmx_flush __A((void));
void		ospf_refresh_schedule __A((task_timer *, time_t));
void		ospf_refresh __A((task_job *));
void		ospf_ls_delete __A((vertex_t *, flag_t));
void		ospf_self_add __A((vertex_t **, ospf_lsa_t *, db_tree_t *,
			rt_entry *));

/* In new_ospf_exch.c
 */
void		ospf_buildDataSum __A((ospf_ngb_t *));
vertex_t	*ospf_global_find __A((db_tree_t *, int, lskey_t *,
			searchdata_t *, db_tree_t *));
void		ospf_ddack __A((ospf_ngb_t *));
void		ospf_ls_needupdate __A((ospf_lsa_t *, ospf_ngb_t *));
void		db_tree_kill __A((db_tree_t *, flag_t));
vertex_t	*ngb_list_next __A((db_tree_t *));
void		ospf_lsa_set __A((ospf_lsa_t *, vertex_t *));
int		ospf_flood __A((ospf_ngb_t *, vertex_t *));
void		ospf_flush_rmx __A((ospf_ngb_t *));
int		ospf_self_originate __A((ospf_ngb_t *,ospf_lsa_t *, vertex_t *,
			searchdata_t *));
int		ospf_lsa_recv __A((ospf_lsa_pck_t *, int, ospf_intf_t *,
			sockaddr_un *));
vertex_t 	*ngb_list_del_rmx __A((ospf_ngb_t *, vertex_t *));
vertex_t	*ngb_list_del_rmx_find __A((ospf_ngb_t *, int, lskey_t *));
int		isflooded __A((ospf_ngb_t *, int, ospf_area_t *));
void		ospf_flash __A((task *, rt_list *));
u_int32		rfc1582apFid __A((rt_head *));
void		ospf_newpolicy __A((task *, rt_list *));

/* In new_ospf_spf.c
 */
void		spf_root __A((ospf_area_t *));
void		psf_all_lssum_route __A((void));
int		spf_inter_route __A((ospf_area_t *, vertex_t *));
void		spf_all_lsext_route __A((void));
void		spf_ext_route __A((ospf_area_t *, vertex_t *));
void		ospf_spf_inc __A((ospf_area_t *, vertex_t *));
void		spf_increment __A((ospf_area_t *, vertex_t *));
vertex_t	*spf_candidate_next __A((ospf_vtxLink_t **, rt_changes **));
void		spf_candidate_update __A((rt_changes *, vertex_t *,
			ospf_vtxLink_t *, cdtList_t *));
cdtList_t	*spf_candidate_find __A((vertex_t *, ospf_vtxLink_t **,
			rt_changes **));
void		spf_next_hop __A((rt_changes *, ospf_area_t *, vertex_t *,
			vertex_t *, vertex_t *, ospf_vtxLink_t *, metric_t));
if_addr		*spf_get_int __A((ospf_area_t *, ospf_vtxLink_t *));
int		spf_update_route __A((rt_changes *, ospf_area_t *, vertex_t *,
			ospf_vtxLink_t *));
ospf_intf_t	*ospf_isendpoint __A((ospf_area_t *, rtid_t));
vertex_t	*ospf_rt_find __A((rtid_t));
void		ospf_gw_update __A((ospf_area_t *, vertex_t *));
void		spf_lssum_gen __A((ospf_area_t *, vertex_t *));
void		ospf_flood_selforiginate __A((ospf_area_t *, vertex_t *));
void		ospf_flood_timer __A((task_timer *, time_t));
void		ospf_flood_job __A((ospf_area_t *, vertex_t *));

/* Patricia tree code. In new_ospf_tree.c.
 */
void		tree_add __A((u_int8 *, int, tree_node_t **, tree_node_t *,
			tree_node_t *, void_t));
tree_node_t	*tree_find __A((u_int8 *, int, tree_node_t *, tree_node_t **));
void		tree_node_del __A((tree_node_t **,  u_int8 *, int, tree_node_t *));
void		tree_delete_rec __A((tree_node_t *, flag_t));
void		ospf_pt_walk __A((ospf_ptwalk_t *));
void_t		tree_del_fast __A((tree_node_t **, u_int8 *, int, void_t));
void		bitdiff __A((bitindex_t *, u_int8 *, int, u_int8 *));

