/*
 * Copyright (c) 1996 The Regents of the University of Michigan
 * All Rights Reserved
 * 
 * License to use, copy, modify, and distribute this software and its
 * documentation can be obtained from Merit at the University of Michigan.
 * 
 * Merit GateDaemon Project
 * 4251 Plymouth Road, Suite C
 * Ann Arbor, MI 48105
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE REGENTS OF THE
 * UNIVERSITY OF MICHIGAN AND MERIT DO NOT WARRANT THAT THE
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the
 * University of Michigan and Merit shall not be liable for
 * any special, indirect, incidental or consequential damages with respect
 * to any claim by Licensee or any third party arising from use of the
 * software. GateDaemon was originated and developed through release 3.0
 * by Cornell University and its collaborators.
 * 
 * Please forward bug fixes, enhancements and questions to the
 * gated mailing list: gated-people@gated.merit.edu.
 * 
 * This copyright has ben automaticly added by the util/addcopyright.pl program.
 * __END_OF_COPYRIGHT__
 */

/* The dijkstra algorithm uses a ordered list of 'candidates'. This list
 * is implemented as a radix tree, which leaves are a list of vertex of
 * the same cost, which ID are the cost.
 */

#include "include.h"
#include "inet.h"
#include "new_ospf.h"
#include "new_ospf_ls.h"

/*
 * typedef struct _vtx_doubleList_t {
 * 	struct _vtx_doubleList_t *dl_prev, *dl_next;
 * 	vertex_t *dl_vtx;
 * } vtx_doubleList_t;
 * 
 * typedef struct _cdt_node_t {                    /* size 20 *
 *         struct _cdt_node_t      *cdt_left;      /* Left sub node *
 *         struct _cdt_node_t      *cdt_right;     /* right sub node *
 * 	struct _cdt_node_t	*cdt_parent;	/* Parent node *
 *         u_int32                 cdt_cost;       /* The node's cost *
 *         u_int32                 cdt_mask;       /* The node's mask *
 *         vtx_doubleList_t        cdt_list;       /* A list of vertices *
 * } cdt_node_t;
 */

cdtList_t *ospf_cdtList = NULL;
#define ISCANDIDATE(vtx) (BIT_TEST((vtx)->vtx_flags, LF_FOUND))
#define CDT_EMPTY(cdlList)	(!(cdlList))
/*
#define CDT_LIST_CHECK(msg) { \
	cdtList_t *__cdt; \
	for(__cdt = ospf_cdtList; __cdt; __cdt = __cdt->l_next) { \
		trace_tf(ospf_trace_options, TR_OSPF_SPF, 0, \
			("*** cdt %s: %p <- %p %p->", msg, __cdt, \
				__cdt->l_prev, __cdt->l_next)); \
		ASSERT(__cdt->l_vtx && \
			(!__cdt->l_next || __cdt->l_next != __cdt->l_prev)); \
	} \
}
 */
#define CDT_LIST_CHECK(msg)


/* SPF algorithm. Compute the Shortest Path Tree for a brain new graph.
 * Follows RFC 1583 p86, #16.1
 * Each vertex has a counter. When doing a new SPF computation, all
 * vertices put in the graph (i.e the shortest path has been computed),
 * its counter is incremented. Therefore a test to check if a vertex is in
 * the graph.
 */
void
spf_root __PF1(area, ospf_area_t *)
{
	vertex_t *nextvtx, *vtx;
	metric_t cost;
	ospf_vtxLink_t *vl;
	cdtList_t *candidate;
	rt_changes rtchg, *rtc;

	ASEXTREE_BUG("spf_root 1");

	/* XXX To be removed. Needed when db_test() is run
	 */
	if (!area->area_graph.gr_root)
		return;
#define SPF_INGRAPH(vtx, area)	((vtx)->vtx_spfcount == (area)->area_spfcount)

	trace_tf(ospf_trace_options, TR_OSPF_SPF, 0,
		("OSPF SPF: Computation for area %A",
		sockbuild_in(0, area->area_id)));

	/* Reset the candidate list. Set SPF to the root node only.
	 * Increment the SPF counter.
	 */
	ASSERT(CDT_EMPTY(ospf_cdtList));

	vtx = area->area_graph.gr_root;
	vtx->vtx_cost = 0;
	vtx->vtx_spfcount = ++area->area_spfcount;
	rtchg.rtc_flags = 0;

	do {
		/* If this is a router, and bit V is set, then set it too
		 * for the area.
		 */
		if (vtx->vtx_type == LST_RTR)
			area->area_rtrbits |= (vtx->vtx_rtr_bits &LSRTR_BITS_V);

		/* Set up the next hops now? XXX ???? We need to know the
		 * parent, and that's not obvious unless we store it with
		 * the vertex.
		 */

		/* Look through all links:
		 */
		for(vl = vtx->vtx_links; vl; vl = vl->vl_next) {

			CDT_LIST_CHECK("spf_root");
			nextvtx = vl->vl_vtx;
			cost = vtx->vtx_cost + vl->vl_cost;

			/* If this is a stub network, update the routing table.
			 * also reset the old cost to infinity.
			 */
			if (vl->vl_type == LSRLT_STUB) {
				(void)bzero((char *)&rtchg, sizeof(rtchg));
				nextvtx->vtx_cost = ospf_lsinfinity;
				spf_next_hop(&rtchg, area, vtx, nextvtx,
					area->area_graph.gr_root, vl, cost);
				spf_update_route(&rtchg, area, nextvtx, NULL);
				continue;
			}

			/* already in the graph vertex, or request onces.
			 */
			if (BIT_TEST(vl->vl_flags, LF_REQUEST) ||
					BIT_TEST(nextvtx->vtx_flags,
						LF_HIDDEN|LF_DELETE)||
					SPF_INGRAPH(nextvtx, area))
				continue;

			/* For hidden, too old vertex, or assymetric, we
			 * set the cost to ospf_lsinfinity to get rid of them
			 * in the routing table.
			 */
			if (BIT_TEST(vl->vl_flags, LF_ASYM|LF_HIDDEN) ||
					VTX_AGE_MAX(nextvtx))
				cost = ospf_lsinfinity;

			/* find out if nextvtx is on the candidate list and
			 * compare the cost.
			 */
			ASSERT(nextvtx->vtx_rt);
			candidate = spf_candidate_find(nextvtx, &vl, &rtc);
			if (candidate && cost > nextvtx->vtx_cost)
				/* We ignore that one.
				 */
				continue;

			/* Get the rt_change structure. Reset the old cost.
			 */
			if (!candidate) {
				RT_CHANGE_ALLOC(rtc);
				(void)bzero((char *)rtc, sizeof(*rtc));
				nextvtx->vtx_cost = ospf_lsinfinity;
			}

			/* If the cost is better, force a metric change.
			 */
			if (!candidate || cost < nextvtx->vtx_cost)
				BIT_SET(rtc->rtc_flags, RTCF_METRIC);

			/* !candidate || cost < nextvtx->vtx_cost
			 * Add the newvtx to the candidate list
			 * (or update its cost), and compute some next
			 * hops. If this guy is not on the list yet,
			 * we force a reset of its cost/nexthops.
			 */
			spf_next_hop(rtc, area, vtx, nextvtx,
				area->area_graph.gr_root, vl, cost);
			spf_candidate_update(rtc, nextvtx, vl, candidate);
		}

		/* If the candidate list is empty, that's the end folks!
		 */
		if (!(vtx = spf_candidate_next(&vl, &rtc)))
			break;

		/* some stuff to do with routing tables and virtual links.
		 */
		spf_update_route(rtc, area, vtx, vl);
		RT_CHANGE_FREE(rtc);
		ASSERT(vtx->vtx_cost);

		/* Add this vertex to the graph
		 */
		vtx->vtx_spfcount = area->area_spfcount;

		ASEXTREE_BUG("spf_root 2");

	} while(1);

        OSPF_TREE_CHECK(&area->area_tree);
}

/* Compute all routes from LS summary advertisements. RFC #16.2
 */
void
psf_all_lssum_route __PF0(void)
{
	int onlybckbone = 0;
	vertex_t *vtx;
	ospf_area_t *area;

	/* If we are attached to multiple area, look at the backbone only.
	 * Don't think so.
	 */
	if (ospf_area_list->area_next)
		onlybckbone = 1;

	OSPF_ALL_AREA(area) {
		OSPF_TREE_CHECK(&area->area_tree);

		/* 
		 * if (onlybckbone && area->area_id != OSPF_AREAID_BACKBONE)
		 *	continue;
		 */

		/* Go through all LS summary advertisements.
		 */
		PT_WALK_TYPE(&area->area_tree, LST_ASIP, vtx) {

			spf_inter_route(area, vtx);

		} PT_WALK_TYPE_END(&area->area_tree, LST_ASIP, vtx);

		/* Same for the AS border summary advertisements.
		 */
                PT_WALK_TYPE(&area->area_tree, LST_ASBR, vtx) {

			spf_inter_route(area, vtx);

                } PT_WALK_TYPE_END(&area->area_tree, LST_ASBR, vtx); 
	}
}

/* Compute the route from any LS summary advertisement. RFC #16.2
 * Return 1 if the route to the destination changed.
 */
int
spf_inter_route __PF2(area, ospf_area_t *, vtx, vertex_t *)
{
	vertex_t *br_vtx;
	int ind;
	rt_changes rtchg;

	/* If the LS summary is from us, we do not need to recompute anything,
	 * the route is already in the routing table.
	 */
	if (vtx->vtx_advrt == ospf_rtid || BIT_TEST(vtx->vtx_flags, LF_REQUEST))
		return(0);

        /* something to do with address ranges (CIDR i guess)
         */
        /* [XXXX] */

        /* Get the cost of the path to the area border router.
         */
        if (!(br_vtx = ospf_rt_find(vtx->vtx_advrt))) {
		trace_tf(ospf_trace_options, TR_OSPF_SPF, 0,
			("OSPF SPF: No route to vertex %s",
			ls_vtx_name(vtx)));
                return(0);
	}

        /* Update the routing entry with the new(?) cost. If the age is
	 * maxage, we need to mark thge route as unreachable.
         */
	(void)bzero((char *)&rtchg, sizeof(rtchg));
	if (VTX_AGE_MAX(vtx))
		rtchg.rtc_metric = ospf_lsinfinity;
	else
		rtchg.rtc_metric = br_vtx->vtx_cost +
				   SLAMETRIC_METRIC(vtx->vtx_tosmet);
	if (rtchg.rtc_metric != vtx->vtx_cost)
		BIT_SET(rtchg.rtc_flags, RTCF_METRIC);

	trace_tf(ospf_trace_options, TR_OSPF_SPF, 0,
		("OSPF SPF: Extra Area Route %s from %s Cost %d set to %d",
			ls_vtx_name(vtx), ls_vtx_name(br_vtx),
			br_vtx->vtx_cost, rtchg.rtc_metric));

	/* And set the next hops equal the the border gateway next hops.
	 */
	for(ind = 0; ind < br_vtx->vtx_nbNextHops; ind++)
		rtchg.rtc_routers[ind] = br_vtx->vtx_nextHops[ind];
	rtchg.rtc_n_gw = br_vtx->vtx_nbNextHops;

        return(spf_update_route(&rtchg, area, vtx, NULL));
}

/* compute all external routes. RFC #16.4
 */
void
spf_all_lsext_route __PF0(void)
{
	vertex_t *vtx;
	ospf_area_t *area;

	/* Go through all external advertisement of all area.
	 */
	OSPF_ALL_AREA(area) {
		PT_WALK_TYPE(&area->area_tree, LST_ASEX, vtx) {

			spf_ext_route(area, vtx);

		} PT_WALK_TYPE_END(&area->area_tree, LST_ASEX, vtx);
	} OSPF_ALL_AREA_END(area);
}

/* compute the route to an external destination. RFC #16.4
 */
void
spf_ext_route __PF2(area, ospf_area_t *, vtx, vertex_t *)
{
	vertex_t *br_vtx;
	int ind;
	rt_changes rtchg;

	/* If the LS external is from us, we do not need to recompute anything,
         * the route is already in the routing table.
         */
        if (vtx->vtx_advrt == ospf_rtid || BIT_TEST(vtx->vtx_flags, LF_REQUEST))
                return;

	/* If too far away, or too old, or not here yet, continue.
         *
        if (BIT_TEST(vtx->vtx_flags, LF_REQUEST) ||
			xxxxxxxxxx(vtx) >= ospf_lsinfinity ||
			VTX_AGE_MAX(vtx))
                return;
	 */

        /* Get the forwarding address
         */
        if (!(br_vtx = ospf_rt_find(vtx->vtx_forward)))
                return;

	/* update the cost... If the age is maxage, mark the route as
	 * unreachable. XXXX? Type of route?
	 */
	(void)bzero((char *)&rtchg, sizeof(rtchg));
	if (VTX_AGE_MAX(vtx))
		rtchg.rtc_metric = ospf_lsinfinity;
	else
		rtchg.rtc_metric = br_vtx->vtx_cost;

	if (rtchg.rtc_metric != vtx->vtx_cost)
		BIT_SET(rtchg.rtc_flags, RTCF_METRIC);

	/* If the gateway is on a directly attached network, the next hop
	 * is meaningless. We need to use the forward address directly.
	 * Same tricky test as in spf_next_hop().
	 * Else the next hops which are inherited from the border gateway.
	 */
	for(ind = 0; ind < br_vtx->vtx_nbNextHops; ind++)
		if (SOCK32(br_vtx->vtx_outIntf[ind]->ifa_addr_local) ==
				SOCK32(br_vtx->vtx_nextHops[ind]))
			rtchg.rtc_routers[ind] = vtx->vtx_forward?
				sockbuild_in(0, vtx->vtx_forward):
				sockbuild_in(0, vtx->vtx_advrt);
		else
			rtchg.rtc_routers[ind] = br_vtx->vtx_nextHops[ind];
	rtchg.rtc_n_gw = br_vtx->vtx_nbNextHops;

        spf_update_route(&rtchg, area, vtx, NULL);
}

/* We got a new LS in the DB. What routes should be updated, and how?
 * Kind of section 16.5 of the RFC, more general.
 */
void
ospf_spf_inc __PF2(area, ospf_area_t *, vtx, vertex_t *)
{
	int change;

	trace_tf(ospf_trace_options, TR_OSPF_SPF, 0,
                ("OSPF SPF: Incremental update area %A LS %s",
				sockbuild_in(0, area->area_id),
				ls_vtx_name(vtx)));
	
	switch(vtx->vtx_type) {
	case LST_RTR:
	case LST_NTW:
		/* We add a new router/network to an area. Run the
		 * incremental SPF algorithm. The RFC says on page 78,
		 * #13.2 that *all* areas routes should be recomputed.
		 * I don't understand why.
		 * Let's do only one area now, and see what happen!
		 */
		ospf_spf_schedule();
		/*spf_increment(area, vtx);*/
		break;

	case LST_ASIP:
	case LST_ASBR:
		/* Let's recompute the route for the inter-area destination.
		 */
		change = spf_inter_route(area, vtx);

		/* If it's a border gateway address, recompute all associated
		 * routes.
		 */
		if (change && vtx->vtx_gw->gw_n_routes)
			ospf_gw_update(area, vtx);
		break;

	case LST_ASEX:
		/* Let recompute the route to the external destination.
		 */
		spf_ext_route(area, vtx);
		break;

	default:
		ASSERT(0);
		break;
	}
}

/* Incremental SPF algorithm. Recompute the Shortest Path Tree when 'adding'
 * a new node. By lpj's own brain. Hope that works....
 *
void
spf_increment __PF2(area, ospf_area_t *, vtx, vertex_t *)
{
	vertex_t *nextvtx;
	metric_t cost;
	ospf_vtxLink_t *vl, *best_vl;
	int change;

	trace_tf(ospf_trace_options, TR_OSPF_SPF, 0,
                ("OSPF SPF: Incremental area %A vertex %s",
			sockbuild_in(0, area->area_id), ls_vtx_name(vtx)));

	ASSERT(CDT_EMPTY(ospf_cdtList));

	/* Get rid of that case now.
	 *
	if (VTX_AGE_MAX(vtx) || BIT_TEST(vtx->vtx_flags, LF_HIDDEN))
		return;

	do {
		/* Look through all links. Find out which one is the
		 * shortest path to the root. but first some stuff as in
		 * spf_root()
		 *
		/* If this is a router, and bit V is set, then set it too
                 * for the area.
                 *
                if (vtx->vtx_type == LST_RTR)
                        area->area_rtrbits |= (vtx->vtx_rtr_bits&LSRTR_BITS_V);

		vtx->vtx_cost = 0;
		best_vl = NULL;
		for(vl = vtx->vtx_links; vl; vl = vl->vl_next) {
                        nextvtx = vl->vl_vtx;

			/* Get rid of unusable links.
			 *
			if (vl->vl_type == LSRLT_STUB ||
				BIT_TEST(vl->vl_flags, LF_ASYM|LF_HIDDEN) ||
				VTX_AGE_MAX(vtx) || ISCANDIDATE(nextvtx))
				continue;

			/* if the cost is lower, that's the best link.
			 *
			cost = vl->vl_cost + vl->vl_vtx->vtx_cost;
			if (cost <= vtx->vtx_cost) {
				spf_next_hop(area, vl->vl_vtx, vtx,
					area->area_graph.gr_root, vl, cost);
				best_vl = vl;
			}
		}

		/* If we do not have a best link now, that means either
		 * vtx is the root, or there is a problem with the graph
		 * (lonely node?)
		 *
		if (!best_vl && vtx != area->area_graph.gr_root) {
			/* Well, it's dead meat. Give up
			 *
			trace_log_tp(ospf_task, 0, LOG_ERR,
				("OSPF DB: Lonely graph vertex %s",
				ls_vtx_name(vtx)));
			return;
		}

		/* We should have the best link by now. We add vtx to the graph,
		 * and check all link but best_vl for changes.
		 *
		change = spf_update_route(area, vtx, NULL);
		if (change && vtx->vtx_gw->gw_n_routes)
			ospf_gw_update(area, vtx);

		for(vl = vtx->vtx_links; vl; vl = vl->vl_next) {
			nextvtx = vl->vl_vtx;
			cost = vtx->vtx_cost + vl->vl_cost;

			if (vl == best_vl ||
				BIT_TEST(vl->vl_flags, LF_ASYM|LF_HIDDEN) ||
				VTX_AGE_MAX(vtx))
				continue;

			/* Stub networks. Compute the route now.
			 *
			if (vl->vl_type == LSRLT_STUB) {
				if (cost <= nextvtx->vtx_cost) {
					spf_next_hop(area, vtx, nextvtx,
						area->area_graph.gr_root, vl,
						cost);
					spf_update_route(area, vtx, NULL);
				}
				continue;
			}

			/* If the cost changed, put it in the candidate list.
			 *
			if (cost <= nextvtx->vtx_cost) {
				spf_next_hop(area, vtx, nextvtx,
					area->area_graph.gr_root, vl, cost);
				spf_candidate_update(nextvtx, vl);
			}
		}

		/* Get the next candidate.
		 *
		if (!(vtx = spf_candidate_next(&vl)))
			break;
	} while(1);
}
*/

/* Give the next vertex on the candidate list.
 */
vertex_t *
spf_candidate_next __PF2(vl, ospf_vtxLink_t **, rtc, rt_changes **)
{
	cdtList_t *cdt;
	vertex_t *vtx;

	CDT_LIST_CHECK("spf_candidate_next");

	if (!ospf_cdtList)
		return(NULL);

	cdt = ospf_cdtList;
	vtx = cdt->l_vtx;
	*vl = cdt->l_vl;
	*rtc = cdt->l_rtchg;
	ospf_cdtList = ospf_cdtList->l_next;
	if (ospf_cdtList)
		ospf_cdtList->l_prev = NULL;
	task_block_free(ospf_cdtList_block, cdt);

	BIT_RESET(vtx->vtx_flags, LF_FOUND);

	CDT_LIST_CHECK("spf_candidate_next 2");
	return(vtx);
}

#define VTXCOST_CMP(vtx1, vtx2)						\
	(((vtx1)->vtx_cost < (vtx2)->vtx_cost)? -1:			\
		((vtx1)->vtx_cost > (vtx2)->vtx_cost)? 1:		\
			((vtx2)->vtx_type == LST_NTW)? 1: 0)

/* Update the cost of a vertex on the candidate list. If oldCdt is set,
 * it point to the old vtx's position on the list.
 */
void
spf_candidate_update __PF4(rtc, rt_changes *, nextvtx, vertex_t *,
		vl, ospf_vtxLink_t *, oldCdt, cdtList_t *)
{
	cdtList_t *newcdt = NULL, *cdt;

	CDT_LIST_CHECK("spf_candidate_update");

	BIT_SET(nextvtx->vtx_flags, LF_FOUND);

	/* If oldcdt is set, remove it.
	 */
	if (oldCdt) {
		/* Well, not if the cost have not changed.
		 */
		if (!BIT_TEST(rtc->rtc_flags, RTCF_METRIC)) {
			oldCdt->l_vl = vl;
			ASSERT(oldCdt->l_vtx);
			CDT_LIST_CHECK("spf_candidate_update 2");
			return;
		}

		if (oldCdt->l_prev)
			oldCdt->l_prev->l_next = oldCdt->l_next;
		else
			ospf_cdtList = oldCdt->l_next;

		if (oldCdt->l_next)
			oldCdt->l_next->l_prev = oldCdt->l_prev;

		newcdt = oldCdt;
		ASSERT(newcdt->l_vtx && newcdt->l_vtx == nextvtx);
	}
	else {
		/* The vertex is not on the list. We create a new one.
		 */
		newcdt = (cdtList_t *)task_block_alloc(ospf_cdtList_block);
		newcdt->l_vtx = nextvtx;
		ASSERT(newcdt->l_vtx);
	}
	newcdt->l_vl = vl;
	newcdt->l_rtchg = rtc;

	/* The list is empty. 
	 */
	if (!ospf_cdtList) {
		ospf_cdtList = newcdt;
		newcdt->l_next = NULL;
		newcdt->l_prev = NULL;
		CDT_LIST_CHECK("spf_candidate_update 3");
		return;
	}
	/* We insert it at the begining
	 */
	if (VTXCOST_CMP(nextvtx, ospf_cdtList->l_vtx) <= 0) {
		newcdt->l_next = ospf_cdtList;
		newcdt->l_prev = NULL;
		if (ospf_cdtList)
			ospf_cdtList->l_prev = newcdt;
		ospf_cdtList = newcdt;
		CDT_LIST_CHECK("spf_candidate_update 4");
		return;
	}

	/* We insert it anywhere... (just after 'cdt')
	 */
	for(cdt = ospf_cdtList; cdt->l_next &&
			VTXCOST_CMP(nextvtx, cdt->l_next->l_vtx) > 0;
			cdt = cdt->l_next);

	newcdt->l_next = cdt->l_next;
	newcdt->l_prev = cdt;
	if (newcdt->l_next)
		newcdt->l_next->l_prev = newcdt;
	cdt->l_next = newcdt;

	CDT_LIST_CHECK("spf_candidate_update 5");
}

/* Find a vertex in the candidate list.
 */
cdtList_t *
spf_candidate_find __PF3(vtx, vertex_t *, vl, ospf_vtxLink_t **,
		rtc, rt_changes **)
{
	cdtList_t *cdt;

	CDT_LIST_CHECK("spf_candidate_find");

	for(cdt = ospf_cdtList; cdt; cdt = cdt->l_next)
		if (cdt->l_vtx == vtx) {
			*vl = cdt->l_vl;
			*rtc = cdt->l_rtchg;
			return(cdt);
		}
	return(NULL);
}

/* Compute the next hop to 'vtx', parent is parent_vtx. vl is the link from
 * the root (router) to vtx.
 */
void
spf_next_hop __PF7(rtc, rt_changes *, area, ospf_area_t *,
		parent_vtx, vertex_t *,
		vtx, vertex_t *, root, vertex_t *,
		vl, ospf_vtxLink_t *, cost, metric_t)
{
	int ind;
	if_addr *ifap;

#define DEBUG_NEXTHOP(str) do {	\
	trace_tf(ospf_trace_options, TR_OSPF_SPF, 0, \
("OSPF SPF: Nexthop %s: from %s to %s, cost changed from %d to %d, #hops changed from %d to %d", \
	(str), ls_vtx_name(parent_vtx), ls_vtx_name(vtx), vtx->vtx_cost, \
	cost, parent_vtx->vtx_nbNextHops, rtc->rtc_n_gw)); \
	trace_tf(ospf_trace_options, TR_OSPF_SPF, 0, \
("OSPF SPF: Nexthop %s: Link is %A/%A, Next hop parent %A, vertex %A, Interface parent %A, vertex %A", \
	(str), sockbuild_in(0, vl->vl_id), sockbuild_in(0, vl->vl_data), \
	parent_vtx->vtx_nextHops[0]? parent_vtx->vtx_nextHops[0]: \
		sockbuild_str("---"), \
	rtc->rtc_routers[0]? rtc->rtc_routers[0]: \
		sockbuild_str("---"), \
	parent_vtx->vtx_outIntf[0]? parent_vtx->vtx_outIntf[0]->ifa_addr: \
		sockbuild_str("---"), \
	rtc->rtc_ifaps[0]? rtc->rtc_ifaps[0]->ifa_addr: \
		sockbuild_str("---"))); \
} while(0);

	/* If we are root, that does not make sense?
	 */
	if (vtx == root) {
		DEBUG_NEXTHOP("ROOT");
		return;
	}

	/* If the cost is better, reset the next hop list
	 */
	rtc->rtc_metric = cost;
	if (cost < vtx->vtx_cost) {
		rtc->rtc_n_gw = 0;
		BIT_SET(rtc->rtc_flags, RTCF_METRIC);
	}

	/* If we have too many next hops, just ignore the next ones.
	 */
	if (rtc->rtc_n_gw == RT_N_MULTIPATH) {
		DEBUG_NEXTHOP("FULL");
		return;
	}

	/* Not really like the RFC. Will see if it works...
	 * First case is a router/network connected to the root.
	 * for an interface route, it seem that the next hop is set to
	 * the interface address itself. Well, according to if.c:if_rtup().
	 */
	if (parent_vtx == root) {
		/* No next hop. Just an interface.
                 */
                ifap = spf_get_int(area, vl);
		if (!ifap) {
			/* That's bad. Usualy it's a bug. But in case it is
			 * not our fault, let do with it. (There should be
			 * an 'ASSERT(ifap)' here!)
			 */
			trace_log_tp(ospf_task, 0, LOG_ERR,
			("OSPF SPF: no interface %A on area %A, to object %s",
				sockbuild_in(0, vl->vl_data),
				sockbuild_in(0, area->area_id),
				ls_vtx_name(vtx)));
			return;
		}

		/* A new next hop...
		 */
		rtc->rtc_ifaps[rtc->rtc_n_gw] = ifap;
		rtc->rtc_routers[rtc->rtc_n_gw] = ifap->ifa_addr;
		rtc->rtc_n_gw++;

		/* XXX We may want to set up a real next hop if vtx is a router
		 * and the p2p link has an address.
		 */
		DEBUG_NEXTHOP("INTF");
                return;
        }

	/* Second case, we copy the next hop/interfaces of the parent.
	 */
	/*ASSERT(parent_vtx->vtx_nbNextHops); */
	for(ind = 0; ind < parent_vtx->vtx_nbNextHops &&
				rtc->rtc_n_gw < RT_N_MULTIPATH;
				ind++, rtc->rtc_n_gw++) {
		/* We need to copy both next hop router or
		 * outbound interfaces (in case of a chain of
		 * unumbered p2p links. XXX. GateD Needs a next hop!!!!
		 *
		 * Case of a dirctly connected (to the root) network. In this
		 * case we create a real next hop.
		 */
		if (vtx->vtx_type == LST_RTR &&
			parent_vtx->vtx_type == LST_NTW &&
			SOCK32(parent_vtx->vtx_outIntf[ind]->ifa_addr_local) ==
				SOCK32(parent_vtx->vtx_nextHops[ind])) {
			rtc->rtc_routers[rtc->rtc_n_gw] =
				sockbuild_in(0, vl->vl_data);
			rtc->rtc_ifaps[rtc->rtc_n_gw] =
				parent_vtx->vtx_outIntf[ind];
		}

		/* Just straighforward inheritence.
		 */
		else  /*if (!vtx->vtx_nextHops[vtx->vtx_nbNextHops] ||
			SOCK32(vtx->vtx_nextHops[vtx->vtx_nbNextHops]) !=
				SOCK32(parent_vtx->vtx_nextHops[ind])) */ {
			rtc->rtc_routers[rtc->rtc_n_gw] =
					parent_vtx->vtx_nextHops[ind];
			rtc->rtc_ifaps[rtc->rtc_n_gw] =
					parent_vtx->vtx_outIntf[ind];
		}
	}
	DEBUG_NEXTHOP("PARENT");
	return;
}

/* Get the interface (of the root) for a given link
 */
if_addr *
spf_get_int __PF2(area, ospf_area_t *, vl, ospf_vtxLink_t *)
{
	ospf_intf_t *ospf_intf;

	OSPF_IF_ADDR_AREA(area, ospf_intf) {
		/* Get rid of sick interfaces...
		 */
		if (ospf_intf->ospf_if_state <= OSPF_IFS_DOWN)
			continue;

		switch(vl->vl_type) {
		case LSRLT_P2P:
			/* Unumbered p2p link used the interface index as data.
			 */
			if (!ospf_intf->ospf_if_addr) {
				if (ospf_intf->ospf_if_ifap->ifa_link->ifl_index
					== vl->vl_data)
					return(ospf_intf->ospf_if_ifap);
				break;
			}
			if (ospf_intf->ospf_if_addr == vl->vl_data)
				return(ospf_intf->ospf_if_ifap);
			break;
		case LSRLT_VLINK:
			if (ospf_intf->ospf_if_addr == vl->vl_data)
				return(ospf_intf->ospf_if_ifap);
			break;
		case LSRLT_TRANS:
			/* It's a link between a router and a network. Data
			 * is the interface IP address.
			 */
			if ((ospf_intf->ospf_if_addr & ospf_intf->ospf_if_mask)
				== (vl->vl_data & ospf_intf->ospf_if_mask))
				return(ospf_intf->ospf_if_ifap);
			break;
		case LSRLT_STUB:
			/* The data is the mask. ID is IP address.
			 */
			if (ospf_intf->ospf_if_mask == vl->vl_data &&
			  (ospf_intf->ospf_if_addr & ospf_intf->ospf_if_mask) ==
				vl->vl_id)
				return(ospf_intf->ospf_if_ifap);
			break;
		case LSRLT_RTR:
			/* This case should never happen.
			 */
		default:
			ASSERT(0);
		}
	} OSPF_IF_ADDR_AREA_END(area, ospf_intf);

	return(NULL);
}

/* Update the routing table. Create a entry in the gated routing table for
 * that 'announcement'. Find out with OSPF route is the best. Update the
 * active route accordingly. Return 1 if the cost or next hop changed.
 */
int
spf_update_route __PF4(rtc, rt_changes *, area, ospf_area_t *,
		vtx, vertex_t *, vl, ospf_vtxLink_t *)
{
	int change;
	ospf_intf_t *ospf_intf;
	pref_t preference;

	ASEXTREE_BUG("spf_update_route 1");

	/* don't need to set a route to the root (ourself)
	 */
	if (vtx == area->area_graph.gr_root) {
		ASEXTREE_BUG("spf_update_route 1.1");
		return(0);
	}

	/* The tag change bit as been set in ospf_ls_update() a long time ago...
	 * And the tag should not change...
	 */
	rtc->rtc_flags |= vtx->vtx_changes;
	rtc->rtc_tag = vtx->vtx_rt->rt_tag;

	/* If the route was not active, and is reachable, and is not a host
	 * route (unless LF_ROUTE is set) activate it!
	 */
	if (vtx->vtx_rt->rt_preference < 0 &&
			rtc->rtc_metric < ospf_lsinfinity &&
			rtc->rtc_n_gw &&
			(vtx->vtx_type != LST_RTR ||
				BIT_TEST(vtx->vtx_flags, LF_ROUTE))) {
		preference = ospf_preference;
		BIT_SET(rtc->rtc_flags, RTCF_METRIC);
	}
	/* And the contrary if the route is unreachable.
	 */
	else if ((!rtc->rtc_n_gw || rtc->rtc_metric >= ospf_lsinfinity ||
			(vtx->vtx_type == LST_RTR &&
				!BIT_TEST(vtx->vtx_flags, LF_ROUTE))) &&
			vtx->vtx_rt->rt_preference >= 0) {
		preference = -1;
		BIT_SET(rtc->rtc_flags, RTCF_METRIC);
	}
	else
		preference = vtx->vtx_rt->rt_preference;

	/*
	if (BIT_TEST(vtx->vtx_rt->rt_state,
			RTS_HIDDEN|RTS_NOTINSTALL|RTS_NOADVISE)) {
		BIT_SET(rtc->rtc_flags, RTCF_METRIC);
		BIT_RESET(vtx->vtx_rt->rt_state,
			RTS_HIDDEN|RTS_NOTINSTALL|RTS_NOADVISE);
	}*/

	/* Make sure something changed.
	 */
	if (!rtc->rtc_flags)
		return(0);

	/* A trick. If RTCF_METRIC is set, we have to force gated to change
	 * the route, even if the old cost is still the same. So we
	 * just change the old cost and make sure it's different from the
	 * new one!
	 */
	if (rtc->rtc_metric == vtx->vtx_cost &&
			BIT_TEST(rtc->rtc_flags, RTCF_METRIC))
		vtx->vtx_cost++;

	/* Update the routing table. XXX For now, uses rt_change()
	 * rt_change() will compare the current next hop with the new
	 * ones and won't do anything if they are the same.
	 */
	ASSERT(vtx == OSPF_RT_VTX(vtx->vtx_rt));
	ASSERT(vtx->vtx_cost);
	rt_change(vtx->vtx_rt, rtc->rtc_metric, rtc->rtc_metric2, rtc->rtc_tag,
		preference, 0, rtc->rtc_n_gw, rtc->rtc_routers);
	ASSERT(vtx->vtx_rt->rt_n_gw ||
		!BIT_TEST(vtx->vtx_rt->rt_state, RTS_ACTIVE));
	ASSERT(vtx == OSPF_RT_VTX(vtx->vtx_rt));
	ASSERT(vtx->vtx_cost);

	/* Now create a LS Summary if needed. Only border routers do that.
	 */
	if (BIT_TEST(ospf_rtrbits, LSRTR_BITS_B) &&
			BIT_TEST(rtc->rtc_flags, RTCF_METRIC))
		spf_lssum_gen(area, vtx);

	change = rtc->rtc_flags;
	vtx->vtx_changes = 0;
	rtc->rtc_flags = 0;

	/* Virtual interface management. If the destination is the end
	 * point of a virtual interface, and the interface was DOWN, generate
	 * an interface UP event. The oposite if the destination is unreachable.
	 */
	if (vtx->vtx_type == LST_RTR &&
			(ospf_intf = ospf_isendpoint(area, vtx->vtx_id))) {
		if (ospf_intf->ospf_if_state == OSPF_IFS_DOWN &&
				vtx->vtx_cost < ospf_lsinfinity) {
			/* New UP interface. Set the physical interface to
			 * the outgoing interface for the destination.
			 */
			ASSERT(vtx->vtx_outIntf[0]);
			ospf_intf->ospf_if_ifap = vtx->vtx_outIntf[0];
			if (!ospf_intf->ospf_if_ngb)
				(void)ospf_new_ngb(ospf_intf,
					ospf_intf->ospf_if_addr,
					ospf_intf->ospf_if_rtpri, vl->vl_data,
					0, 0, OSPF_NGBS_DOWN);
			else
				ospf_intf->ospf_if_ngb->ngb_addr = vl->vl_data;

			ospf_if_event(EV_IF_UP, ospf_intf);
		}
		else if (ospf_intf->ospf_if_state == OSPF_IFS_P2P &&
				vtx->vtx_cost >= ospf_lsinfinity) {
			/* New DOWN interface.
			 */
			ospf_if_event(EV_IF_DOWN, ospf_intf);
		}
	}

	ASEXTREE_BUG("spf_update_route 2");

	return(change);
}

/* Find out if 'rtid' is the endpoint of a virtual link.
 */
ospf_intf_t *
ospf_isendpoint __PF2(area, ospf_area_t *, rtid, rtid_t)
{
	ospf_intf_t *ospf_intf;

	OSPF_IF_ADDR_AREA(area, ospf_intf) {
		if (OSPF_ISVIRTUAL(ospf_intf) && ospf_intf->ospf_if_addr== rtid)
			return(ospf_intf);
	} OSPF_IF_ADDR_AREA_END(area, ospf_intf);

	return(NULL);
}

/* Find an OSPF route [to a border router] in the routing table.
 */
vertex_t *
ospf_rt_find __PF1(rtid, rtid_t)
{
	rt_entry *rt;

	rt = rt_lookup(RTS_NETROUTE, 0, sockbuild_in(0, rtid),
		RTPROTO_BIT(RTPROTO_OSPF));

	ASSERT(!rt || (OSPF_RT_VTX(rt) && OSPF_RT_VTX(rt)->vtx_rt));

	return(rt? OSPF_RT_VTX(rt): NULL);
}

/* update all the routes owned by a certain gateway.
 */
void
ospf_gw_update __PF2(area, ospf_area_t *, vtx, vertex_t *)
{
	rt_entry *rt;
	vertex_t *rtvtx;
	metric_t cost;
	int ind;
	rt_changes rtchg;

	/* Go through all the routes owned by the gateway.
	 */
	RTQ_LIST(&vtx->vtx_gw->gw_rtq, rt) {

		rtvtx = OSPF_RT_VTX(rt);
		rtchg.rtc_flags = 0;

		/* Update the cost
		 */
		if (rtvtx->vtx_type == LST_ASIP ||
				rtvtx->vtx_type == LST_ASBR)
			cost = vtx->vtx_cost +
				SLAMETRIC_METRIC(rtvtx->vtx_tosmet);
		else if (rtvtx->vtx_type == LST_ASEX)
			cost = vtx->vtx_cost;
		else
			/* Only summaries and external routes should be in
			 * a gateway list.
			 */
			ASSERT(0);

		if (cost != rtvtx->vtx_cost) {
			rtchg.rtc_metric = cost;
			BIT_SET(rtchg.rtc_flags, RTCF_METRIC);
		}

		/* Set the next hops
		 */
		for(ind = 0; ind < vtx->vtx_nbNextHops; ind++)
			rtchg.rtc_routers[ind] =
				vtx->vtx_nextHops[ind];
		rtchg.rtc_n_gw = vtx->vtx_nbNextHops;

		spf_update_route(&rtchg, area, rtvtx, NULL);

	} RTQ_LIST_END(&vtx->vtx_gw->gw_rtq, rt);
}

/* Create a LS Summary for an intra area path. Should follows RFC 1583, #12.4.3
 * p72, but it does not seem optimal. So let try something close.... :)
 */
void
spf_lssum_gen __PF2(area, ospf_area_t *, vtx, vertex_t *)
{
	ospf_lsa_t lsa;
	vertex_t *sumvtx;
	searchdata_t searchdata;

	/* LS Summary are created only for NTW vertex, and RTR vertex
	 * with the LSRTR_BITS_E bit (external AS border) set.
	 */
	if (vtx->vtx_type != LST_NTW &&
			(vtx->vtx_type != LST_RTR ||
			!BIT_TEST(vtx->vtx_rtr_bits, LSRTR_BITS_E)))
		return;

	/* Create the LSA, install it in the DB. XXX Disable the checksum.
	 */
	LSA_SL_SET((ospf_lsa_sla_t *)&lsa,
		(vtx->vtx_type == LST_NTW)? vtx->vtx_mask: HOSTMASK,
		0, vtx->vtx_cost);
	LSA_SET_H(&lsa, 0, area->area_options,
		(vtx->vtx_type == LST_NTW)? LST_ASIP: LST_ASBR,
		(vtx->vtx_type == LST_NTW)? SOCK32(vtx->vtx_rt->rt_dest):
			vtx->vtx_id,
		ospf_rtid, RES_LS_SEQ, 0, sizeof(ospf_lsa_sla_t));

	/* If the vertex is unreachable, get rid of it.
	 */
	if (vtx->vtx_cost >= ospf_lsinfinity) {
		if ((sumvtx = ospf_global_find(&area->area_tree, lsa.lsa_type,
				&lsa.lsa_key, &searchdata, &area->area_tree))) {
			VTX_AGE_SET(sumvtx, ospf_maxAge);
			BIT_SET(sumvtx->vtx_flags, LF_DELETE);
			ospf_flood_selforiginate(area, sumvtx);
		}
		return;
	}

	/* 
	/* Update the self originated LS!
	 */
	sumvtx = NULL;
	ospf_self_add(&sumvtx, &lsa, &area->area_tree, NULL);

	/* Set the cost to something. It's useless 'cause the route is
	 * set for the corresponding RTR/MTW. The AIP will have a hidden route.
	 */
	sumvtx->vtx_cost = vtx->vtx_cost;
}

#define TIMERFLOOD_QUEUE_ADD(queue, vtx) do { \
	vtxList_t *__new; \
	__new = (vtxList_t *)task_block_alloc(ospf_vtxlist_block); \
	__new->l_next = (queue); \
	__new->l_vtx = (vtx); \
	(queue) = __new; \
} while(0)
#define TIMERFLOOD_QUEUE_NEXT(queue, vtx) do { \
	 if (!(queue)) \
		(vtx) = NULL; \
	else { \
	 	vtxList_t *__deadQ; \
	 	__deadQ = (queue); \
	 	(queue) = (queue)->l_next; \
		(vtx) = __deadQ->l_vtx; \
		task_block_free(ospf_vtxlist_block, __deadQ); \
	} \
} while(0)

/* Flood a self originated LS to the neighbours. Make sure 2 announcements
 * are separated by ospf_minLSInt second at least.
 */
void
ospf_flood_selforiginate __PF2(area, ospf_area_t *, vtx, vertex_t *)
{
	time_t nextSlot;

	/* If the LS is scheduled for flooding, nothing to do!
	 */
	if (BIT_TEST(vtx->vtx_flags, LF_FLUSH)) {
		ASSERT(!BIT_TEST(ospf_floodTimer1->task_timer_flags,
				TIMERF_INACTIVE) ||
			!BIT_TEST(ospf_floodTimer2->task_timer_flags,
				TIMERF_INACTIVE));
		return;
	}

	/* If the LS can be announced now, (because the last
	 * announcement was more than ospf_minLSInt second ago),
	 * flood it now.
	 */
	if (vtx->vtx_nextRecv <  time_sec) {
		ospf_flood_job(area, vtx);
		return;
	}

	/* Schedule a timer to flood the LSA later on. The timers are set
	 * every ospf_minLSInt second. We can have up to 2 timers running
	 * at a given time. timer1 is the earliest, then timer2. 
	 */
	nextSlot = (time_sec / ospf_minLSInt + 2) * ospf_minLSInt;
	ASSERT(nextSlot > time_sec);
	if (BIT_TEST(ospf_floodTimer1->task_timer_flags, TIMERF_INACTIVE)) {
		/* Activate the first timer for the correct slot
		 */
		task_timer_set(ospf_floodTimer1, 0, nextSlot - time_sec);
		TIMERFLOOD_QUEUE_ADD(ospf_floodQueue1, vtx);
		BIT_SET(vtx->vtx_flags, LF_FLUSH);
		return;
	}
	else if (ospf_floodTimer1->task_timer_next_time >= nextSlot) {
		/* We can use that timer. Just add the vtx to the queue.
		 */
		TIMERFLOOD_QUEUE_ADD(ospf_floodQueue1, vtx);
		BIT_SET(vtx->vtx_flags, LF_FLUSH);
		return;
	}
	else if (BIT_TEST(ospf_floodTimer2->task_timer_flags,TIMERF_INACTIVE)) {
		/* Activate the second timer.
		 */
		task_timer_set(ospf_floodTimer2, 0, nextSlot - time_sec);
		TIMERFLOOD_QUEUE_ADD(ospf_floodQueue2, vtx);
		BIT_SET(vtx->vtx_flags, LF_FLUSH);
		return;
	}
	else {
		/* The second timer activation time should be ok.
		 */
		ASSERT(ospf_floodTimer2->task_timer_next_time >= nextSlot);
		ASSERT(!BIT_TEST(ospf_floodTimer2->task_timer_flags,
			TIMERF_INACTIVE));

		/* Just add the vtx to the queue.
		 */
		TIMERFLOOD_QUEUE_ADD(ospf_floodQueue2, vtx);
		BIT_SET(vtx->vtx_flags, LF_FLUSH);
		return;
	}
}

/* The job of flooding all the LSA in the flodding queue.
 */
void
ospf_flood_timer __PF2(ti, task_timer *, t, time_t)
{
	vtxList_t *queue;
	vertex_t *vtx;

	/* If the are the first timer, move the second timer to the first
	 * so the second will become the more recent!!!!
	 */
	if (ti == ospf_floodTimer1) {
		ospf_floodTimer1 = ospf_floodTimer2;
		ospf_floodTimer2 = ti;
		queue = ospf_floodQueue1;
		ospf_floodQueue1 = ospf_floodQueue2;
		ospf_floodQueue2 = queue;
	}

	/* Take all vertices from the queue, and flood them!
	 */
	TIMERFLOOD_QUEUE_NEXT(ospf_floodQueue2, vtx);
	while(vtx) {
		BIT_RESET(vtx->vtx_flags, LF_FLUSH);
		ospf_flood_job(VTX_AREA(vtx), vtx);

		TIMERFLOOD_QUEUE_NEXT(ospf_floodQueue2, vtx);
	}

	/* Flush the neighbour's queues
	 */
	ospf_ngbrmx_flush();
}

/* Put the LSA on the neighbour's transmission list, for flooding.
 * From ospf_flood().
 */
void
ospf_flood_job __PF2(area, ospf_area_t *, vtx, vertex_t *)
{
	ospf_intf_t *ospf_intf;
	ospf_ngb_t *ngb;
	vertex_t *req;
	int cmp, nbNgb = 0;

	/* Go trhough all area, all interfaces, all neighbours.
	 */
	OSPF_IF_ADDR(ospf_intf) {
		OSPF_INTF_NGB(ngb, ospf_intf) {

			/* Is the neighbour OK to recive it?
			 */
			if (ngb->ngb_state < OSPF_NGBS_EXCHANGE)
				continue;

			/* Do we want to send that LS to that ngb?
			 */
			if (!isflooded(ngb, vtx->vtx_type, area))
				continue;

			/* The adjacency is not full. Check the LS 
			 * request list. Self originated LS should not be
			 * on the LS request list!!!
			 */
			if (ngb->ngb_state < OSPF_NGBS_FULL &&
				(NGB_LIST_FIND_LSR(req, ngb, vtx))) {
				cmp = ls_cmpage(VTX_AGE(req), req->vtx_seq,
					req->vtx_cksum, VTX_AGE(vtx),
					vtx->vtx_seq, vtx->vtx_cksum);
				/* It's less recent?
				 */
				if (cmp > 0)
					continue;

				/* The same? Remove it from the list.
                                 */
                                if (!cmp) {
                                        NGB_LIST_DEL_LSR(ngb, vtx);
                                        continue;
                                }

                                /* The new LSA is more recent.
                                 */
                                NGB_LIST_DEL_LSR(ngb, vtx);
                        }
			/* Add the LSA to the retransmission list.
                         */
			nbNgb++;
                        NGB_LIST_ADD_RMX(ngb, vtx->vtx_type, &vtx->vtx_key,vtx);
			NGB_FLUSH_ADD(ngb);
		}

	} OSPF_IF_ADDR_END(ospf_intf);

	ASSERT(nbNgb < OSPF_MAX_NGB);

	/* If the LF_DELETE bit is set, set the vtx_qref field. It's
	 * used to physicly remove the vertex when all the neighbour
	 * have acknowledged the LSU.
	 */
	if (BIT_TEST(vtx->vtx_flags, LF_DELETE))
		vtx->vtx_qref = nbNgb;

	/* Set the next time we are allowed to announce the LS again.
	 */
	if (nbNgb)
		vtx->vtx_nextRecv = time_sec + ospf_minLSInt;

	trace_tf(ospf_trace_options, TR_OSPF_FLOOD, 0,
		("OSPF FLOOD: Flooded %s for area %A to %d neighbours",
		ls_vtx_name(vtx), sockbuild_in(0, area->area_id), nbNgb));
}






