/*
 * Copyright (c) 1996 The Regents of the University of Michigan
 * All Rights Reserved
 * 
 * License to use, copy, modify, and distribute this software and its
 * documentation can be obtained from Merit at the University of Michigan.
 * 
 * Merit GateDaemon Project
 * 4251 Plymouth Road, Suite C
 * Ann Arbor, MI 48105
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE REGENTS OF THE
 * UNIVERSITY OF MICHIGAN AND MERIT DO NOT WARRANT THAT THE
 * FUNCTIONS CONTAINED IN THE SOFTWARE WILL MEET LICENSEE'S REQUIREMENTS OR
 * THAT OPERATION WILL BE UNINTERRUPTED OR ERROR FREE. The Regents of the
 * University of Michigan and Merit shall not be liable for
 * any special, indirect, incidental or consequential damages with respect
 * to any claim by Licensee or any third party arising from use of the
 * software. GateDaemon was originated and developed through release 3.0
 * by Cornell University and its collaborators.
 * 
 * Please forward bug fixes, enhancements and questions to the
 * gated mailing list: gated-people@gated.merit.edu.
 * 
 * This copyright has ben automaticly added by the util/addcopyright.pl program.
 * __END_OF_COPYRIGHT__
 */

/* OSPF by lpj
 */
#include "include.h"
#include "inet.h"
#include "new_ospf.h"
#include "new_ospf_ls.h"

/* Nice trace of an OSPF packet. Packets are in network order. The packet
 * may be screwed up (we check the packet afterward).
 */
void
ospf_trace_packet __PF6(pck, ospf_pck_t *, pcklen, int, err, int, flag, flag_t,
                peer, sockaddr_un *, from, sockaddr_un *)
{
        ospf_hello_pck_t *pck_hello;
        ospf_dd_pck_t *pck_dd;
        int MaxNgb, nbNgb, nblsh, nblsr, nblsa, i;
        ospf_lsa_hdr_t *lsh;
        ospf_lsa_t *lsa;
        ospf_lsr_data_t *lsr;
 
        /* Do we want to trace?
         */
        if (!TRACE_TF(ospf_trace_options, flag))
                return;
 
        trace_only_tf(ospf_trace_options, 0,
        ("OSPF %s: %A -> %A: Version %d, Type %s (%d), Lenght %d",
                (flag & TR_PACKET_SEND)? "SEND": "RECV",
                (flag & TR_PACKET_SEND)? from: peer,
                (flag & TR_PACKET_SEND)? peer: from,
                pck->pck_version,
                trace_value(ospf_pck_types, pck->pck_type),
                pck->pck_type,
                ntohs(pck->pck_len)));
	trace_only_tf(ospf_trace_options, 0,
		("\tRouter ID %A, Area %A, Authentication %d",
                sockbuild_in(0, pck->pck_rtid),
                sockbuild_in(0, pck->pck_area),
                ntohs(pck->pck_authtype)));
 
        /* Trace all kind of packets!
         */
        switch(pck->pck_type) {
        case OSPF_PCK_HELLO:
                if (!TRACE_TF(ospf_trace_options,
                                TR_OSPF_PCK_ALL|TR_OSPF_PCK_HELLO))
                        break;
                pck_hello = (ospf_hello_pck_t *)pck;
                MaxNgb = NB_HELLO_NGB(ntohs(pck->pck_len));
                trace_only_tf(ospf_trace_options, 0,
                ("\tMask %A, Options <%s> (%X), Priority %d, Neighbours %d",
                        sockbuild_in(0, pck_hello->pck_hello_mask),
                        trace_bits(ospf_pck_options,
                                pck_hello->pck_hello_options),
                        pck_hello->pck_hello_options,
                        pck_hello->pck_hello_rtpri, MaxNgb));

		trace_only_tf(ospf_trace_options, 0,
("\tInvervals: Hello %ds, Dead Router %ds, Designed Router %A, Backup %A",
                        ntohs(pck_hello->pck_hello_int),
                        ntohl(pck_hello->pck_hello_rtdead),
                        sockbuild_in(0, pck_hello->pck_hello_dr),
                        sockbuild_in(0, pck_hello->pck_hello_backupDr)));
 
                tracef("\t");
                for(nbNgb = 0; nbNgb < MaxNgb;) {
			tracef("\tNeighbours: ");
			for(i = 0; i < 4 && nbNgb < MaxNgb; i++, nbNgb++)
                        	tracef("%A ", sockbuild_in(0,
                                	pck_hello->pck_hello_ngb[nbNgb]));
                	trace_only_tf(ospf_trace_options, 0, (""));
		}
                break;
 
        case OSPF_PCK_DD:
                if (!TRACE_TF(ospf_trace_options,
                                TR_OSPF_PCK_ALL|TR_OSPF_PCK_DD))
                        break;
                pck_dd = (ospf_dd_pck_t *)pck;
                trace_only_tf(ospf_trace_options, 0,
                        ("\tOptions <%s> (%X), Flags <%s> (%X), Sequence %d",
                        trace_bits(ospf_pck_options, pck_dd->pck_dd_options),
                        pck_dd->pck_dd_options,
                        trace_bits(ospf_pck_flags, pck_dd->pck_dd_flags),
                        pck_dd->pck_dd_flags,
                        ntohl(pck_dd->pck_dd_seq)));

                /* Print all the LS in the packet.
                 */
                (ospf_dd_pck_t *)lsh = pck_dd + 1;
                nblsh = (ntohs(pck_dd->pck_dd_len) - sizeof(ospf_dd_pck_t)) /
                        sizeof(ospf_lsa_hdr_t);
                for(; nblsh; nblsh--, lsh++)
                        ospf_trace_lsh(lsh, "\t\t");
                break;
 
        case OSPF_PCK_LSR:
                if (!TRACE_TF(ospf_trace_options,
                                TR_OSPF_PCK_ALL|TR_OSPF_PCK_LSR))
                        break;
                lsr = ((ospf_lsr_pck_t *)pck)->pck_lsr_data;
                nblsr = PCK_NBLSR(ntohs(pck->pck_len));
                for(; nblsr; nblsr--, lsr++)
                        trace_only_tf(ospf_trace_options, 0,
                                ("\tType %s (%d) ID %A Advertising Router %A",
                                trace_value(ospf_lsa_types,
                                        ntohl(lsr->lsrd_type)),
                                ntohl(lsr->lsrd_type),
                                sockbuild_in(0, lsr->lsrd_id),
                                sockbuild_in(0, lsr->lsrd_advrt)));
                break;
 
        case OSPF_PCK_LSU:
                if (!TRACE_TF(ospf_trace_options,
                                TR_OSPF_PCK_ALL|TR_OSPF_PCK_LSU))
                        break;
                nblsa = ntohl(((ospf_lsu_pck_t *)pck)->pck_lsu_nbadv);
                trace_only_tf(ospf_trace_options, 0,
			("\tNumber of Link State: %d:", nblsa));
                lsa = ((ospf_lsu_pck_t *)pck)->pck_lsu_lsa;
 
                for(; nblsa; nblsa--) {
                        ospf_trace_lsa(lsa, "\t");
                        ASSERT(lsa->lsa_type);
                        (int)lsa += (int)ntohs(lsa->lsa_len);
                }
                break;
        case OSPF_PCK_LSA:
                if (!TRACE_TF(ospf_trace_options,
                                TR_OSPF_PCK_ALL|TR_OSPF_PCK_LSA))
                        break;
                nblsh = PCK_LSA_NBLSH(ntohs(pck->pck_len));
                ASSERT(nblsh < 100);
                lsh = ((ospf_lsa_pck_t *)pck)->pck_lsa_lsh;
                for(; nblsh; nblsh--) {
                        ospf_trace_lsh(lsh, "\t");
                        lsh++;
                }
                break;
        default:
                break;
        }
}
 
/* trace a LSA header. The LSA header is in the network byte order.
 */
void
ospf_trace_lsh __PF2(lsh, ospf_lsa_hdr_t *, msg, char *)
{
	vertex_t vtx;

	vtx.vtx_type = lsh->lsh_type;
	vtx.vtx_id = lsh->lsh_id;
	vtx.vtx_advrt = lsh->lsh_advrt;
	vtx.vtx_rt = NULL;

        trace_only_tf(ospf_trace_options, 0,
("%sLS %s Age %d Sequence %x CkSum %x Lenght %d Options <%s> (%x)",
                msg, ls_vtx_name(&vtx),
                ntohs(lsh->lsh_age),
                ntohl(lsh->lsh_seq), ntohs(lsh->lsh_cksum),
                ntohs(lsh->lsh_len),
                trace_bits(ospf_pck_options, lsh->lsh_options),
                lsh->lsh_options));
}
 
/* Nice print of a vertex.
 */
void
ospf_trace_vtx __PF2(vtx, vertex_t *, msg, char *)
{
        ospf_vtxLink_t *vl;
 
        if (!TRACE_TF(ospf_trace_options, TR_OSPF_DB))
                return;
 
        /* General info...
         */
        trace_only_tf(ospf_trace_options, 0,
("OSPF %s: VTX %s bit [%d:%X] Age %d Seq %x CkSum %x Flags <%s>",
                msg, ls_vtx_name(vtx), vtx->vtx_bitindex.bi_byte,
		vtx->vtx_bitindex.bi_bits, VTX_AGE(vtx),
                vtx->vtx_seq, vtx->vtx_cksum,
                trace_bits(ospf_vtx_flags, vtx->vtx_flags)));
 
        trace_only_tf(ospf_trace_options, 0,
                ("\tNode %p Left %p Right %p", vtx, vtx->vtx_node_left,
                vtx->vtx_node_right));
 
        tracef(
"\tArea %A Gateway %A Nb Routes %d install %d time %d NextRecv %d SPFcount %d opd %d RefCt %d",
                sockbuild_in(0, VTX_AREA(vtx)->area_id),
                (vtx->vtx_gw->gw_addr)?
                        vtx->vtx_gw->gw_addr:
                        sockbuild_str("NONE"),
                vtx->vtx_gw->gw_n_routes,
                vtx->vtx_install, time_sec, vtx->vtx_nextRecv,
                vtx->vtx_spfcount, vtx->vtx_options, vtx->vtx_qref);
 
        /* Links...
         */
        switch(vtx->vtx_type) {
        case LST_RTR:
                trace_only_tf(ospf_trace_options, 0, (" bits <%x>",
                        vtx->vtx_rtr_bits));
                break;
        case LST_NTW:
                trace_only_tf(ospf_trace_options, 0, (" mask %A",
                        vtx->vtx_rt? sockbuild_in(0, vtx->vtx_mask):
                                sockbuild_str("<NONE>")));
                break;
        case LST_ASBR:
        case LST_ASIP:
                trace_only_tf(ospf_trace_options, 0, (" TOS %d Metric %d",
                        SLAMETRIC_TOS(vtx->vtx_tosmet),
                        SLAMETRIC_METRIC(vtx->vtx_tosmet)));
                break;
        case LST_ASEX:
                trace_only_tf(ospf_trace_options, 0, (""));
                trace_only_tf(ospf_trace_options, 0,
                        ("\tType %d TOS %d Metric %d Forward %A",
                        EXTMETRIC_TYPE(vtx->vtx_tosmet),
                        EXTMETRIC_TOS(vtx->vtx_tosmet),
                        EXTMETRIC_METRIC(vtx->vtx_tosmet),
                        sockbuild_in(0, vtx->vtx_forward)));
                break;
        default:
                ASSERT(0);
        }
 
        if (vtx->vtx_rt) {
                trace_only_tf(ospf_trace_options, 0,
                ("\t%sRoute %A (%p) From %A Cost %d <%s> Tag %d",
                        (vtx->vtx_rt == ospf_nullrt)?
                                "*** Junk route: Ignore *** ": "",
                        vtx->vtx_rt->rt_dest,
                        vtx->vtx_rt,
                        vtx->vtx_rt_gw && vtx->vtx_rt_gw->gw_addr?
                                vtx->vtx_rt_gw->gw_addr:
                                sockbuild_str("NONE"),
                        vtx->vtx_cost,
                        trace_bits(rt_state_bits, vtx->vtx_rt->rt_state),
                        vtx->vtx_tag));
        }
        else
                trace_only_tf(ospf_trace_options, 0,
                        ("\t[No Route. Self originated LS or Incomplete LS]"));
 
        if (vtx->vtx_type == LST_RTR || vtx->vtx_type == LST_NTW)
                for(vl = vtx->vtx_links; vl; vl = vl->vl_next)
                        trace_only_tf(ospf_trace_options, 0,
                                ("\t -> %s (%d) %A cost %d data %A <%s>",
                                trace_state(ospf_lslinks_types, vl->vl_type),
                                vl->vl_type,
                                vl->vl_vtx?
                                        sockbuild_in(0, vl->vl_id):
                                        sockbuild_str("<Unavailable>"),
                                vl->vl_cost,
                                sockbuild_in(0, vl->vl_data),
                                trace_bits(ospf_vtx_flags, vl->vl_flags)));
}
 
/* trace a LSA. LSA is in network byte order.
 */
void
ospf_trace_lsa __PF2(lsa, ospf_lsa_t *, msg, char *)
{
        int nbRtrs, ind, nbLinks;
        ospf_lsa_rtlink_t *link;
        u_int32 tosmet32;
 
        /* Print the header...
         */
        ospf_trace_lsh(&lsa->lsa_hdr, msg);
 
        switch(lsa->lsa_type) {
        case LST_RTR:
                trace_only_tf(ospf_trace_options, 0,
                        ("%s\t#links %d:", msg, ntohs(lsa->lsa_rtr_nblinks)));
                nbLinks = ntohs(lsa->lsa_rtr_nblinks);
                link = lsa->lsa_rtr_links;
 
                for(; nbLinks; nbLinks--) {
                        trace_only_tf(ospf_trace_options, 0,
                        ("%s\tType %s (%d) ID %A Data %A NbTOS %d Cost %d",
                                msg,
                                trace_value(ospf_lslinks_types,link->lsrl_type),
                                link->lsrl_type,
                                sockbuild_in(0, link->lsrl_id),
                                sockbuild_in(0, link->lsrl_data),
                                link->lsrl_nbtos, ntohs(link->lsrl_metric)));
 
                        (int)link += sizeof(ospf_lsa_rtlink_t)
                                + (link->lsrl_nbtos - 1) *
                                        sizeof(ospf_tosmetric_t);
                }
                break;
 
        case LST_NTW:
                nbRtrs = LSA_NTW_NBRTR(lsa, ntohs(lsa->lsa_len));
                trace_only_tf(ospf_trace_options, 0,
                        ("%s\tMask %A, %d routers:", msg,
                                sockbuild_in(0, lsa->lsa_ntw_mask),
                                nbRtrs));
                for(ind = 0; ind < nbRtrs; ind++)
                        trace_only_tf(ospf_trace_options, 0,
                        ("%s\t\tRouter %A", msg,
                        sockbuild_in(0, lsa->lsa_ntw_rtid[ind])));
                break;
 
        case LST_ASBR:
        case LST_ASIP:
                tosmet32 = ntohl(lsa->lsa_sla_tosmet32);
 
                trace_only_tf(ospf_trace_options, 0,
                        ("%s\tMask %A, TOS %d Metric %d", msg,
                                sockbuild_in(0, lsa->lsa_sla_mask),
                                SLAMETRIC_TOS(tosmet32),
                                SLAMETRIC_METRIC(tosmet32)));
                break;
 
        case LST_ASEX:
                tosmet32 = ntohl(lsa->lsa_ext_tosmet32);
 
                trace_only_tf(ospf_trace_options, 0,
                ("%s\tMask %A, TOS %d Type %d Metric %d Forward %A Tag %X",
                        msg,
                        sockbuild_in(0, lsa->lsa_ext_mask),
                        EXTMETRIC_TOS(tosmet32),
                        EXTMETRIC_TYPE(tosmet32),
                        EXTMETRIC_METRIC(tosmet32),
                        sockbuild_in(0, lsa->lsa_ext_forw),
                        ntohl(lsa->lsa_ext_tag)));
                break;
 
        default:
                break;
        }
}
 
/* Print the name/type of a vertex.
 */
char *
ls_vtx_name __PF1(vtx, vertex_t *)
{
#define VTXNAMEMAX      5
#define LS_MAX_VTXLEN   80
        static char lbuff[VTXNAMEMAX][LS_MAX_VTXLEN];
        static int ind = 0;
 
        ind = (ind + 1) % VTXNAMEMAX;
 
	if (!LS_BOUND(vtx->vtx_type)) {
		(void)sprintf(lbuff[ind], "(%d)???", vtx->vtx_type);
		return(lbuff[ind]);
	}

        switch(vtx->vtx_type) {
        case LST_NTW:
        case LST_ASEX:
        case LST_ASIP:
                (void)sprintf(lbuff[ind], "%s:%A/%d [%A]",
                        trace_state(ospf_lsa_types, vtx->vtx_type),
                        sockbuild_in(0, vtx->vtx_id),
                        vtx->vtx_rt? mask_bits_32(vtx->vtx_mask): 0,
			sockbuild_in(0, vtx->vtx_advrt));
                break;
        case LST_RTR:
        case LST_ASBR:
                (void)sprintf(lbuff[ind], "%s:%A [%A]",
                        trace_state(ospf_lsa_types, vtx->vtx_type),
                        sockbuild_in(0, vtx->vtx_id),
			sockbuild_in(0, vtx->vtx_advrt));
                break;
        default:
                ASSERT(0);
        }
 
        return(lbuff[ind]);
}
 

