/*
 * DRET-IPv6
 *
 * Contacts:	
 *              INRIA      <Christophe.Diot@sophia.inria.fr>
 *              MASI       <Eric.Horlait@masi.ibp.fr>
 *              This software has been developped with the financial support
 *              of DRET (French Military Research Agency).
 *
 * Version:    $Id: ripng.h,v 1.3 1996/12/18 10:57:50 eh Exp $ 
 *	
 * Authors:	
 *	        Frank Zago
 *
 * Description:
 *
 *              ripng.h is derived from rip.h
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *              This notice must be copied with the distributed package.
 *
 */				

#ifdef PROTO_RIPNG

#define	RIPNG_VERSION_1	1

#define RIPNG_METRIC_UNREACHABLE	16
#define	RIPNG_METRIC_SHUTDOWN	(RIPNG_METRIC_UNREACHABLE - 1)
#define	RIPNG_METRIC_NEXTHOP	0xff

#define	RIPNGTF_MCSET	TARGETF_USER1 /* MC has been enabled on this interface */
#define	RIPNGTF_POLL	TARGETF_USER2 /* must send a poll packet */
#define	RIPNGTF_V1	TARGETF_USER1 /* RIPng v1 */

/* RIPng Draft section 2.5 */
extern struct in6_addr INET6_RIPNG_ALLROUTERS;

/* Structures of packets: header and route table entry (RTE) */
struct ripng_hdr {
	u_int8	ripng_cmd;	/* request/response */
	u_int8	ripng_vers;	/* protocol version # */
	u_int16	ripng_zero;	/* unused, always 0 */
};

struct ripng_rte {
	struct in6_addr prefix;	/* IPv6 prefix */
	u_int16 rte_tag;	/* route tag */
	u_int8 preflen;		/* prefix length */
	u_int8 metric;		/* metric */
};

struct ripng_packet {
	struct ripng_hdr header;
	struct ripng_rte rte[0]; /* variable number of rte */
};

/*
 * Packet types (RIPng draft section 2.1)
 */
#define	RIPNGCMD_REQUEST	1	/* want info */
#define	RIPNGCMD_RESPONSE	2	/* responding to request */
#define	RIPNGCMD_MAX		3

/* FIXME: wrong if authentication header present */
#define	RIPNG_MAXSIZE(ifap)	(ifap->ifa_mtu - sizeof(struct ipv6hdr) - sizeof (struct udphdr) - sizeof(struct ripng_hdr))

#define RIPNG_T_UPDATE	(time_t) 30
#define	RIPNG_T_HOLDOWN	(time_t) 120
#define	RIPNG_T_FLASH	(time_t) (grand((u_int32) (RIPNG_T_MAX - RIPNG_T_MIN + 1)) + RIPNG_T_MIN)
#define	RIPNG_T_MAX	(time_t) 5
#define	RIPNG_T_MIN	(time_t) 1
#define	RIPNG_T_EXPIRE	(time_t) 180

#define RIPNG_PORT	521
#define	RIPNG_HOP	1	/* Minimum hop count when passing through */
#define	RIPNG_LIMIT_METRIC	RIPNG_HOP, RIPNG_METRIC_UNREACHABLE	/* For parser */

#define	RIPNG_CONFIG_IN			1
#define	RIPNG_CONFIG_OUT		2
#define	RIPNG_CONFIG_METRICIN		3
#define	RIPNG_CONFIG_METRICOUT		4
#define	RIPNG_CONFIG_FLAG		5
#define	RIPNG_CONFIG_MAX		6

extern flag_t ripng_flags;		/* Option flags */
extern trace *ripng_trace_options;	/* RIP tracing configuration */
extern metric_t ripng_default_metric;	/* Default metric to use when propogating */
extern pref_t ripng_preference;		/* Preference for RIP routes */
extern int ripng_n_trusted;		/* Number of Trusted RIP gateways */
extern int ripng_n_source;		/* Number of gateways to receive explicate RIP info */
extern adv_entry *ripng_import_list;	/* List of nets to import and not import */
extern adv_entry *ripng_export_list;	/* List of nets to export */
extern adv_entry *ripng_int_policy;	/* List of interface policy */
extern gw_entry *ripng_gw_list;		/* List of RIPng gateways */
extern const bits ripng_trace_types[];	/* List of RIPng specific trace flags*/

/* Values for rip_flags */
#define	RIPNGF_ON		BIT(0x01)		/* RIPng is enabled */
#define	RIPNGF_SOURCE		BIT(0x02)		/* Source packets to our peers */
#define	RIPNGF_FLASHDUE		BIT(0x04)		/* Flash update is due */
#define	RIPNGF_NOFLASH		BIT(0x08)		/* Can not do a flash update until after the next normal update */
#define	RIPNGF_RECONFIG		BIT(0x10)		/* Initial processing or reconfiguration */
#define	RIPNGF_TERMINATE	BIT(0x20)		/* RIPng is terminating */
#define RIPNGF_MULTI            BIT(0x40)               /* Supply on all interfaces */

#define	RIPNG_HOLDCOUNT		(RIPNG_T_HOLDOWN/RIPNG_T_UPDATE)	/* Number of updates per garbage collection */

#define	ifa_ripng_mccount	ifa_ps[RTPROTO_RIPNG].ips_datas[0]

#define	RIPNG_IFPS_NOMC	IFPS_KEEP1	/* Unable to enable MC on this IF */

/* Tracing */
#define	TR_RIPNG_INDEX_PACKETS	0	/* All packets */
#define	TR_RIPNG_INDEX_REQUEST	1	/* Request packets (REQUEST) */
#define	TR_RIPNG_INDEX_RESPONSE	2	/* Response packets (RESPONSE) */

#define	TR_RIPNG_DETAIL_REQUEST	 TR_DETAIL_1
#define	TR_RIPNG_DETAIL_RESPONSE TR_DETAIL_2

/**/
PROTOTYPE(ripng_init,
	  extern void,
	  (void));
PROTOTYPE(ripng_var_init,
	  extern void,
	  (void));
PROTOTYPE(ripng_config_free,
	  extern void,
	  (config_entry *));

#if defined(PROTO_SNMP) && defined(MIB_RIPNG)
u_int ripng_global_changes;
u_int ripng_global_responses;

PROTOTYPE(ripng_init_mib,
	  extern void,
	  (int));
#endif	/* defined(PROTO_SNMP) && defined(MIB_RIP) */

#endif	/* PROTO_RIPNG */











